Require Import WeakSCAP.
Require Import LinLog.
Require Import Machine.
Require Import List.
Require Import SpecUtil.
Require Import GC.
Require Import NatListSet.
Require Import NatUtil.
Require Import EqNat.
Require Import StateTheory.


(** ** Labels and register aliases *)

(** Mark stack operations *)
Notation STACKEMPTY := 17.
Notation STACKEMPTYT := 18.
Notation STACKPUSH := 19.
Notation STACKLOOP := 20.
Notation STACKPOP := 21.

(** Registers *)
Notation rtemp := ast.
Notation rtop := t2.
Notation rbot := t3.
Notation rbuff := t4.


(** ** Basic heap specs *)

Fixpoint arrayStack (h : nat) (elems : list nat) {struct elems} : hpred :=
  match elems with
    | nil => unit
    | x::elems' => h |-> x ** arrayStack (h - addrOffset) elems'
  end.

Definition buffArrayStack x y z E :=
  !(x + addrOffset * length E = y /\ y <= z) **
  arrayStack (y - addrOffset) E ** buffer y z.

Definition stackHp objs s :=
  buffArrayStack (rget s rbot) (rget s rtop) (rget s rbuff) objs.

Notation stackPre :=
  (fun s => exists objs, s |||- stackHp objs s ** top).


(** *** Mark stack operations *)


(** isEmpty() *)
Definition isEmptyPre := stackPre.

Definition isEmptyGuar s s' :=
  (forall objs A, s |||- stackHp objs s ** A ->
    sbool (rget s' v0) (objs = nil)) /\ 
  heapOf s = heapOf s' /\
  (forall r, r <> v0 -> rget s r = rget s' r).

Definition isEmptyImpl :=
  beq rbot rtop STACKEMPTYT;;
  addiu v0 zero 0;
  jr ra.

Definition isEmptyTPre s :=
  s |||- stackHp nil s ** top.

Definition isEmptyTGuar s s' :=
  rget s' v0 = 1 /\ heapOf s = heapOf s' /\
  (forall r, r <> v0 -> rget s r = rget s' r).

Definition isEmptyTImpl :=
  addiu v0 zero 1;
  jr ra.


(** push() *)
Definition pushPre := stackPre.

Definition pushGuar s s' :=
  (forall objs A, s |||- stackHp objs s ** A -> 
    s' |||- stackHp ((rget s a0)::objs) s' ** A) /\
  (forall r, r <> rtop -> r <> rtemp -> rget s r = rget s' r).

Definition pushImpl :=
  addiu rtemp rtop 4;
  sltu rtemp rbuff rtemp;
  bne rtemp zero STACKLOOP;;
  sw a0 0 rtop;
  addiu rtop rtop 4;
  jr ra.

(* infinite loop, in case we run out of space *)

Definition loopPre (s : state) := True.
Definition loopGuar (s s' : state):= False.

Definition loopImpl :=
  j STACKLOOP.


(** pop() *)
Definition popPre s :=
  exists objs, (objs <> nil) /\
    s |||- stackHp objs s ** top.

Definition popGuar s s' :=
  (forall objs A v, s |||- stackHp (v::objs) s ** A -> 
    s' |||- stackHp objs s' ** A /\ rget s' v0 = v) /\
  (forall r, r <> v0 -> r <> rtop -> rget s r = rget s' r).

Definition popImpl :=
  addiu v0 zero 4;
  subu rtop rtop v0;
  lw v0 0 rtop;
  jr ra.


(** ** Local specification map *)

Definition stackP f :=
  if EqNat.beq_nat STACKLOOP f then
    SomeT (bspeci loopPre loopGuar)
    else if EqNat.beq_nat STACKEMPTYT f then
      SomeT (bspeci isEmptyTPre isEmptyTGuar)
      else
	NoneT _.

Ltac PsimplStack := 
  unfold stackP; 
  rewrite beq_true; 
  repeat (rewrite beq_false; [idtac | discriminate]); 
  reflexivity.


(** ** Well-formedness theorems *)

Theorem isEmptyImplOk : 
  iseqOkS stackP isEmptyPre isEmptyGuar isEmptyImpl.
Proof.
  unfold isEmptyImpl.
  iseqWeakSolver PsimplStack.
  
  unfold isEmptyPre; unfold isEmptyGuar; 
    unfold isEmptyTPre; unfold isEmptyTGuar.
  unfold stackHp; unfold buffArrayStack.
  intros s [objs H].
  assert (H' := H).
  linLogElimOld H.
  unfold doBtest.
  compare (rget s rbot) (rget s rtop);intro e.
  rewrite e;rewrite beq_true.
  assert (objs = nil).
  destruct H2;rewrite e in H0.
  assert (length objs = 0).
  omega.
  induction objs.
  reflexivity.
  simpl in H2;discriminate H2.
  
  split.
  generalize H'.
  rewrite <- H0;rewrite <- e.
  auto.

  clear H' H2 H objs H0.
  intros s' [Hv0 [Hhp Hr]].
  split;auto.
  intros.
  linLogElimOld H.
  assert (objs = nil).
  assert (length objs = 0).
  omega.
  induction objs.
  reflexivity.
  simpl in H0;discriminate H0.
  rewrite Hv0;auto.

  rewrite beq_false;auto.
  rewrite addiuStep;stSimpl.
  split;auto.

  clear H' H2 H objs.
  simpl;intros.
  subst s';split.
  intros.
  stSimpl; simpl.
  linLogElimOld H.
  assert (length objs <> 0).
  omega.
  intro H';apply H0;subst objs;simpl;auto.
  split;auto.
  intros; stSimpl; reflexivity.
Qed.

Theorem isEmptyTImplOk : 
  iseqOkS stackP isEmptyTPre isEmptyTGuar isEmptyTImpl.
Proof.
  unfold isEmptyTImpl.
  iseqWeakSolver PsimplStack.
  
  unfold isEmptyTPre; unfold isEmptyTGuar.
  unfold stackHp; unfold buffArrayStack.
  
  intros s [Hra H].
  rewrite addiuStep;stSimpl.
  split;auto.
  simpl;intros s' He;subst s';stSimpl.
  intuition.
  stSimpl; reflexivity.
Qed.

Theorem loopImplOk : 
  iseqOkS stackP loopPre loopGuar loopImpl.
Proof.
  unfold loopImpl.
  iseqWeakSolver PsimplStack.
  tauto.
Qed.

Theorem popImplOk : iseqOkS stackP popPre popGuar popImpl.
Proof.
  unfold popImpl.
  iseqWeakSolver PsimplStack.

  unfold popPre; unfold popGuar; unfold stackHp.
  unfold buffArrayStack; 
    unfold buffer; unfold stackHp;
      unfold buffArrayStack; unfold buffer.

  intros s [objs [Hobjs Hpred]].
  linLogElimOld Hpred.

  rewrite addiuStep.
  stSimpl.
  rewrite subuStep.
  stSimpl.
  replace (4 + 0) with 4;auto.
  destruct objs as [ | x objs'].
  elimtype False; apply Hobjs; reflexivity.
  simpl in Hpred.
  assert (H := sconjAssocL Hpred);clear Hpred.
  rewrite (lwStep H).  
  stSimpl.
  split;auto.

  intros s' Hs'.
  rewrite <- Hs'; stSimpl.
  assert (Heap.get (heapOf s) (rget s rtop - 4) = Some x).
  eapply contGetEqL.
  apply H.
  rename H0 into XXX.
  clear Hobjs H.
  split.
  intros objs A v Hpred.
  linLogElimOld Hpred.
  split.
  linLogIntroOld.
  destruct H1 as [HA HB].
  simpl in H2; omega.
  simpl in Hpred.
  apply sconjAssocL.
  eapply sconjImplL.
  apply (sconjAssocR Hpred).
  clear Hpred;intros h H.
  eapply sconjImplR.
   apply (sconjAssocL (sconjComm (sconjAssocL H))).
  clear H;intros h0 H.
  apply fasEqvSet with (set_add (rget s rtop - 4) 
    (addrRange (rget s rtop) (rget s rbuff))).
  apply fasIntro.
  apply sconjComm.
  linLogIntroOld.
  exists v;apply H.
  intro Hab. 
  destruct (addrRangeElim Hab) as [X1 [k X2]].
  simpl in H2.
  omega.
  constructor;intros.
  destruct H2;simpl in H1;
  simpl in H0;destruct H0. 
  apply addrRangeIntro with (k := 0);
    omega.
  decompose [and ex] (addrRangeElim H0).
  apply addrRangeIntro with (k := x0 + 1);
    omega.
  decompose [and ex] (addrRangeElim H0).
  destruct x0.
  left; subst a; auto.
  simpl in H2; inversion_clear H2.
  right;apply addrRangeIntro with (k := x0);
    omega.
  assert (Heap.get (heapOf s) (rget s rtop - 4) = Some v).
  unfold arrayStack in Hpred; fold arrayStack in Hpred.
  assert (H := sconjAssocL Hpred);clear Hpred.
  eapply contGetEqL.
  apply H.
  rewrite H in XXX; inversion_clear XXX;reflexivity.
  
  intros r He1 He2.
  stSimpl; reflexivity.
  reflexivity.
  stSimpl;auto.
Qed.

    
Theorem pushImplOk : 
  iseqOkS stackP pushPre pushGuar pushImpl.
Proof.
  unfold pushImpl.
  iseqWeakSolver PsimplStack.
  
  unfold pushPre; unfold pushGuar; 
    unfold loopPre; unfold loopGuar.
  unfold stackHp; unfold buffArrayStack; unfold buffer.

  intros s [objs Hpred].
  linLogElimOld Hpred.
  rewrite addiuStep; stSimpl.
  rewrite sltuStep; stSimpl.

  doBtestSimpl.

  split;auto.
  intros;elimtype False;auto.

  assert (H := sconjAssocL (sconjComm Hpred));clear Hpred.
  assert (s |||- (((rget s rtop) |->_) ** 
    foralls z in addrRange 
    (rget s rtop + 4) (rget s rbuff), z |->_) **
    top ** arrayStack (rget s rtop - 4) objs).
  eapply sconjImplL.
  apply H.
  intros h Hp.
  assert (set_In (rget s rtop) 
    (addrRange (rget s rtop) (rget s rbuff))).
  apply addrRangeIntro with (k := 0); trivial.
  hypReplace Hp (fasElim H0 Hp).
  apply (sconjImplR Hp).
  clear h Hp; intros h Hp.
  apply (fasEqvSet _ _ Hp (addrRangeRemoveMin 
    (rget s rtop) (rget s rbuff))).

  assert (Hpred := sconjAssocL H0);clear H0.
  linLogElimOld Hpred.

  rewrite (swStep Hpred);auto.
  rewrite addiuStep.
  stSimpl.
  split;auto.
  clear H1 H Hpred x objs.
  intros s' Hp.
  rewrite <- Hp;stSimpl.
  split.
  intros objs A H.
  linLogElimOld H.
  linLogIntroOld.
  simpl.
  omega.
  simpl.
  replace (rget s rtop - 0) with (rget s rtop).
  2:omega.
  repeat (apply sconjAssocR).
  assert (H' := sconjAssocL (sconjComm H));clear H.
  assert (s |||- (((rget s rtop) |->_) ** 
    foralls z in addrRange (rget s rtop + 4) 
    (rget s rbuff), z |->_) **
    A ** arrayStack (rget s rtop - 4) objs).
  eapply sconjImplL.
  apply H'.
  intros h Hp'.
  assert (set_In (rget s rtop) 
    (addrRange (rget s rtop) (rget s rbuff))).
  apply addrRangeIntro with (k := 0); trivial.
  hypReplace Hp' (fasElim H Hp').
  apply (sconjImplR Hp').
  clear h Hp'; intros h Hp'.
  apply (fasEqvSet _ _ Hp' (addrRangeRemoveMin 
    (rget s rtop) (rget s rbuff))).

  assert (Hpred := sconjAssocL H);clear H.
  linLogElimOld Hpred.
  apply contUpdateSt with x.
  clear H'.
  assert (H' := sconjAssocR (sconjComm 
    (sconjAssocR (sconjAssocR Hpred))));clear Hpred.
  repeat (apply sconjAssocL).
  eapply sconjImplL.
  apply H'.
  clear H';intros h H.
  assert (H' := sconjAssocR H);clear H.
  eapply sconjImplL.
  replace (S (S (S (S (rget s rtop))))) with
    (rget s rtop + 4).
  apply H'.
  omega.
  clear H';intros h0 H.
  apply sconjComm;auto.
  
  intros; stSimpl; reflexivity.
Qed.
