Require Import WeakSCAP.
Require Import LinLog.
Require Import Machine.
Require Import List.
Require Import SpecUtil.
Require Import Arith.
Require Import GC.
Require Import NatUtil.
Require Import StateTheory.
Require Import MarkSweepGC.
Require Import MarkSweepLemma.
Require Import BuffArrayStack.
Require Import StackSet.
Require Import IMarkSweepGC.

(* Barriers and shades *)
Notation ADDHDRETURN := 700.
Notation SHADEWHITE := 701.
Notation YUSSABARRIER := 702.
Notation YUSSABARRIERRET := 703.

(** *** shadeWhite() *)
(* shadeWhite pre-condition:
  1. MARK_PHASE /\ arg0 may be in black, gray or white set
*)

Definition shadeWhitePreP s ohpSt ohpEnd B G W free A :=
  rget s rphase = MARK_PHASE /\
  set_eqv ((B |_| G) |_| (W |_| free)) (rangeMSObjs ohpSt ohpEnd) /\
  set_In (rget s a0) ((B |_| G) |_| W) /\
  s |||- objHdrs B MARKED ** objHdrs G MARKED ** objHdrs (W |_| free) UNMARK **
    setHp G s ** A.

Definition shadeWhitePre := fun s =>
  exists ohpSt, exists ohpEnd, exists B,
    exists G, exists W, exists free,
      shadeWhitePreP s ohpSt ohpEnd B G W free top.

(* shadeWhite guarantees:
  1. MARK_PHASE /\ arg0 is in white set, shade it gray
    
  2. otherwise, nothing changes.
*)
Definition shadeWhiteGuar s s' :=
  (forall ohpSt ohpEnd B G W free A,
    shadeWhitePreP s ohpSt ohpEnd B G W free A ->
    (set_In (rget s a0) W -> 
      s' |||- objHdrs B MARKED ** objHdrs (set_add (rget s a0) G) MARKED **
	objHdrs (set_remove (rget s a0) (W |_| free)) UNMARK **
	setHp (set_add (rget s a0) G) s' ** A) /\
    (~ set_In (rget s a0) W ->
      heapOf s' = heapOf s /\ rget s' rtop = rget s rtop)) /\
  (forall r, r <> rtemp -> r <> rtop -> r <> rcst3 -> 
    rget s' r = rget s r).

(* new implementation of shade and yussa barrier 

  void shadeWhite (obj_t* obj) {
   if (obj->color == WHITE) {   
     obj->color = BLACK;        
     stack_push(obj);           
   }                            
  }                             
                                
  void yuasaStore (obj_t** field, obj_t* newVal) {
    if (phase == MARK_PHASE && !isAtom( *field )) 
      shadeWhite( *field);
    *field = newVal;
  } *)

Definition shadeWhiteImpl :=
  addiu rcst3 ra 0;
  addiu r31 zero HDSIZE;
  subu a0 a0 r31;
  lw rtemp 0 a0;
  addiu r31 zero MARKED;
  beq rtemp r31 ADDHDRETURN;; 
  sw r31 0 a0;
  addiu a0 a0 HDSIZE;
  jal STACKPUSH RETURN3.

Definition addreturnPre (s : state) := True.

Definition addreturnGuar := fun s s' =>
  s' = rset (rset s a0 (HDSIZE + rget s a0)) ra (rget (rset s a0 (HDSIZE + rget s a0)) rcst3).

Definition addreturnImpl :=
  addiu a0 a0 HDSIZE;
  addiu ra rcst3 0;
  jr ra.

Definition shadeWhiteP f :=
  if EqNat.beq_nat STACKPUSH f then
    SomeT (bspeci pushPre pushGuar)
    else if EqNat.beq_nat RETURN3 f then
      SomeT (bspeci return3Pre return3Guar)
      else if EqNat.beq_nat ADDHDRETURN f then
	SomeT (bspeci addreturnPre addreturnGuar)
	else
	  NoneT _.

Ltac PsimplshadeWhite := 
  unfold shadeWhiteP; 
  rewrite beq_true; 
  repeat (rewrite beq_false; [idtac | discriminate]); 
  reflexivity.

Theorem addreturnImplOk :
  iseqOkS shadeWhiteP addreturnPre addreturnGuar addreturnImpl.
Proof.
  unfold addreturnImpl.
  iseqWeakSolver PsimplReturn.

  unfold addreturnPre, addreturnGuar.
  intros.
  rewrite addiuStep.
  rewrite addiuStep.
  split; trivial.
  intros; subst s'.
  trivial.
Qed.

Theorem shadeWhiteImplOk :
  iseqOkS shadeWhiteP shadeWhitePre shadeWhiteGuar shadeWhiteImpl.
Proof.
  unfold shadeWhiteImpl.
  iseqWeakSolver PsimplshadeWhite.

  unfold pushGuar; intros s s' _ [_ Hr].
  apply Hr; discriminate.

  intros s [ohpSt [ohpEnd [B [G [W [free [Hphase [Hset_eq [Hin Hp]]]]]]]]].

  rewrite addiuStep.
  rewrite addiuStep.
  rewrite subuStep; stSimpl.
  
  assert (Hgt : rget s a0 >= 4).
  assert (set_In (rget s a0) (rangeMSObjs ohpSt ohpEnd)).
  elimSetIn.
  eapply MSPtrGtFour.
  apply H.
  reflexivity.

  destruct (set_union_elim Hin) as [HinBG | HinW].
  (* phase == MARK_PHASE /\ a0 is in B or G *)
  assert (s |||- rget s a0 - 4 |-> MARKED ** top).
  destruct (set_union_elim HinBG) as [HinB | HinG].
  hypReplace Hp (sconjAssocL (fasElimL HinB Hp)).
  apply (sconjImplR Hp).
  intros; linLogIntro.
  sconjLift2 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinG Hp)).
  apply (sconjImplR Hp).
  intros; linLogIntro.

  rewrite (lwStep H); trivial.
  rewrite addiuStep; stSimpl.
  unfold doBtest; rewrite beq_true.

  unfold addreturnPre, addreturnGuar.
  split; trivial.

  clear ohpSt ohpEnd free B G W Hset_eq Hin Hp HinBG Hphase.
  intros s' Hs.
  subst s'; stSimpl.
  split.

  intros ohpSt ohpEnd B G W free A [Hphase [Hset_eq [Hin Hp]]].
  unfold setHp; stSimpl.
  split.
  intro HinW.
  sconjLift3 Hp.
  hypReplace Hp (sconjAssocL (fasElimL (set_union_intro1 _ _ free HinW) Hp)).
  assert (Hgeta0 := contGetEqL Hp).
  assert (Hgeta0' := contGetEqL H).
  rewrite Hgeta0' in Hgeta0; inversion_clear Hgeta0.
  intros.
  split; reflexivity.
  intros.
  induction r; trivial.
  elim H0; reflexivity.
  stSimpl; omega.
  elim H2; reflexivity.
  
  (* phase == MARK_PHASE /\ a0 in white ==> shade it gray *)
  sconjLift3 Hp.
  hypReplace Hp (sconjAssocL (fasElimL (set_union_intro1 _ _ free HinW) Hp)).
  rewrite (lwStep Hp); trivial.
  rewrite addiuStep; stSimpl.
  unfold doBtest; rewrite beq_false.
  rewrite (swStep Hp); trivial.
  rewrite addiuStep; stSimpl.
  
  split.
  unfold pushPre, pushGuar, return3Pre, return3Guar, stackHp; stSimpl.
  split; trivial.
  unfold setHp, buffStackSet in Hp.
  hypReplace Hp (contUpdateSt Hp (1 + 0)).
  sconjLift5 Hp.
  linLogElim Hp.
  exists x.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  linLogIntro.
  assert (Ha0 := contGetEqL Hp).

  clear ohpSt ohpEnd B G W free Hset_eq Hin HinW Hp Hphase.
  destruct pushStackIsPushSet as [G1 G2].
  intros s'' [s' [HGpush HGret]].

  split.
  intros ohpSt ohpEnd B G W free A [Hphase [Hset_eq [Hin Hp]]].
  replace (4 + 0) with 4 in HGpush; [idtac | trivial].
  rewrite <-  (le_plus_minus 4 (rget s a0)) in HGpush.
  simpl in HGpush.

  split.
  intros HinW.
  rewrite <- HGret; stSimpl.
  sconjLift3 Hp.
  unfold objHdrs, setHp in Hp.
  assert (HnotinG : ~ set_In (rget s a0) G).
  intro HinG.
  sconjLift3 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinG Hp)).
  assert (Ha0' := contGetEqL Hp).
  rewrite Ha0' in Ha0; inversion Ha0.
  hypReplace Hp (sconjAssocL (fasElimL (set_union_intro1 _ _ free HinW) Hp)).
  hypReplace Hp (contUpdateSt Hp MARKED).
  hypReplace HGpush (G2 _ _ HGpush).
  generalize HGpush; clear HGpush; unfold stackPushGuar, setHp; stSimpl.
  intros [HGpush _].
  sconjLift5 Hp.
  hypReplace Hp (HGpush _ _ Hp HnotinG).
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  sconjLift4 Hp.
  sconjLift3 Hp.
  apply (sconjImpl (sconjAssocR Hp)); clear h Hp; intros h Hp.
  apply objHdrsIntro; trivial.
  sconjLift2 Hp.
  apply Hp.
  intro HnotinW.
  assert (HinBG := set_InUnion_notinR_inL _ _ _ Hin HnotinW).
  assert (s |||- rget s a0 - 4 |-> 1 ** top).
  destruct (set_union_elim HinBG) as [HinB | HinG].
  hypReplace Hp (sconjAssocL (fasElimL HinB Hp)).
  apply (sconjImplR Hp).
  intros; linLogIntro.
  sconjLift2 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinG Hp)).
  apply (sconjImplR Hp).
  intros; linLogIntro.
  assert (Ha0' := contGetEqL H).
  rewrite Ha0' in Ha0; inversion Ha0.
  apply Hgt.

  intros.
  unfold return3Guar in HGret.
  subst s''.
  generalize HGpush; clear HGpush; unfold pushGuar; stSimpl.
  intros [_ Hr].
  rewrite <- Hr.
  compare ra r; intro Hre.
  subst r; trivial.
  stSimpl.
  rewrite <- Hr; trivial.
  compare a0 r; intros Hre'.
  subst r; stSimpl; omega.
  stSimpl; trivial.
  apply beq_rnum_false_neq; reflexivity.
  apply beq_rnum_false_neq; reflexivity.
  omega.
Qed.

(* Yussa' snapshot write barrier *)

(* MARK_PHASE => nothing touch ifreeList *)
Definition BrMPHp s freePtr B G W free A :=
  s |||- objHdrs B MARKED ** objHdrs G MARKED ** objHdrs (W |_| free) UNMARK **
    setHp G s ** objHp (B |_| G) B **
    objHp ((B |_| G) |_| W) (G |_| W) **
    ifreeList freePtr free ** A.

(* ~ MARK_PHASE => may point to B *)
Definition BrNotMPHp s ohpSt ohpEnd freePtr sweepAt B W free A :=
  s |||- bobjHdrs B sweepAt UNMARK MARKED ** objHdrs (W |_| free) UNMARK **
    setHp empty_set s ** objHp B B **
    objHp (rangeMSObjs ohpSt ohpEnd) W **
    ifreeList freePtr free ** A.

Definition BrMPRemoveA0Hp s freePtr B G W free A :=
  s |||- objHdrs B MARKED ** objHdrs G MARKED ** objHdrs (W |_| free) UNMARK **
    setHp G s ** objHp (B |_| G) (set_remove (rget s a0) B) **
    objHp ((B |_| G) |_| W) (set_remove (rget s a0) (G |_| W)) **
    objHp ((B |_| G) |_| W) ((rget s a0)::empty_set) **
    ifreeList freePtr free ** A.
  
Definition BrHp s freePtr B G W free x0 x1 objHp1 objHp2 A :=
  s |||- objHdrs B MARKED ** objHdrs G MARKED ** objHdrs (W |_| free) UNMARK **
    setHp G s ** rget s a0 |-> x0 ** 4 + rget s a0 |-> x1 ** objHp1 ** objHp2 **
    ifreeList freePtr free ** A.

Definition BarrierPreP s ohpSt ohpEnd freePtr B G W free A :=
  (rget s rphase = MARK_PHASE /\
    okFieldVal ((B |_| G) |_| W) (rget s a1) /\
    set_eqv ((B |_| G) |_| (W |_| free)) (rangeMSObjs ohpSt ohpEnd) /\
    set_In (rget s a0) ((B |_| G) |_| W) /\
    BrMPHp s freePtr B G W free A) \/
  (rget s rphase <> MARK_PHASE /\  (* IDLE_PHASE =>  SWEEP_PHASE => *)
    okFieldVal B (rget s a1) /\
    set_eqv (B |_| (W |_| free)) (rangeMSObjs ohpSt ohpEnd) /\
    set_In (rget s a0) B /\ 
    exists sweepAt,
      BrNotMPHp s ohpSt ohpEnd freePtr sweepAt B W free A).

Definition BarrierPre := fun s =>
  exists ohpSt, exists ohpEnd, exists freePtr, exists B, 
    exists G, exists W, exists free,
      BarrierPreP s ohpSt ohpEnd freePtr B G W free top.

Notation yussaBarrierPreP := BarrierPreP.
Notation yussaBarrierPre := BarrierPre.

Definition yussaBarrierRetPre (s : state) :=
  exists v,
    s |||- rget s v0 |-> v ** top.

Definition yussaBarrierRetGuar := fun s s': state =>
  (forall v A,
    s |||- rget s v0 |-> v ** A ->
      s' |||- rget s' a0 |-> rget s a1 ** A) /\
  (forall r, r <> ra -> r <> a0 ->
    rget s' r = rget s r) /\ rget s' a0 = rget s v0 /\
  rget s' ra = rget s rcst2.

Definition yussaBarrierRetImpl :=
  addiu a0 v0 0;
  sw a1 0 a0;
  addiu ra rcst2 0;
  jr ra.

Definition BarrierHpInv s s' ohpSt ohpEnd free:=
  hget s' (rget s a0) = Some (rget s a1) /\
  (forall x, set_In x (rangeMSObjs ohpSt ohpEnd) /\ ~ set_In x free /\ x <> rget s a0 -> 
    hget s x = hget s' x /\ hget s (addrOffset + x) = hget s' (addrOffset + x)).

Definition yussaBarrierGuar s s' :=
  (forall ohpSt ohpEnd freePtr B G W free A,
    yussaBarrierPreP s ohpSt ohpEnd freePtr B G W free A ->
    exists v, hget s (rget s a0) = Some v /\ okFieldVal ((B |_| G) |_| W) v /\
      (rget s rphase = MARK_PHASE ->
	(atomP v -> 
	  BarrierHpInv s s' ohpSt ohpEnd free /\ BrMPRemoveA0Hp s' freePtr B G W free A) /\
	(~ atomP v ->
	  (set_In v W -> 
	    BarrierHpInv s s' ohpSt ohpEnd free /\
	    BrMPHp s' freePtr B (set_add v G) (set_remove v W) free A) /\
	  (~ set_In v W ->
	    BarrierHpInv s s' ohpSt ohpEnd free /\
	    BrMPRemoveA0Hp s' freePtr B G W free A))) /\
      (rget s rphase <> MARK_PHASE ->
	BarrierHpInv s s' ohpSt ohpEnd free /\
	exists sweepAt,
	  BrNotMPHp s' ohpSt ohpEnd freePtr sweepAt B W free A)) /\
  (forall r, r <> rtemp -> r <> rtop -> r <> rcst2 -> r <> rcst3 -> r <> v0 ->
    rget s' r = rget s r).

Definition yussaBarrierImpl :=
  addiu rcst2 ra 0;
  addiu v0 a0 0;
  addiu r31 zero MARK_PHASE;
  bne rphase r31 YUSSABARRIERRET;;
  lw a0 0 a0;
  andi1 r31 a0;
  bne r31 zero YUSSABARRIERRET;;
  jal SHADEWHITE YUSSABARRIERRET.

Definition yussaBarrierP f :=
  if EqNat.beq_nat SHADEWHITE f then
    SomeT (bspeci shadeWhitePre shadeWhiteGuar)
    else if EqNat.beq_nat YUSSABARRIER f then
      SomeT (bspeci yussaBarrierPre yussaBarrierGuar)
      else if EqNat.beq_nat YUSSABARRIERRET f then
	SomeT (bspeci yussaBarrierRetPre yussaBarrierRetGuar)
	else
	  NoneT _.

Ltac Psimplyussa := 
  unfold yussaBarrierP; 
  rewrite beq_true; 
  repeat (rewrite beq_false; [idtac | discriminate]); 
  reflexivity.

Theorem yussaBarrierRetImplOk :
  iseqOkS yussaBarrierP yussaBarrierRetPre yussaBarrierRetGuar yussaBarrierRetImpl. 
Proof.
  unfold yussaBarrierRetImpl.
  iseqWeakSolver PsimplReturn.
  
  unfold yussaBarrierRetPre.
  intros s [v Hp].
  rewrite addiuStep.
  rewrite (swStep Hp); trivial.
  rewrite addiuStep.
  split; trivial.
  clear v Hp.
  intros.
  subst s'.
  split; stSimpl.
  intros v A Hp.
  apply uhsetHOS.
  apply contUpdateSt with v; trivial.
  split; trivial.
  intros r Hra Ha0.
  stSimpl; reflexivity.
  split; reflexivity.
Qed.

Theorem yussaBarrierImplOk :
  iseqOkS yussaBarrierP yussaBarrierPre yussaBarrierGuar yussaBarrierImpl.
Proof.
  unfold yussaBarrierImpl.
  iseqWeakSolver Psimplyussa.
  
  unfold shadeWhiteGuar; intros s s' _ [_ Hr].
  regEqSolver.

  intros s [ohpSt [ohpEnd [freePtr [B [G [W [free Hpre]]]]]]].
  destruct Hpre as [[Hphase [Ha1 [Hset_eq [Hin Hp]]]] | [Hphase [Ha1 [Hset_eq [Hin Hp]]]]].

  rewrite addiuStep.
  rewrite addiuStep.
  rewrite addiuStep; stSimpl.
  (* MARK_PHASE *)
  rewrite Hphase.
  doBtestSimpl.
  elimtype False; apply Bt; reflexivity.
  clear Bt.

  assert (Hpsa0 : exists v, s |||- rget s a0 |-> v ** top /\ okFieldVal ((B |_| G) |_| W) v).
  destruct (set_In_dec (rget s a0) B) as [HinB | HnotinB].
  sconjLift5 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinB Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  exists x.
  split.
  apply (sconjImplR Hp).
  intros; linLogIntro.
  destruct H3; [left; trivial | right; elimSetIn].
  assert (Hin' : set_In (rget s a0) (B |_| (G |_| W))).
  elimSetIn.
  assert (HinGW := set_InUnion_notinL_inR _ _ _ Hin' HnotinB).
  sconjLift6 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinGW Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  exists x.
  split; trivial.
  apply (sconjImplR Hp).
  intros; linLogIntro.
  destruct Hpsa0 as [v [Hpsa0 HFV]].
  rewrite (lwStep Hpsa0); trivial.
  rewrite andi1Step; stSimpl.
  destruct (even_odd_dec v) as [Hveven | Hvodd].
  assert (Hvnotatom : ~ atomP v).
  unfold atomP.
  intro.
  apply not_even_and_odd with v; trivial.
  destruct HFV as [Hatom | Hvin].
  elimtype False; apply Hvnotatom; trivial.
  (* ~ atomP v  /\ MARK_PHASE *)
  rewrite (notAtomNotOdd Hvnotatom).
  doBtestSimpl.
  elimtype False; apply Bt; reflexivity.
  clear Bt Hveven.
  destruct (set_In_dec v W) as [HvinW | HvnotinW].
  assert (HvnotinBG : ~ set_In v (B |_| G)).
  intro HvinBG.
  destruct (set_union_elim HvinBG) as [HvinB | HvinG].
  hypReplace Hp (sconjAssocL (fasElimL HvinB Hp)).
  sconjLift4 Hp.
  hypReplace Hp (sconjAssocL (fasElimL (set_union_intro1 _ _ free HvinW) Hp)).
  sconjLift3 Hp.
  assert (v - 4 <> v - 4).
  eapply sconjNeq.
  linLogIntro.
  apply Hp.
  apply H; reflexivity.
  sconjLift2 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HvinG Hp)).
  sconjLift4 Hp.
  hypReplace Hp (sconjAssocL (fasElimL (set_union_intro1 _ _ free HvinW) Hp)).
  sconjLift3 Hp.
  assert (v - 4 <> v - 4).
  eapply sconjNeq.
  linLogIntro.
  apply Hp.
  apply H; reflexivity.
  assert (Hgetvhd : Heap.get (heapOf s) (v - 4) = Some 0).
  sconjLift3 Hp.
  hypReplace Hp (sconjAssocL (fasElimL (set_union_intro1 _ _ free HvinW) Hp)).
  apply (contGetEqL Hp).
  
  (* MARK_PHASE /\ v in W => shade it gray *)
  split.
  split.
  exists ohpSt; exists ohpEnd; exists B;
    exists G; exists W; exists free.
  split; trivial.
  split; trivial.
  split; trivial.
  unfold setHp; stSimpl.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  linLogIntro.
  
  unfold shadeWhiteGuar, shadeWhitePreP, BrMPHp, setHp.
  stSimpl.
  intros s' [Hg Hr].
  exists v; regRewrite v0; stSimpl; simpl.
  destruct (set_In_dec (rget s a0) B) as [HinB | HnotinB].
  sconjLift5 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinB Hp)). 
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Ha0 := contGetEqL Hp).
  assert (Ha0' := contGetEqL Hpsa0).
  rewrite Ha0' in Ha0; inversion Ha0.
  subst x; clear Ha0' Ha0.
  destruct H3 as [Hatom | Hi].
  elimtype False; apply Hvnotatom; trivial.
  elimtype False; apply HvnotinBG; trivial.
  assert (Hin' : set_In (rget s a0) (B |_| (G |_| W))).
  elimSetIn.
  assert (HinGW := set_InUnion_notinL_inR _ _ _ Hin' HnotinB).
  sconjLift6 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinGW Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Ha0 := contGetEqL Hp).
  assert (Ha0' := contGetEqL Hpsa0).
  rewrite Ha0' in Ha0; inversion Ha0.
  subst x; clear Ha0' Ha0.
  sconjLift7 Hp.
  sconjLift7 Hp.
  sconjLift7 Hp.
  sconjLift7 Hp.
  assert (Htmp : rget s rphase = MARK_PHASE /\
    set_eqv ((B |_| G) |_| (W |_| free)) (rangeMSObjs ohpSt ohpEnd) /\
    set_In v ((B |_| G) |_| W) /\
    BrHp s freePtr B G W free v x0 
    (objHp ((B |_| G) |_| W) (set_remove (rget s a0) (G |_| W)))
    (objHp (B |_| G) B) top).
  split; trivial.
  split; trivial.
  split; trivial.
  destruct (Hg _ _ _ _ _ _ _ Htmp) as  [Ht _].
  hypReplace Hp (Ht HvinW); clear Hg Hr Htmp Ht.
  sconjLift5 Hp.
  apply (sconjImplR Hp).
  intros; linLogIntro.
  
  clear Hp Ha1 Hin Hset_eq freePtr free B G W ohpSt ohpEnd Hvin HvinW HvnotinBG.
  unfold shadeWhiteGuar, shadeWhitePreP, BrMPHp, setHp.
  stSimpl.
  intros s'' [s' [[HGshade Hr] [HGret [Hr' [Ha0'' Hra]]]]].
  generalize HGret.
  rewrite Ha0''.
  regRewrite v0.
  regRewrite a1.
  stSimpl.
  simpl; clear HGret; intro HGret.
  split.
  unfold BarrierPreP.
  intros ohpSt ohpEnd freePtr B G W free A [Hpre | Hpre].
  (* rphase = MARK_PHASE *)
  destruct Hpre as [_ [Ha1 [Hset_eq [Hin Hp]]]].
  exists v.
  assert (Hgeta0 := contGetEqL Hpsa0).
  assert (Hv : okFieldVal ((B |_| G) |_| W) v).
  destruct (set_In_dec (rget s a0) B) as [HinB | HnotinB].
  sconjLift5 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinB Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgeta0' := contGetEqL Hp).
  rewrite Hgeta0' in Hgeta0; inversion Hgeta0.
  subst x.
  destruct H3; [left; trivial | right; elimSetIn].
  assert (Hin' : set_In (rget s a0) (B |_| (G |_| W))).
  elimSetIn.
  assert (HinGW := set_InUnion_notinL_inR _ _ _ Hin' HnotinB).
  sconjLift6 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinGW Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgeta0' := contGetEqL Hp).
  rewrite Hgeta0' in Hgeta0; inversion Hgeta0.
  subst x.
  assumption.
  unfold hget; split; trivial.
  split; trivial.
  split; intro Hphase'.
  split; intro.
  elimtype False; apply Hvnotatom; trivial.
  destruct Hv as [Ha | Hvin].
  elimtype False; apply Hvnotatom; trivial.
  split.
  intro HvinW.
  unfold setHp, BrMPHp, setHp.
  regRewrite rbot.
  regRewrite rtop.
  regRewrite rbuff.
  (* MARK_PHASE /\ v in W *)
  destruct (set_In_dec (rget s a0) B) as [HinB | HnotinB].
  (* MARK_PHASE /\ v in W /\ rget s a0 |-> v /\ rget s a0 in B => false *)
  sconjLift5 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinB Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgeta0' := contGetEqL Hp).
  rewrite Hgeta0' in Hgeta0; inversion Hgeta0.
  subst x; clear Hgeta0 Hgeta0'.
  destruct H4 as [Ha | Hi].
  elimtype False; apply Hvnotatom; trivial.
  destruct (set_union_elim Hi) as [HvinB | HvinG].
  sconjLift4 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HvinB Hp)).
  assert (Hgetvhd' := contGetEqL Hp).
  rewrite Hgetvhd' in Hgetvhd; inversion Hgetvhd.
  sconjLift5 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HvinG Hp)).
  assert (Hgetvhd' := contGetEqL Hp).
  rewrite Hgetvhd' in Hgetvhd; inversion Hgetvhd.
  assert (Hin' : set_In (rget s a0) (B |_| (G |_| W))).
  elimSetIn.
  assert (HinGW := set_InUnion_notinL_inR _ _ _ Hin' HnotinB).
  (* MARK_PHASE /\ v in W /\ rget s a0 |-> v /\ rget s a0 in G |_| W *)
  assert (HvnotinF : ~ set_In v free).
  clear HGshade Hr Hr' HGret.
  intro HvinF.
  sconjLift7 Hp.
  assert (Hf := inFreeListVal _ _ _ _ _ Hp HvinF).
  linLogElim Hf.
  sconjLift7 Hp.
  hypReplace Hp (sconjAssocL (fasElimL (set_union_intro2 _ G _ HvinW) Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  sconjLift2 Hp.
  sconjLift2 Hf.
  assert (Hget := contGetEqL Hp).
  assert (Hget' := contGetEqL Hf).
  rewrite Hget' in Hget; inversion Hget.
  subst x1; clear Hget' Hget Hf Hp.
  destruct H7 as [Ha | Hi].
  inversion_clear Ha.
  inversion_clear H0.
  inversion_clear H1.
  assert (set_In 2 (rangeMSObjs ohpSt ohpEnd)).
  elimSetIn.
  assert (~ set_In INFREELIST (rangeMSObjs ohpSt ohpEnd)).
  eapply inFreeListnotinMSobjs.
  reflexivity.
  apply H1; trivial.

  sconjLift6 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinGW Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgeta0' := contGetEqL Hp).
  rewrite Hgeta0' in Hgeta0; inversion Hgeta0.
  subst x; clear Hgeta0 Hgeta0'.
  sconjLift7 Hp.
  sconjLift7 Hp.
  sconjLift7 Hp.
  sconjLift7 Hp.
  assert (Hpb := Hp).
  assert (Htmp :rget s rphase = MARK_PHASE /\
    set_eqv ((B |_| G) |_| (W |_| free)) (rangeMSObjs ohpSt ohpEnd) /\
    set_In v ((B |_| G) |_| W) /\
    BrHp s freePtr B G W free v x0 
    (objHp ((B |_| G) |_| W) (set_remove (rget s a0) (G |_| W)))
    (objHp (B |_| G) B) A).
  split; trivial.
  split; trivial.
  split; trivial.
  destruct (HGshade _ _ _ _ _ _ _ Htmp) as [Hp' _].
  clear Hp Htmp H; rename Hp' into Hp.
  hypReplace Hp (Hp HvinW).
  sconjLift5 Hp.
  hypReplace Hp (HGret _ _ Hp); clear Hr Hr'.
  split.
  split.
  apply (contGetEqL Hp).
  clear Hphase' Ha0'' Hra Hpsa0 Hgetvhd Hp; rename Hpb into Hp.
  intros x [Hxin [HxnotinF Hex]].
  assert (Hxin' : set_In x (B |_| (G |_| W))).
  apply set_InUnion_notinR_inL with free; trivial.
  assert (set_In x ((B |_| G) |_| W |_| free)).
  setoid_rewrite Hset_eq; assumption.
  elimSetIn.
  destruct (set_union_elim Hxin') as [HxinB | HxinGW]; clear Hxin'.
  sconjLift8 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HxinB Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgetxl := contGetEqL Hp).
  sconjLift2 Hp.
  assert (Hgetxr := contGetEqL Hp).
  sconjLift7 Hp.
  sconjLift7 Hp.
  sconjLift7 Hp.
  sconjLift7 Hp.
  assert (Htmp :rget s rphase = MARK_PHASE /\
    set_eqv ((B |_| G) |_| (W |_| free)) (rangeMSObjs ohpSt ohpEnd) /\
    set_In v ((B |_| G) |_| W) /\
    s |||- objHdrs B 1 ** objHdrs G 1 ** objHdrs (W |_| free) 0 ** setHp G s **
      4 + x |-> x2 ** x |-> x1 ** objHp (B |_| G) (set_remove x B) **
	rget s a0 |-> v ** 4 + rget s a0 |-> x0 **
          objHp ((B |_| G) |_| W) (set_remove (rget s a0) (G |_| W)) **
          ifreeList freePtr free ** A).
  split; trivial.
  split; trivial.
  split; trivial.
  destruct (HGshade _ _ _ _ _ _ _ Htmp) as [Hp' _]; clear HGshade Htmp.
  hypReplace Hp (Hp' HvinW); clear Hp'.
  sconjLift8 Hp.
  hypReplace Hp (HGret _ _ Hp); clear HGret.
  sconjLift7 Hp.
  assert (Hgetxl' := contGetEqL Hp).
  sconjLift7 Hp.
  assert (Hgetxr' := contGetEqL Hp).
  unfold hget.
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.

  assert (HxinGW' : set_In x (set_remove (rget s a0) (G |_| W))).
  apply set_remove_intro; trivial.
  sconjLift7 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HxinGW' Hp)).
  linLogElim Hp.
  assert (Hgetxl := contGetEqL Hp).
  sconjLift2 Hp.
  assert (Hgetxr := contGetEqL Hp).
  sconjLift7 Hp.
  sconjLift7 Hp.
  sconjLift7 Hp.
  sconjLift7 Hp.
  assert (Htmp :rget s rphase = MARK_PHASE /\
    set_eqv ((B |_| G) |_| (W |_| free)) (rangeMSObjs ohpSt ohpEnd) /\
    set_In v ((B |_| G) |_| W) /\
    s |||- objHdrs B 1 ** objHdrs G 1 ** objHdrs (W |_| free) 0 ** setHp G s **
      4 + x |-> x2 ** x |-> x1 ** 
	objHp ((B |_| G) |_| W) (set_remove x (set_remove (rget s a0) (G |_| W))) **
	rget s a0 |-> v ** 4 + rget s a0 |-> x0 **
          objHp (B |_| G) B ** ifreeList freePtr free ** A).
  split; trivial.
  split; trivial.
  split; trivial.
  destruct (HGshade _ _ _ _ _ _ _ Htmp) as [Hp' _]; clear HGshade Htmp.
  hypReplace Hp (Hp' HvinW); clear Hp'.
  sconjLift8 Hp.
  hypReplace Hp (HGret _ _ Hp); clear HGret.
  sconjLift7 Hp.
  assert (Hgetxl' := contGetEqL Hp).
  sconjLift7 Hp.
  assert (Hgetxr' := contGetEqL Hp).
  unfold hget.
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.
  clear HGshade HGret Hpb Hphase Hphase' Hpsa0 Hgetvhd.

  sconjLift2 Hp.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  sconjLift2 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift2 Hp.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  rewrite <- (not_In_removeEq _ _ HvnotinF).
  rewrite hpredTrueEq.
  setoid_rewrite <- (set_removeUnionDistrib v W free).
  apply Hp.
  sconjLift2 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImpl Hp); clear h Hp.
  unfold objHp; intros h Hp; apply (fasImpl Hp).
  unfold okObjAt, okFieldAt.
  intros.
  linLogElim H.
  linLogIntro.
  2:apply H.
  destruct H5; [left; trivial | right; elimSetIn].
  destruct H6; [left; trivial | right; elimSetIn].
  intros h Hp.
  apply (sconjImplL (sconjAssocR (sconjAssocR Hp))); clear h Hp; intros h Hp.
  rewrite hpredTrueEq.
  setoid_replace (set_add v G |_| set_remove v W) with (G |_| W).
  split; intro; elimSetIn.
  setoid_replace ((B |_| set_add v G) |_| set_remove v W) with ((B |_| G) |_| W).
  split; intro; elimSetIn.
  setoid_rewrite <- (addRemoveEqv HinGW).
  apply objHpIntro.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  unfold okObjAt, okFieldAt.
  linLogIntro.
  2 : apply Hp.
  apply Ha1.
  assumption.
  (* MARK_PHASE /\ v in W /\ ~ v in W) => false *)
  intro  HvnotinW.
  assert (HvinBG := set_InUnion_notinR_inL _ _ _ Hvin HvnotinW).
  destruct (set_union_elim HvinBG) as [HvinB | HvinG]; clear HvinBG.
  hypReplace Hp (sconjAssocL (fasElimL HvinB Hp)).
  assert (Hgetvhd' := contGetEqL Hp).
  rewrite Hgetvhd' in Hgetvhd; inversion Hgetvhd.
  sconjLift2 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HvinG Hp)).
  assert (Hgetvhd' := contGetEqL Hp).
  rewrite Hgetvhd' in Hgetvhd; inversion Hgetvhd.
  rewrite Hphase in Hphase'; elim Hphase'; reflexivity.

  destruct Hpre as [Hphase' _].
  rewrite Hphase in Hphase'; elim Hphase'; reflexivity.
  clear Hphase Hvnotatom HGshade HGret Hpsa0 Hgetvhd.
  intros.
  compare r ra; intro.
  subst r.
  rewrite Hra.
  regRewrite rcst2; reflexivity.
  compare r a0; intro.
  subst r.
  rewrite Ha0''.
  regRewrite v0; reflexivity.
  regEqSolver; trivial.
  regEqSolver; trivial.
  stSimpl; reflexivity.
  
  (* MARK_PHASE /\ ~ atomP v /\ ~ set_In v W *)
  split.
  split.
  exists ohpSt; exists ohpEnd; exists B; exists G; 
    exists W; exists free.
  split; trivial.
  split; trivial.
  split; trivial.
  unfold setHp; stSimpl.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  linLogIntro.
  
  unfold shadeWhiteGuar, shadeWhitePreP, BrMPHp, setHp.
  stSimpl.
  intros s' [HGshade Hr].
  exists v; regRewrite v0; stSimpl; simpl.
  assert (Htmp : rget s rphase = MARK_PHASE /\
    set_eqv ((B |_| G) |_| (W |_| free)) (rangeMSObjs ohpSt ohpEnd) /\
    set_In v ((B |_| G) |_| W) /\
    BrMPHp s freePtr B G W free top).
  split; trivial.
  split; trivial.
  split; trivial.
  hypReplace Hp (HGshade _ _ _ _ _ _ _ Htmp); clear HGshade Htmp Hr.
  destruct Hp as [_ Hp].
  destruct (Hp HvnotinW) as [Hs Hrswp].
  rewrite Hs; trivial.
  assert (Hgetvhd : Heap.get (heapOf s) (v - 4) = Some 1).
  assert (HvinBG := set_InUnion_notinR_inL _ _ _ Hvin HvnotinW).
  destruct (set_union_elim HvinBG) as [HvinB | HvinG]; clear HvinBG.
  hypReplace Hp (sconjAssocL (fasElimL HvinB Hp)).
  apply (contGetEqL Hp).
  sconjLift2 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HvinG Hp)).
  apply (contGetEqL Hp).

  clear ohpSt ohpEnd freePtr B G W free Ha1 Hin Hp Hvin Hset_eq HvnotinW.
  unfold shadeWhiteGuar, shadeWhitePreP, BrMPHp, setHp.
  stSimpl.
  intros s'' [s' [[HGshade Hr] [HGret [Hr' [Ha0'' Hra]]]]].
  generalize HGret.
  rewrite Ha0''.
  regRewrite v0.
  regRewrite a1.
  stSimpl.
  simpl; clear HGret; intro HGret.
  split.
  unfold BarrierPreP.
  intros ohpSt ohpEnd freePtr B G W free A [Hpre | Hpre].
  destruct Hpre as [_ [Ha1 [Hset_eq [Hin Hp]]]].
  exists v.
  assert (Hgeta0 := contGetEqL Hpsa0).
  assert (Hv : okFieldVal ((B |_| G) |_| W) v).
  destruct (set_In_dec (rget s a0) B) as [HinB | HnotinB].
  sconjLift5 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinB Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgeta0' := contGetEqL Hp).
  rewrite Hgeta0' in Hgeta0; inversion Hgeta0.
  subst x.
  destruct H3; [left; trivial | right; elimSetIn].
  assert (Hin' : set_In (rget s a0) (B |_| (G |_| W))).
  elimSetIn.
  assert (HinGW := set_InUnion_notinL_inR _ _ _ Hin' HnotinB).
  sconjLift6 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinGW Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgeta0' := contGetEqL Hp).
  rewrite Hgeta0' in Hgeta0; inversion Hgeta0.
  subst x.
  assumption.
  unfold hget; split; trivial.
  split; trivial.
  split; intro Hphase'.
  split; intro.
  elimtype False; apply Hvnotatom; trivial.
  destruct Hv as [Ha | Hvin].
  elimtype False; apply Hvnotatom; trivial.
  clear H; split.
  intro HvinW.
  sconjLift3 Hp.
  hypReplace Hp (sconjAssocL (fasElimL (set_union_intro1 _ _ free HvinW) Hp)).
  assert (Hgetvhd' := contGetEqL Hp).
  rewrite Hgetvhd' in Hgetvhd; inversion Hgetvhd.
  intro HvnotinW.
  assert (Htmp : rget s rphase = MARK_PHASE /\
    set_eqv ((B |_| G) |_| (W |_| free)) (rangeMSObjs ohpSt ohpEnd) /\
    set_In v ((B |_| G) |_| W) /\ BrMPHp s freePtr B G W free A).
  split; trivial.
  split; trivial.
  split; trivial.
  assert (Hp' := HGshade _ _ _ _ _ _ _ Htmp); clear HGshade Htmp.
  destruct Hp' as [_ Hp'].
  destruct (Hp' HvnotinW) as [Hs Hrswp]; clear Hp' Hphase'.
  unfold BrMPRemoveA0Hp.
  rewrite Ha0''; regRewrite v0; stSimpl; simpl.
  unfold setHp.
  regRewrite rbot.
  regRewrite rtop.
  regRewrite rbuff.
  regRewrite rbot.
  rewrite Hrswp.
  regRewrite rbuff.
  stSimpl.
  destruct (set_In_dec (rget s a0) B) as [HinB | HnotinB].  
  (* MARK_PHASE /\ ~ atomP v /\ ~ v in W /\ rget s a0 |-> v /\ rget s a0 in B *)
  sconjLift5 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinB Hp)).
  assert (Hpb := Hp).
  assert (HnotinG : ~ set_In (rget s a0) G).
  intro HinG.
  sconjLift7 Hp.
  hypReplace Hp (sconjAssocL (fasElimL (set_union_intro1 _ _ W HinG) Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  sconjLift4 Hp.
  assert (rget s a0 <> rget s a0).
  eapply sconjNeq.
  linLogIntro.
  apply Hp.
  apply H; reflexivity.
  assert (HnotinW : ~ set_In (rget s a0) W).
  intro HinW.
  sconjLift7 Hp.
  hypReplace Hp (sconjAssocL (fasElimL (set_union_intro2 _ G _ HinW) Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  sconjLift4 Hp.
  assert (rget s a0 <> rget s a0).
  eapply sconjNeq.
  linLogIntro.
  apply Hp.
  apply H; reflexivity.
  rewrite <- Hs in Hp.
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  hypReplace Hp (HGret _ _ Hp).
  split.
  split.
  apply (contGetEqL Hp).
  clear H3 x x0 H6 Hp Hr Hgetvhd Hpsa0 Hr' Ha0'' Hra; rename Hpb into Hp.
  intros x [Hxin [HxnotinF Hex]].
  assert (Hxin' : set_In x (B |_| (G |_| W))).
  apply set_InUnion_notinR_inL with free; trivial.
  assert (set_In x ((B |_| G) |_| W |_| free)).
  setoid_rewrite Hset_eq; assumption.
  elimSetIn.
  destruct (set_union_elim Hxin') as [HxinB | HxinGW]; clear Hxin'.
  sconjLift2 Hp.
  assert (HxinB' : set_In x (set_remove (rget s a0) B)).
  apply set_remove_intro; trivial.
  hypReplace Hp (sconjAssocL (fasElimL HxinB' Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgetxl := contGetEqL Hp).
  sconjLift2 Hp.
  assert (Hgetxr := contGetEqL Hp).

  rewrite Hs in HGret.
  sconjLift4 Hp.
  hypReplace Hp (HGret _ _ Hp); clear HGret Hs.
  sconjLift3 Hp.
  assert (Hgetxl' := contGetEqL Hp).
  sconjLift3 Hp.
  assert (Hgetxr' := contGetEqL Hp).
  unfold hget.
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.

  sconjLift7 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HxinGW Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgetxl := contGetEqL Hp).
  sconjLift2 Hp.
  assert (Hgetxr := contGetEqL Hp).
  rewrite Hs in HGret.
  sconjLift4 Hp.
  hypReplace Hp (HGret _ _ Hp); clear Hs HGret.
  sconjLift3 Hp.
  assert (Hgetxl' := contGetEqL Hp).
  sconjLift3 Hp.
  assert (Hgetxr' := contGetEqL Hp).
  unfold hget.
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.
  clear Hpb Hphase Hr Hr' Ha0'' Hra HGret Hpsa0 Hgetvhd.

  sconjLift4 Hp.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift3 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift3 Hp.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  rewrite hpredTrueEq.
  setoid_rewrite (set_removeUnionDistrib (rget s a0) G W).
  rewrite (not_In_removeEq (rget s a0) G HnotinG).
  rewrite (not_In_removeEq (rget s a0) W HnotinW).
  apply Hp.
  apply (sconjImplL (sconjAssocR Hp)); clear h Hp; intros h Hp.
  unfold objHp, okObjAt, okFieldAt, fas, fas'.
  simpl.
  linLogIntro.
  2 : apply Hp.
  apply Ha1.
  destruct H6; [left; trivial | right; elimSetIn].
  assert (Hin' : set_In (rget s a0) (B |_| (G |_| W))).
  elimSetIn.
  assert (HinGW := set_InUnion_notinL_inR _ _ _ Hin' HnotinB).
  (* MARK_PHASE /\ ~ atomP v /\ ~ v in W /\ rget s a0 |-> v /\ rget s a0 in G |_| W *)
  sconjLift6 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinGW Hp)).
  assert (Hpb := Hp).
  rewrite <- Hs in Hp.
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  hypReplace Hp (HGret _ _ Hp).
  split.
  split.
  apply (contGetEqL Hp).
  clear x H3 Hpsa0 Hgetvhd H6 x0 Hp Hr Hr' Ha0'' Hra; rename Hpb into Hp.
  intros x [Hxin [HxnotinF Hex]].
  assert (Hxin' : set_In x (B |_| (G |_| W))).
  apply set_InUnion_notinR_inL with free; trivial.
  assert (set_In x ((B |_| G) |_| W |_| free)).
  setoid_rewrite Hset_eq; assumption.
  elimSetIn.
  destruct (set_union_elim Hxin') as [HxinB | HxinGW]; clear Hxin'.
  sconjLift7 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HxinB Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgetxl := contGetEqL Hp).
  sconjLift2 Hp.
  assert (Hgetxr := contGetEqL Hp).

  rewrite Hs in HGret.
  sconjLift4 Hp.
  hypReplace Hp (HGret _ _ Hp); clear HGret Hs.
  sconjLift3 Hp.
  assert (Hgetxl' := contGetEqL Hp).
  sconjLift3 Hp.
  assert (Hgetxr' := contGetEqL Hp).
  unfold hget.
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.

  sconjLift2 Hp.
  assert (HxinGW' : set_In x (set_remove (rget s a0) (G |_| W))).
  apply set_remove_intro; trivial.
  hypReplace Hp (sconjAssocL (fasElimL HxinGW' Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgetxl := contGetEqL Hp).
  sconjLift2 Hp.
  assert (Hgetxr := contGetEqL Hp).
  rewrite Hs in HGret.
  sconjLift4 Hp.
  hypReplace Hp (HGret _ _ Hp); clear Hs HGret.
  sconjLift3 Hp.
  assert (Hgetxl' := contGetEqL Hp).
  sconjLift3 Hp.
  assert (Hgetxr' := contGetEqL Hp).
  unfold hget.
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.
  clear Hpb Hphase Hr Hr' Ha0'' Hra HGret Hpsa0 Hgetvhd.

  sconjLift4 Hp.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  rewrite hpredTrueEq.
  rewrite (not_In_removeEq (rget s a0) B HnotinB).
  apply Hp.
  sconjLift3 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplL (sconjAssocR Hp)); clear h Hp; intros h Hp.
  unfold objHp, okObjAt, okFieldAt, fas, fas'.
  simpl.
  linLogIntro.
  2 : apply Hp.
  apply Ha1.
  assumption.

  rewrite Hphase in Hphase'; elim Hphase'; reflexivity.
  destruct Hpre as [Hphase' _].
  rewrite Hphase in Hphase'; elim Hphase'; reflexivity.
  clear HGshade HGret Hphase Hpsa0 Hgetvhd Hvnotatom.
  intros.
  compare r a0; intro.
  subst r.
  rewrite Ha0''.
  regRewrite v0; reflexivity.
  compare r ra; intro.
  subst r.
  rewrite Hra.
  regRewrite rcst2; reflexivity.
  regEqSolver; trivial.
  regEqSolver; trivial.
  stSimpl; reflexivity.
  (* MARK_PHASE /\ atomP v *)
  assert (Hatom : atomP v).
  trivial.
  rewrite (atomIsOdd Hatom).
  doBtestSimpl.
  clear Bt.
  split.
  exists v; stSimpl; simpl; trivial.
  clear HFV Hvodd Hp Hin Hset_eq Ha1 ohpSt ohpEnd freePtr B G W free.
  unfold yussaBarrierRetGuar.
  stSimpl; simpl.
  intros s' [HGret [Hr [Ha0 Hra]]].
  rewrite Ha0 in HGret.

  split.
  unfold BarrierPreP.
  intros ohpSt ohpEnd freePtr B G W free A [Hpre | Hpre].
  destruct Hpre as [_ [Ha1 [Hset_eq [Hin Hp]]]].
  exists v.
  assert (Hgeta0 := contGetEqL Hpsa0).
  assert (Hv : okFieldVal ((B |_| G) |_| W) v).
  destruct (set_In_dec (rget s a0) B) as [HinB | HnotinB].
  sconjLift5 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinB Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgeta0' := contGetEqL Hp).
  rewrite Hgeta0' in Hgeta0; inversion Hgeta0.
  subst x.
  destruct H3; [left; trivial | right; elimSetIn].
  assert (Hin' : set_In (rget s a0) (B |_| (G |_| W))).
  elimSetIn.
  assert (HinGW := set_InUnion_notinL_inR _ _ _ Hin' HnotinB).
  sconjLift6 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinGW Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgeta0' := contGetEqL Hp).
  rewrite Hgeta0' in Hgeta0; inversion Hgeta0.
  subst x.
  assumption.
  unfold hget; split; trivial.
  split; trivial.
  split; intro Hphase'.
  split; intro.
  unfold BrMPRemoveA0Hp, setHp.
  rewrite Ha0.
  regRewrite rbot.
  regRewrite rtop.
  regRewrite rbuff.
  stSimpl.
  destruct (set_In_dec (rget s a0) B) as [HinB | HnotinB].
  (* MARK_PHASE /\ atomP v /\ rget s a0 |-> v /\ rget s a0 in B *)
  assert (Hp' := Hp).
  sconjLift5 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinB Hp)).
  assert (HnotinG : ~ set_In (rget s a0) G).
  intro HinG.
  sconjLift7 Hp.
  hypReplace Hp (sconjAssocL (fasElimL (set_union_intro1 _ _ W HinG) Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  sconjLift4 Hp.
  assert (rget s a0 <> rget s a0).
  eapply sconjNeq.
  linLogIntro.
  apply Hp.
  apply H0; reflexivity.
  assert (HnotinW : ~ set_In (rget s a0) W).
  intro HinW.
  sconjLift7 Hp.
  hypReplace Hp (sconjAssocL (fasElimL (set_union_intro2 _ G _ HinW) Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  sconjLift4 Hp.
  assert (rget s a0 <> rget s a0).
  eapply sconjNeq.
  linLogIntro.
  apply Hp.
  apply H0; reflexivity.
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  hypReplace Hp (HGret _ _ Hp).
  split.
  split.
  apply (contGetEqL Hp).

  clear Hpsa0 H4 H7 x x0 Hp Hr H Ha0 Hra Hphase' Hphase.
  intros x [Hxin [HxnotinF Hex]].
  assert (Hxin' : set_In x (B |_| (G |_| W))).
  apply set_InUnion_notinR_inL with free; trivial.
  assert (set_In x ((B |_| G) |_| W |_| free)).
  setoid_rewrite Hset_eq; assumption.
  elimSetIn.
  sconjLift5 Hp'.
  hypReplace Hp' (sconjAssocL (fasElimL HinB Hp')).
  destruct (set_union_elim Hxin') as [HxinB | HxinGW]; clear Hxin'.
  assert (HxinB' : set_In x (set_remove (rget s a0) B)).
  apply set_remove_intro; auto.
  sconjLift2 Hp'.
  hypReplace Hp' (sconjAssocL (fasElimL HxinB' Hp')).
  unfold okObjAt, okFieldAt in Hp'.
  linLogElim Hp'.
  assert (Hgetxl := contGetEqL Hp').
  sconjLift2 Hp'.
  assert (Hgetxr := contGetEqL Hp').
  sconjLift4 Hp'.
  hypReplace Hp' (HGret _ _ Hp'); clear HGret.
  sconjLift3 Hp'.
  assert (Hgetxl' := contGetEqL Hp').
  sconjLift3 Hp'.
  assert (Hgetxr' := contGetEqL Hp').
  unfold hget.
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.
  sconjLift7 Hp'.
  hypReplace Hp' (sconjAssocL (fasElimL HxinGW Hp')).
  unfold okObjAt, okFieldAt in Hp'.
  linLogElim Hp'.
  assert (Hgetxl := contGetEqL Hp').
  sconjLift2 Hp'.
  assert (Hgetxr := contGetEqL Hp').
  sconjLift4 Hp'.
  hypReplace Hp' (HGret _ _ Hp'); clear HGret.
  sconjLift3 Hp'.
  assert (Hgetxl' := contGetEqL Hp').
  sconjLift3 Hp'.
  assert (Hgetxr' := contGetEqL Hp').
  unfold hget.
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.

  sconjLift4 Hp.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift3 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift3 Hp.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  rewrite hpredTrueEq.
  setoid_rewrite (set_removeUnionDistrib (rget s a0) G W).
  rewrite (not_In_removeEq (rget s a0) G HnotinG).
  rewrite (not_In_removeEq (rget s a0) W HnotinW).
  apply Hp.
  apply (sconjImplL (sconjAssocR Hp)); clear h Hp; intros h Hp.
  unfold objHp, okObjAt, okFieldAt, fas, fas'.
  simpl.
  linLogIntro.
  2 : apply Hp.
  apply Ha1.
  destruct H7; [left; trivial | right; elimSetIn].
  assert (Hin' : set_In (rget s a0) (B |_| (G |_| W))).
  elimSetIn.
  assert (HinGW := set_InUnion_notinL_inR _ _ _ Hin' HnotinB).
  (* MARK_PHASE /\ atomP v /\ rget s a0 |-> v /\ rget s a0 in G |_| W *)
  sconjLift6 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinGW Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hpb := Hp).
  hypReplace Hp (HGret _ _ Hp).
  split.
  split.
  apply (contGetEqL Hp).

  clear Hpsa0 Hp Hr H Ha0 Hra Hphase' Hphase; rename Hpb into Hp.
  intros m [Hmin [HmnotinF Hem]].
  assert (Hmin' : set_In m (B |_| (G |_| W))).
  apply set_InUnion_notinR_inL with free; trivial.
  assert (set_In m ((B |_| G) |_| W |_| free)).
  setoid_rewrite Hset_eq; assumption.
  elimSetIn.
  destruct (set_union_elim Hmin') as [HminB | HminGW]; clear Hmin'.
  sconjLift8 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HminB Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgetxl := contGetEqL Hp).
  sconjLift2 Hp.
  assert (Hgetxr := contGetEqL Hp).
  sconjLift4 Hp.
  hypReplace Hp (HGret _ _ Hp); clear HGret.
  sconjLift3 Hp.
  assert (Hgetxl' := contGetEqL Hp).
  sconjLift3 Hp.
  assert (Hgetxr' := contGetEqL Hp).
  unfold hget.
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.
  sconjLift3 Hp.
  assert (HminGW' : set_In m (set_remove (rget s a0) (G |_| W))).
  apply set_remove_intro; auto.
  hypReplace Hp (sconjAssocL (fasElimL HminGW' Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgetxl := contGetEqL Hp).
  sconjLift2 Hp.
  assert (Hgetxr := contGetEqL Hp).
  sconjLift4 Hp.
  hypReplace Hp (HGret _ _ Hp); clear HGret.
  sconjLift3 Hp.
  assert (Hgetxl' := contGetEqL Hp).
  sconjLift3 Hp.
  assert (Hgetxr' := contGetEqL Hp).
  unfold hget.
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.
  clear Hpb Hr HGret.

  sconjLift4 Hp.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  rewrite hpredTrueEq.
  rewrite (not_In_removeEq (rget s a0) B HnotinB).
  apply Hp.
  sconjLift3 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplL (sconjAssocR Hp)); clear h Hp; intros h Hp.
  unfold objHp, okObjAt, okFieldAt, fas, fas'.
  simpl.
  linLogIntro.
  2 : apply Hp.
  apply Ha1.
  assumption.

  elimtype False; apply H; trivial.
  rewrite Hphase in Hphase'; elim Hphase'; reflexivity.
  destruct Hpre as [Hphase' _].
  rewrite Hphase in Hphase'; elim Hphase'; reflexivity.
  clear HGret Hpsa0 Hphase.
  intros.
  compare r a0; intro.
  subst; trivial.
  compare r ra; intro.
  subst; trivial.
  regEqSolver; trivial.
  stSimpl; trivial.
  
  inversion Bt.

  rewrite addiuStep.
  rewrite addiuStep.
  rewrite addiuStep; stSimpl.
  doBtestSimpl.
  2 : rewrite Bt in Hphase; elim Hphase; reflexivity.
  (* ~ MARK_PHASE *)
  destruct Hp as [sweepAt Hp].
  split.
  unfold yussaBarrierRetPre.
  stSimpl; simpl.
  sconjLift4 Hp.
  hypReplace Hp (sconjAssocR (fasElimL Hin Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  exists x.
  apply (sconjImplR Hp).
  intros; linLogIntro.

  clear ohpSt ohpEnd freePtr B G W free Ha1 Hset_eq Hin sweepAt Hp Bt.
  unfold yussaBarrierRetGuar.
  stSimpl; simpl.
  intros s' [HGret [Hr [Ha0 Hra]]].
  rewrite Ha0 in HGret.
  split.
  intros.
  destruct H as [HinM | HnotM].
  destruct HinM as [Hphase' _].
  rewrite Hphase' in Hphase; elim Hphase; reflexivity.
  destruct HnotM as [_ [Ha1 [Hset_eqv [Hin [sweepAt Hp]]]]].
  sconjLift4 Hp.
  hypReplace Hp (sconjAssocR (fasElimL Hin Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  exists x.
  split.
  unfold hget.
  apply (contGetEqL Hp).
  split.
  destruct H3; [left; trivial | right; elimSetIn].
  split; intro Hphase'.
  rewrite Hphase' in Hphase; elim Hphase; reflexivity.
  destruct (even_odd_dec x) as [Hxeven | Hxodd].
  assert (Hxnotatom : ~ atomP x).
  unfold atomP; intro.
  apply not_even_and_odd with x; trivial.
  destruct H3 as [Hatom | HinB].
  elimtype False; apply Hxnotatom; trivial.
  assert (Hpb := Hp).
  hypReplace Hp (HGret _ _ Hp).
  split.
  split.
  unfold hget.
  apply (contGetEqL Hp).

  clear Hp Hphase' Hxeven Ha0 Hra Hr; rename Hpb into Hp.
  intros m [Hmin [HmnotinF Hem]].
  assert (Hmin' : set_In m (B |_| W)).
  apply set_InUnion_notinR_inL with free; trivial.
  assert (set_In m (B |_| W |_| free)).
  setoid_rewrite Hset_eqv; assumption.
  elimSetIn.
  destruct (set_union_elim Hmin') as [HminB | HminW]; clear Hmin'.
  sconjLift3 Hp.
  assert (HminB' : set_In m (set_remove (rget s a0) B)).
  apply set_remove_intro; trivial.
  hypReplace Hp (sconjAssocL (fasElimL HminB' Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgetxl := contGetEqL Hp).
  sconjLift2 Hp.
  assert (Hgetxr := contGetEqL Hp).
  sconjLift4 Hp.
  hypReplace Hp (HGret _ _ Hp); clear HGret.
  sconjLift3 Hp.
  assert (Hgetxl' := contGetEqL Hp).
  sconjLift3 Hp.
  assert (Hgetxr' := contGetEqL Hp).
  unfold hget.
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.
  sconjLift7 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HminW Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgetxl := contGetEqL Hp).
  sconjLift2 Hp.
  assert (Hgetxr := contGetEqL Hp).
  sconjLift4 Hp.
  hypReplace Hp (HGret _ _ Hp); clear HGret.
  sconjLift3 Hp.
  assert (Hgetxl' := contGetEqL Hp).
  sconjLift3 Hp.
  assert (Hgetxr' := contGetEqL Hp).
  unfold hget.
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.
  clear HGret Hpb.

  exists sweepAt.
  sconjLift4 Hp.
  unfold BrNotMPHp.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift4 Hp.
  unfold setHp.
  regRewrite rbot.
  regRewrite rtop.
  regRewrite rbuff.
  stSimpl.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplL (sconjAssocR (sconjAssocR Hp))).
  clear h Hp; intros h Hp.
  rewrite hpredTrueEq.
  setoid_rewrite <- (addRemoveEqv Hin).
  apply objHpIntro.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  unfold okObjAt, okFieldAt.
  linLogIntro.
  2 : apply Hp.
  setoid_rewrite (addRemoveEqv Hin).
  apply Ha1.
  setoid_rewrite (addRemoveEqv Hin).
  assumption.
  rewrite hpredTrueEq.
  setoid_rewrite (addRemoveEqv Hin).
  apply Hp.
  (* atomP x *)
  assert (Hpb := Hp).
  hypReplace Hp (HGret _ _ Hp).
  split.
  split.
  unfold hget.
  apply (contGetEqL Hp).

  clear Hphase' Ha0 Hra Hp; rename Hpb into Hp.
  intros m [Hmin [HmnotinF Hem]].
  assert (Hmin' : set_In m (B |_| W)).
  apply set_InUnion_notinR_inL with free; trivial.
  assert (set_In m (B |_| W |_| free)).
  setoid_rewrite Hset_eqv; assumption.
  elimSetIn.
  destruct (set_union_elim Hmin') as [HminB | HminW]; clear Hmin'.
  sconjLift3 Hp.
  assert (HminB' : set_In m (set_remove (rget s a0) B)).
  apply set_remove_intro; trivial.
  hypReplace Hp (sconjAssocL (fasElimL HminB' Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgetxl := contGetEqL Hp).
  sconjLift2 Hp.
  assert (Hgetxr := contGetEqL Hp).
  sconjLift4 Hp.
  hypReplace Hp (HGret _ _ Hp); clear HGret.
  sconjLift3 Hp.
  assert (Hgetxl' := contGetEqL Hp).
  sconjLift3 Hp.
  assert (Hgetxr' := contGetEqL Hp).
  unfold hget.
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.
  sconjLift7 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HminW Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgetxl := contGetEqL Hp).
  sconjLift2 Hp.
  assert (Hgetxr := contGetEqL Hp).
  sconjLift4 Hp.
  hypReplace Hp (HGret _ _ Hp); clear HGret.
  sconjLift3 Hp.
  assert (Hgetxl' := contGetEqL Hp).
  sconjLift3 Hp.
  assert (Hgetxr' := contGetEqL Hp).
  unfold hget.
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.
  clear HGret Hpb Hphase' Ha0 Hra.

  exists sweepAt.
  unfold BrNotMPHp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift4 Hp.
  unfold setHp.
  regRewrite rbot.
  regRewrite rtop.
  regRewrite rbuff.
  stSimpl.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplL (sconjAssocR (sconjAssocR Hp))).
  clear h Hp; intros h Hp.
  rewrite hpredTrueEq.
  setoid_rewrite <- (addRemoveEqv Hin).
  apply objHpIntro.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  unfold okObjAt, okFieldAt.
  linLogIntro.
  2 : apply Hp.
  setoid_rewrite (addRemoveEqv Hin).
  apply Ha1.
  setoid_rewrite (addRemoveEqv Hin).
  assumption.
  rewrite hpredTrueEq.
  setoid_rewrite (addRemoveEqv Hin).
  apply Hp.
  intros.
  compare r ra; intro.
  subst r; trivial.
  compare r a0; intro.
  subst r; trivial.
  regEqSolver; trivial.
  stSimpl; reflexivity.
Qed.

(* Djikstra's write barrier *)
(* a modified Djikstra barrier
void djikstraStore (obj_t** field, obj_t* newVal) {
  *field = newVal;
  if (phase <> MARK_PHASE && isAtom(newVal)) return;
  shadeWhite(newVal);
}
*)
Notation djikstraBarrierPreP := BarrierPreP.
Notation djikstraBarrierPre := BarrierPre.

Definition djikstraBarrierGuar s s' :=  
  (forall ohpSt ohpEnd freePtr B G W free A,
    djikstraBarrierPreP s ohpSt ohpEnd freePtr B G W free A ->
    (rget s rphase = MARK_PHASE ->
      (atomP (rget s a1) -> 
	BarrierHpInv s s' ohpSt ohpEnd free /\
	BrMPHp s' freePtr B G W free A) /\
      (~ atomP (rget s a1) ->
	(set_In (rget s a1) W -> 
	  BarrierHpInv s s' ohpSt ohpEnd free /\
	  BrMPHp s' freePtr B (set_add (rget s a1) G) (set_remove (rget s a1) W) free A) /\
	(~ set_In (rget s a1) W ->
	  BarrierHpInv s s' ohpSt ohpEnd free /\
	  BrMPHp s' freePtr B G W free A))) /\
    (rget s rphase <> MARK_PHASE ->
      BarrierHpInv s s' ohpSt ohpEnd free /\
      exists sweepAt,
	BrNotMPHp s' ohpSt ohpEnd freePtr sweepAt B W free A)) /\
  (forall r, r <> rtemp -> r <> rtop -> r <> rcst2 -> r <> rcst3 -> r <> a0 ->
    rget s' r = rget s r).

Definition djikstraBarrierImpl :=
  addiu rcst2 ra 0;
  sw a1 0 a0;
  addiu r31 zero MARK_PHASE;
  bne rphase r31 RETURN2;;
  andi1 r31 a1;
  bne r31 zero RETURN2;;
  addiu a0 a1 0;
  jal SHADEWHITE RETURN2.

Definition djikstraBarrierP f :=
  if EqNat.beq_nat SHADEWHITE f then
    SomeT (bspeci shadeWhitePre shadeWhiteGuar)
    else if EqNat.beq_nat RETURN2 f then
      SomeT (bspeci return2Pre return2Guar)
      else
	NoneT _.

Ltac Psimpldjikstra := 
  unfold djikstraBarrierP; 
  rewrite beq_true; 
  repeat (rewrite beq_false; [idtac | discriminate]); 
  reflexivity.

Theorem djikstraBarrierImplOk :
  iseqOkS djikstraBarrierP djikstraBarrierPre djikstraBarrierGuar djikstraBarrierImpl.
Proof.
  unfold djikstraBarrierImpl.
  iseqWeakSolver Psimpldjikstra.

  unfold shadeWhiteGuar; intros s s' _ [_ Hr].
  regEqSolver.

  intros s [ohpSt [ohpEnd [freePtr [B [G [W [free Hpre]]]]]]].
  destruct Hpre as [[Hphase [Ha1 [Hset_eq [Hin Hp]]]] | [Hphase [Ha1 [Hset_eq [Hin Hp]]]]].
  
  (* MARK_PHASE *)
  rewrite addiuStep.
  assert (Hpsa0 : exists v, s |||- rget s a0 |-> v ** top).
  destruct (set_union_elim Hin) as [HinBG | HinW].
  destruct (set_union_elim HinBG) as [HinB | HinG]; clear HinBG.
  sconjLift5 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinB Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  exists x.
  apply (sconjImplR Hp).
  intros; linLogIntro.
  sconjLift6 Hp.
  hypReplace Hp (sconjAssocL (fasElimL (set_union_intro1 _ _ W HinG) Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  exists x.
  apply (sconjImplR Hp).
  intros; linLogIntro.
  sconjLift6 Hp.
  hypReplace Hp (sconjAssocL (fasElimL (set_union_intro2 _ G _ HinW) Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  exists x.
  apply (sconjImplR Hp).
  intros; linLogIntro.
  destruct Hpsa0 as [v Hpsa0].
  rewrite (swStep Hpsa0); trivial.
  rewrite addiuStep; stSimpl.
  rewrite Hphase.
  doBtestSimpl.
  elim Bt; reflexivity.
  clear Bt.
  rewrite andi1Step; stSimpl.
  destruct (even_odd_dec (rget s a1)) as [Ha1even | Ha1odd].
  (* MARK_PHASE /\ ~ atomP (rget s a1) *)
  assert (Ha1notatom : ~ atomP (rget s a1)).
  unfold atomP; intro.
  apply not_even_and_odd with (rget s a1); trivial.
  rewrite (notAtomNotOdd Ha1notatom).
  doBtestSimpl.
  elim Bt; reflexivity.
  clear Bt.
  rewrite addiuStep.
  destruct Ha1 as [Ha1atom | Ha1in].
  elimtype False; apply not_even_and_odd with (rget s a1); trivial.
  split.
  split.
  exists ohpSt; exists ohpEnd; exists B;
    exists G; exists W; exists free.
  split; trivial.
  split; trivial.
  split; trivial.
  unfold setHp; stSimpl.
  destruct (set_In_dec (rget s a0) B) as [Ha0inB | Ha0notinB].
  sconjLift5 Hp.
  hypReplace Hp (sconjAssocL (fasElimL Ha0inB Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  hypReplace Hp (contUpdateSt Hp (rget s a1)).
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  linLogIntro.
  assert (Hin' : set_In (rget s a0) (B |_| (G |_| W))).
  elimSetIn.
  assert (Ha0inGW := set_InUnion_notinL_inR _ _ _ Hin' Ha0notinB). 
  sconjLift6 Hp.
  hypReplace Hp (sconjAssocL (fasElimL Ha0inGW Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  hypReplace Hp (contUpdateSt Hp (rget s a1)).
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  linLogIntro.
  unfold return2Pre; auto.

  clear ohpSt ohpEnd freePtr B G W free Ha1in Hset_eq Hin Hp v
    Hpsa0 Ha1even.
  unfold shadeWhiteGuar, shadeWhitePreP, return2Guar, setHp; stSimpl.
  intros s'' [s' [[HGshade Hr] HGret]].
  subst s''.
  split; intros.
  destruct H as  [[_ [Ha1 [Hset_eq [Hin Hp]]]] | [Hphase' [Ha1 [Hset_eq [Hin Hp]]]]].
  split; intro Hphase'.
  destruct Ha1 as [Ha1 | Ha1].
  elimtype False; apply Ha1notatom; trivial.
  split; intro.
  elimtype False; apply Ha1notatom; trivial.
  split.
  intro Ha1inW.
  unfold setHp; stSimpl.
  (* MARK_PHASE /\ (rget s a1) in W *)
  assert (Ha1notinF : ~ set_In (rget s a1) free).
  clear HGshade Hr.
  intro Ha1inF.
  sconjLift7 Hp.
  assert (Hf := inFreeListVal _ _ _ _ _ Hp Ha1inF).
  linLogElim Hf.
  sconjLift7 Hp.
  hypReplace Hp (sconjAssocL (fasElimL (set_union_intro2 _ G _ Ha1inW) Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  sconjLift2 Hp.
  sconjLift2 Hf.
  assert (Hget := contGetEqL Hp).
  assert (Hget' := contGetEqL Hf).
  rewrite Hget' in Hget; inversion Hget.
  subst x1; clear Hget' Hget Hf Hp.
  destruct H7 as [Ha | Hi].
  inversion_clear Ha.
  inversion_clear H0.
  inversion_clear H1.
  assert (set_In 2 (rangeMSObjs ohpSt ohpEnd)).
  elimSetIn.
  assert (~ set_In INFREELIST (rangeMSObjs ohpSt ohpEnd)).
  eapply inFreeListnotinMSobjs.
  reflexivity.
  apply H1; trivial.

  destruct (set_In_dec (rget s a0) B) as [Ha0inB | Ha0notinB].
  (* MARK_PHASE /\ (rget s a1) in W /\ (rget s a0) in B *)
  sconjLift5 Hp.
  hypReplace Hp (sconjAssocL (fasElimL Ha0inB Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hpb := Hp).
  hypReplace Hp (contUpdateSt Hp (rget s a1)).
  sconjLift7 Hp.
  sconjLift7 Hp.
  sconjLift7 Hp.
  sconjLift7 Hp.
  assert (Htmp : rget s rphase = 0 /\
    set_eqv ((B |_| G) |_| W |_| free) (rangeMSObjs ohpSt ohpEnd) /\
    set_In (0 + rget s a1) ((B |_| G) |_| W) /\
    BrHp (uhset s (rget s a0)(rget s a1)) freePtr B G W free 
    (rget s a1) x0 
    (objHp (B |_| G) (set_remove (rget s a0) B))
    (objHp ((B |_| G) |_| W) (G |_| W)) A).
  split; trivial.
  split; trivial.
  split; trivial.
  destruct (HGshade _ _ _ _ _ _ _ Htmp) as [Hp' _]; clear Htmp.
  hypReplace Hp (Hp' Ha1inW); clear Hp'.
  unfold BrMPHp.
  Opaque set_add.
  generalize Hp; clear Hp H Hphase'; stSimpl; simpl.
  intro Hp.
  split.
  split.
  sconjLift5 Hp.
  unfold hget.
  apply (contGetEqL Hp).

  clear Hp Hr; rename Hpb into Hp.
  intros v [Hvin [HvnotinF Hev]].
  assert (Hvin' : set_In v (B |_| (G |_| W))).
  apply set_InUnion_notinR_inL with free; trivial.
  assert (set_In v ((B |_| G) |_| W |_| free)).
  setoid_rewrite Hset_eq; assumption.
  elimSetIn.
  sconjLift7 Hp.
  sconjLift7 Hp.
  sconjLift7 Hp.
  sconjLift7 Hp.
  destruct (set_union_elim Hvin') as [HvinB | HvinGW]; clear Hvin'.
  sconjLift7 Hp.
  assert (HvinB' : set_In v (set_remove (rget s a0) B)).
  apply set_remove_intro; trivial.
  hypReplace Hp (sconjAssocL (fasElimL HvinB' Hp)).
  linLogElim Hp.
  assert (Hgetxl := contGetEqL Hp).
  sconjLift2 Hp.
  assert (Hgetxr := contGetEqL Hp).
  sconjLift8 Hp.
  hypReplace Hp (contUpdateSt Hp (rget s a1)).
  sconjLift8 Hp.
  sconjLift8 Hp.
  sconjLift8 Hp.
  sconjLift8 Hp.
  assert (Htmp : rget s rphase = MARK_PHASE /\
    set_eqv ((B |_| G) |_| W |_| free) (rangeMSObjs ohpSt ohpEnd) /\
    set_In (rget s a1) ((B |_| G) |_| W) /\
    Heap.uset (heapOf s) (rget s a0) (rget s a1) ||- objHdrs B 1 **
      objHdrs G 1 ** objHdrs (W |_| free) 0 ** setHp G s **
      rget s a0 |-> rget s a1 ** 4 + v |-> x2 ** v |-> x1 **
	objHp (B |_| G) (set_remove v (set_remove (rget s a0) B)) **
	4 + rget s a0 |-> x0 **
          objHp ((B |_| G) |_| W) (G |_| W) ** ifreeList freePtr free ** A).
  split; trivial.
  split; trivial.
  split; trivial.
  simpl in HGshade.
  destruct (HGshade _ _ _ _ _ _ _ Htmp) as [Hp' _]; clear HGshade Htmp.
  hypReplace Hp (Hp' Ha1inW); clear Hp'.
  sconjLift7 Hp.
  assert (Hgetxl' := contGetEqL Hp).
  sconjLift7 Hp.
  assert (Hgetxr' := contGetEqL Hp).
  unfold hget; stSimpl.
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.
  sconjLift8 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HvinGW Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgetxl := contGetEqL Hp).
  sconjLift2 Hp.
  assert (Hgetxr := contGetEqL Hp).
  sconjLift8 Hp.
  hypReplace Hp (contUpdateSt Hp (rget s a1)).
  sconjLift8 Hp.
  sconjLift8 Hp.
  sconjLift8 Hp.
  sconjLift8 Hp.
  assert (Htmp : rget s rphase = MARK_PHASE /\
    set_eqv ((B |_| G) |_| W |_| free) (rangeMSObjs ohpSt ohpEnd) /\
    set_In (rget s a1) ((B |_| G) |_| W) /\
    Heap.uset (heapOf s) (rget s a0) (rget s a1) ||- objHdrs B 1 **
      objHdrs G 1 ** objHdrs (W |_| free) 0 ** setHp G s **
      rget s a0 |-> rget s a1 ** 4 + v |-> x2 ** v |-> x1 **
        objHp ((B |_| G) |_| W) (set_remove v (G |_| W)) **
	4 + rget s a0 |-> x0 **
	  objHp (B |_| G) (set_remove (rget s a0) B) ** ifreeList freePtr free ** A).
  split; trivial.
  split; trivial.
  split; trivial.
  simpl in HGshade.
  destruct (HGshade _ _ _ _ _ _ _ Htmp) as [Hp' _]; clear HGshade Htmp.
  hypReplace Hp (Hp' Ha1inW); clear Hp'.
  sconjLift7 Hp.
  assert (Hgetxl' := contGetEqL Hp).
  sconjLift7 Hp.
  assert (Hgetxr' := contGetEqL Hp).
  unfold hget; stSimpl.
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.
  clear Hpb HGshade Hr.

  apply (sconjImplR Hp); clear Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  rewrite <- (not_In_removeEq (rget s a1) free Ha1notinF).
  rewrite hpredTrueEq.
  setoid_rewrite <- (set_removeUnionDistrib (rget s a1) W free).
  apply Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImpl (sconjAssocR (sconjAssocR Hp))); clear h Hp; intros h Hp.
  rewrite hpredTrueEq.
  setoid_rewrite <- (addRemoveEqv Ha0inB).
  apply objHpIntro.
  apply (sconjImpl Hp); clear h Hp.
  intros h Hp.
  unfold okObjAt, okFieldAt.
  linLogIntro.
  2: apply Hp.
  right; elimSetIn.
  setoid_rewrite (addRemoveEqv Ha0inB).
  destruct H7; [left; trivial | right; elimSetIn].
  unfold objHp.
  intros h Hp; apply (fasImpl Hp).
  unfold okObjAt, okFieldAt.
  intros.
  linLogElim H.
  linLogIntro.
  2 : apply H.
  destruct H5; [left; trivial | right; elimSetIn].
  destruct H6; [left; trivial | right; elimSetIn].
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  rewrite hpredTrueEq.
  setoid_replace (set_add (rget s a1) G |_| set_remove (rget s a1) W) with (G |_| W).
  split; intro; elimSetIn.
  setoid_replace ((B |_| set_add (rget s a1) G) |_| set_remove (rget s a1) W)
    with ((B |_| G) |_| W).
  split; intro; elimSetIn.
  apply Hp.
  (* MARK_PHASE /\ (rget s a1) in W /\ (rget s a0) in G |_| W *)
  assert (Hin' : set_In (rget s a0) (B |_| (G |_| W))).
  elimSetIn.
  assert (Ha0inGW := set_InUnion_notinL_inR _ _ _ Hin' Ha0notinB).
  sconjLift6 Hp.
  hypReplace Hp (sconjAssocL (fasElimL Ha0inGW Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hpb := Hp).
  hypReplace Hp (contUpdateSt Hp (rget s a1)).
  sconjLift7 Hp.
  sconjLift7 Hp.
  sconjLift7 Hp.
  sconjLift7 Hp.
  assert (Htmp : rget s rphase = 0 /\
    set_eqv ((B |_| G) |_| W |_| free) (rangeMSObjs ohpSt ohpEnd) /\
    set_In (0 + rget s a1) ((B |_| G) |_| W) /\
    BrHp (uhset s (rget s a0)(rget s a1)) freePtr B G W free 
    (rget s a1) x0 
    (objHp ((B |_| G) |_| W) (set_remove (rget s a0) (G |_| W)))
    (objHp (B |_| G) B) A).
  split; trivial.
  split; trivial.
  split; trivial.
  destruct (HGshade _ _ _ _ _ _ _ Htmp) as [Hp' _]; clear Htmp.
  hypReplace Hp (Hp' Ha1inW); clear Hp'.
  unfold BrMPHp.
  generalize Hp; clear Hp H Hphase'; stSimpl; simpl.
  intro Hp.
  split.
  split.
  sconjLift5 Hp.
  unfold hget.
  apply (contGetEqL Hp).

  clear Hp Hr; rename Hpb into Hp.
  intros v [Hvin [HvnotinF Hev]].
  assert (Hvin' : set_In v (B |_| (G |_| W))).
  apply set_InUnion_notinR_inL with free; trivial.
  assert (set_In v ((B |_| G) |_| W |_| free)).
  setoid_rewrite Hset_eq; assumption.
  elimSetIn.
  sconjLift7 Hp.
  sconjLift7 Hp.
  sconjLift7 Hp.
  sconjLift7 Hp.
  destruct (set_union_elim Hvin') as [HvinB | HvinGW]; clear Hvin'.
  sconjLift8 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HvinB Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgetxl := contGetEqL Hp).
  sconjLift2 Hp.
  assert (Hgetxr := contGetEqL Hp).
  sconjLift8 Hp.
  hypReplace Hp (contUpdateSt Hp (rget s a1)).
  sconjLift8 Hp.
  sconjLift8 Hp.
  sconjLift8 Hp.
  sconjLift8 Hp.
  assert (Htmp : rget s rphase = MARK_PHASE /\
    set_eqv ((B |_| G) |_| W |_| free) (rangeMSObjs ohpSt ohpEnd) /\
    set_In (rget s a1) ((B |_| G) |_| W) /\
    Heap.uset (heapOf s) (rget s a0) (rget s a1) ||- objHdrs B 1 **
      objHdrs G 1 ** objHdrs (W |_| free) 0 ** setHp G s **
      rget s a0 |-> rget s a1 ** 4 + v |-> x2 ** v |-> x1 **
	objHp (B |_| G) (set_remove v B) **
	4 + rget s a0 |-> x0 **
          objHp ((B |_| G) |_| W) (set_remove (rget s a0) (G |_| W))
	  ** ifreeList freePtr free ** A).
  split; trivial.
  split; trivial.
  split; trivial.
  simpl in HGshade.
  destruct (HGshade _ _ _ _ _ _ _ Htmp) as [Hp' _]; clear HGshade Htmp.
  hypReplace Hp (Hp' Ha1inW); clear Hp'.
  sconjLift7 Hp.
  assert (Hgetxl' := contGetEqL Hp).
  sconjLift7 Hp.
  assert (Hgetxr' := contGetEqL Hp).
  unfold hget; stSimpl.
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.
  sconjLift7 Hp.
  assert (HvinGW' : set_In v (set_remove (rget s a0) (G |_| W))).
  apply set_remove_intro; trivial.
  hypReplace Hp (sconjAssocL (fasElimL HvinGW' Hp)).
  linLogElim Hp.
  assert (Hgetxl := contGetEqL Hp).
  sconjLift2 Hp.
  assert (Hgetxr := contGetEqL Hp).
  sconjLift8 Hp.
  hypReplace Hp (contUpdateSt Hp (rget s a1)).
  sconjLift8 Hp.
  sconjLift8 Hp.
  sconjLift8 Hp.
  sconjLift8 Hp.
  assert (Htmp : rget s rphase = MARK_PHASE /\
    set_eqv ((B |_| G) |_| W |_| free) (rangeMSObjs ohpSt ohpEnd) /\
    set_In (rget s a1) ((B |_| G) |_| W) /\
    Heap.uset (heapOf s) (rget s a0) (rget s a1) ||- objHdrs B 1 **
      objHdrs G 1 ** objHdrs (W |_| free) 0 ** setHp G s **
      rget s a0 |-> rget s a1 ** 4 + v |-> x2 ** v |-> x1 **
        objHp ((B |_| G) |_| W) (set_remove v (set_remove (rget s a0) (G |_| W))) **
	4 + rget s a0 |-> x0 ** objHp (B |_| G) B ** ifreeList freePtr free ** A).
  split; trivial.
  split; trivial.
  split; trivial.
  simpl in HGshade.
  destruct (HGshade _ _ _ _ _ _ _ Htmp) as [Hp' _]; clear HGshade Htmp.
  hypReplace Hp (Hp' Ha1inW); clear Hp'.
  sconjLift7 Hp.
  assert (Hgetxl' := contGetEqL Hp).
  sconjLift7 Hp.
  assert (Hgetxr' := contGetEqL Hp).
  unfold hget; stSimpl.
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.
  clear Hpb HGshade Hr.

  apply (sconjImplR Hp); clear Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  rewrite <- (not_In_removeEq (rget s a1) free Ha1notinF).
  rewrite hpredTrueEq.
  setoid_rewrite <- (set_removeUnionDistrib (rget s a1) W free).
  apply Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImpl Hp); clear h Hp.
  unfold objHp.
  intros h Hp; apply (fasImpl Hp).
  unfold okObjAt, okFieldAt.
  intros.
  linLogElim H.
  linLogIntro.
  2 : apply H.
  destruct H5; [left; trivial | right; elimSetIn].
  destruct H6; [left; trivial | right; elimSetIn].
  intros h Hp.
  apply (sconjImplL (sconjAssocR (sconjAssocR Hp))); clear h Hp; intros h Hp.
  rewrite hpredTrueEq.
  setoid_replace (set_add (rget s a1) G |_| set_remove (rget s a1) W) with (G |_| W).
  split; intro; elimSetIn.
  setoid_replace ((B |_| set_add (rget s a1) G) |_| set_remove (rget s a1) W)
    with ((B |_| G) |_| W).
  split; intro; elimSetIn.
  setoid_rewrite <- (addRemoveEqv Ha0inGW).
  apply objHpIntro.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  unfold okObjAt, okFieldAt.
  linLogIntro.
  2: apply Hp.
  right; elimSetIn.
  assumption.
  (* MARK_PHASE /\ ~ atomP (rget s a1) /\ ~ (rget s a1) in W *)
  intro Ha1notinW.
  destruct (set_In_dec (rget s a0) B) as [Ha0inB | Ha0notinB].
  (* MARK_PHASE /\ ~ atomP (rget s a1) /\ ~ (rget s a1) in W /\ (rget s a0) in B *)
  sconjLift5 Hp.
  hypReplace Hp (sconjAssocL (fasElimL Ha0inB Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hpb := Hp).
  hypReplace Hp (contUpdateSt Hp (rget s a1)).
  sconjLift7 Hp.
  sconjLift7 Hp.
  sconjLift7 Hp.
  sconjLift7 Hp.
  assert (Htmp : rget s rphase = 0 /\
    set_eqv ((B |_| G) |_| W |_| free) (rangeMSObjs ohpSt ohpEnd) /\
    set_In (0 + rget s a1) ((B |_| G) |_| W) /\
    BrHp (uhset s (rget s a0)(rget s a1)) freePtr B G W free 
    (rget s a1) x0 
    (objHp (B |_| G) (set_remove (rget s a0) B))
    (objHp ((B |_| G) |_| W) (G |_| W)) A).
  split; trivial.
  split; trivial.
  split; trivial.
  destruct (HGshade _ _ _ _ _ _ _ Htmp) as [_ Hp']; clear HGshade Htmp.
  destruct (Hp' Ha1notinW) as [Hs Hrswp]; clear H Hphase' Hp'.
  simpl in Hs.
  rewrite <- Hs in Hp.
  split.
  split.
  sconjLift5 Hp.
  unfold hget.
  apply (contGetEqL Hp).

  clear Hp Hr; rename Hpb into Hp.
  intros v [Hvin [HvnotinF Hev]].
  assert (Hvin' : set_In v (B |_| (G |_| W))).
  apply set_InUnion_notinR_inL with free; trivial.
  assert (set_In v ((B |_| G) |_| W |_| free)).
  setoid_rewrite Hset_eq; assumption.
  elimSetIn.
  destruct (set_union_elim Hvin') as [HvinB | HvinGW]; clear Hvin'.
  sconjLift3 Hp.
  assert (HvinB' : set_In v (set_remove (rget s a0) B)).
  apply set_remove_intro; trivial.
  hypReplace Hp (sconjAssocL (fasElimL HvinB' Hp)).
  linLogElim Hp.
  assert (Hgetxl := contGetEqL Hp).
  sconjLift2 Hp.
  assert (Hgetxr := contGetEqL Hp).
  sconjLift4 Hp.
  hypReplace Hp (contUpdateSt Hp (rget s a1)).
  rewrite <- Hs in Hp.
  sconjLift3 Hp.
  assert (Hgetxl' := contGetEqL Hp).
  sconjLift3 Hp.
  assert (Hgetxr' := contGetEqL Hp).
  unfold hget; stSimpl.
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.
  sconjLift8 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HvinGW Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgetxl := contGetEqL Hp).
  sconjLift2 Hp.
  assert (Hgetxr := contGetEqL Hp).
  sconjLift4 Hp.
  hypReplace Hp (contUpdateSt Hp (rget s a1)).
  rewrite <- Hs in Hp.
  sconjLift3 Hp.
  assert (Hgetxl' := contGetEqL Hp).
  sconjLift3 Hp.
  assert (Hgetxr' := contGetEqL Hp).
  unfold hget; stSimpl.
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.
  clear Hpb.

  unfold BrMPHp, setHp; stSimpl.
  assert (Ha1inBG := set_InUnion_notinR_inL _ _ _ Ha1 Ha1notinW).
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  regRewrite rbot.
  rewrite Hrswp.
  regRewrite rbuff.
  stSimpl.
  apply (sconjImplR Hp); clear h Hp Hr; intros h Hp.
  apply (sconjImplL (sconjAssocR (sconjAssocR Hp))); clear h Hp; intros h Hp.
  rewrite hpredTrueEq.
  setoid_rewrite <- (addRemoveEqv Ha0inB).
  apply objHpIntro.
  apply (sconjImpl Hp); clear h Hp.
  intros h Hp.
  unfold okObjAt, okFieldAt.
  linLogIntro.
  2: apply Hp.
  right; elimSetIn.
  setoid_rewrite (addRemoveEqv Ha0inB).
  assumption.
  intros h Hp.
  rewrite hpredTrueEq.
  setoid_rewrite (addRemoveEqv Ha0inB).
  apply Hp.
  (* MARK_PHASE /\ ~atomP (rget s a1) /\ ~ (rget s a1) in W /\ (rget s a0) in G |_| W *)
  assert (Hin' : set_In (rget s a0) (B |_| (G |_| W))).
  elimSetIn.
  assert (Ha0inGW := set_InUnion_notinL_inR _ _ _ Hin' Ha0notinB).
  sconjLift6 Hp.
  hypReplace Hp (sconjAssocL (fasElimL Ha0inGW Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hpb := Hp).
  hypReplace Hp (contUpdateSt Hp (rget s a1)).
  sconjLift7 Hp.
  sconjLift7 Hp.
  sconjLift7 Hp.
  sconjLift7 Hp.
  assert (Htmp : rget s rphase = 0 /\
    set_eqv ((B |_| G) |_| W |_| free) (rangeMSObjs ohpSt ohpEnd) /\
    set_In (0 + rget s a1) ((B |_| G) |_| W) /\
    BrHp (uhset s (rget s a0)(rget s a1)) freePtr B G W free 
    (rget s a1) x0 
    (objHp ((B |_| G) |_| W) (set_remove (rget s a0) (G |_| W)))
    (objHp (B |_| G) B) A).
  split; trivial.
  split; trivial.
  split; trivial.
  destruct (HGshade _ _ _ _ _ _ _ Htmp) as [_ Hp']; clear Htmp H Hphase'.
  destruct (Hp' Ha1notinW) as [Hs Hrswp]; clear Hp' HGshade Ha1notinW.
  simpl in Hs.
  rewrite <- Hs in Hp.
  split.
  split.
  sconjLift5 Hp.
  unfold hget.
  apply (contGetEqL Hp).

  clear Hp Hr; rename Hpb into Hp.
  intros v [Hvin [HvnotinF Hev]].
  assert (Hvin' : set_In v (B |_| (G |_| W))).
  apply set_InUnion_notinR_inL with free; trivial.
  assert (set_In v ((B |_| G) |_| W |_| free)).
  setoid_rewrite Hset_eq; assumption.
  elimSetIn.
  destruct (set_union_elim Hvin') as [HvinB | HvinGW]; clear Hvin'.
  sconjLift8 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HvinB Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgetxl := contGetEqL Hp).
  sconjLift2 Hp.
  assert (Hgetxr := contGetEqL Hp).
  sconjLift4 Hp.
  hypReplace Hp (contUpdateSt Hp (rget s a1)).
  rewrite <- Hs in Hp.
  sconjLift3 Hp.
  assert (Hgetxl' := contGetEqL Hp).
  sconjLift3 Hp.
  assert (Hgetxr' := contGetEqL Hp).
  unfold hget; stSimpl.
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.
  sconjLift3 Hp.
  assert (HvinGW' : set_In v (set_remove (rget s a0) (G |_| W))).
  apply set_remove_intro; trivial.
  hypReplace Hp (sconjAssocL (fasElimL HvinGW' Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgetxl := contGetEqL Hp).
  sconjLift2 Hp.
  assert (Hgetxr := contGetEqL Hp).
  sconjLift4 Hp.
  hypReplace Hp (contUpdateSt Hp (rget s a1)).
  rewrite <- Hs in Hp.
  sconjLift3 Hp.
  assert (Hgetxl' := contGetEqL Hp).
  sconjLift3 Hp.
  assert (Hgetxr' := contGetEqL Hp).
  unfold hget; stSimpl.
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.
  clear Hpb.

  unfold BrMPHp, setHp; stSimpl.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  regRewrite rbot.
  rewrite Hrswp.
  regRewrite rbuff.
  stSimpl.
  apply (sconjImplR Hp); clear h Hp Hr; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplL (sconjAssocR (sconjAssocR Hp))); clear h Hp; intros h Hp.
  rewrite hpredTrueEq.
  setoid_rewrite <- (addRemoveEqv Ha0inGW).
  apply objHpIntro.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  unfold okObjAt, okFieldAt.
  linLogIntro.
  2: apply Hp.
  right; elimSetIn.
  trivial.
  rewrite Hphase in Hphase'; elim Hphase'; reflexivity.
  rewrite Hphase in Hphase'; elim Hphase'; reflexivity.
  clear HGshade Ha1notatom Hphase.
  compare r ra; intro.
  subst r; regRewrite rcst2; reflexivity.
  transitivity (rget s' r).
  apply rgetRsetNeq; auto.
  regEqSolver; trivial.
  stSimpl; reflexivity.
  (* MARK_PHASE /\ atomP (rget s a1) *)
  assert (Hatom : atomP (rget s a1)).
  apply Ha1odd.
  rewrite (atomIsOdd Hatom).
  doBtestSimpl.
  clear Bt.
  unfold return2Guar, return2Pre.
  split; trivial.
  stSimpl; simpl.

  clear Ha1 Hset_eq Hin Hp v Hpsa0 Ha1odd ohpSt ohpEnd B G W free freePtr.
  intros.
  subst s'.
  unfold djikstraBarrierGuar, djikstraBarrierPreP; stSimpl.
  split; intros.
  destruct H as  [[_ [Ha1 [Hset_eq [Hin Hp]]]] | [Hphase' [Ha1 [Hset_eq [Hin Hp]]]]].
  split; intros Hphase'.
  split; intros.
  unfold BrMPHp, BarrierHpInv, hget, setHp; stSimpl.
  destruct (set_In_dec (rget s a0) B) as [Ha0inB | Ha0notinB].
  (* MARK_PHASE /\ atomP (rget s a1) /\ (rget s a0) in B *)
  sconjLift5 Hp.
  hypReplace Hp (sconjAssocL (fasElimL Ha0inB Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hpb := Hp).
  hypReplace Hp (contUpdateSt Hp (rget s a1)).
  split.
  split.
  apply (contGetEqL Hp).

  clear Hp; rename Hpb into Hp.
  intros v [Hvin [HvnotinF Hev]].
  assert (Hvin' : set_In v (B |_| (G |_| W))).
  apply set_InUnion_notinR_inL with free; trivial.
  assert (set_In v ((B |_| G) |_| W |_| free)).
  setoid_rewrite Hset_eq; assumption.
  elimSetIn.
  destruct (set_union_elim Hvin') as [HvinB | HvinGW]; clear Hvin'.
  sconjLift3 Hp.
  assert (HvinB' : set_In v (set_remove (rget s a0) B)).
  apply set_remove_intro; trivial.
  hypReplace Hp (sconjAssocL (fasElimL HvinB' Hp)).
  linLogElim Hp.
  assert (Hgetxl := contGetEqL Hp).
  sconjLift2 Hp.
  assert (Hgetxr := contGetEqL Hp).
  sconjLift4 Hp.
  hypReplace Hp (contUpdateSt Hp (rget s a1)).
  sconjLift3 Hp.
  assert (Hgetxl' := contGetEqL Hp).
  sconjLift3 Hp.
  assert (Hgetxr' := contGetEqL Hp).
  unfold uhset; stSimpl.
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.
  sconjLift8 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HvinGW Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgetxl := contGetEqL Hp).
  sconjLift2 Hp.
  assert (Hgetxr := contGetEqL Hp).
  sconjLift4 Hp.
  hypReplace Hp (contUpdateSt Hp (rget s a1)).
  sconjLift3 Hp.
  assert (Hgetxl' := contGetEqL Hp).
  sconjLift3 Hp.
  assert (Hgetxr' := contGetEqL Hp).
  unfold uhset; stSimpl.
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.
  clear Hpb.

  sconjLift7 Hp.
  sconjLift7 Hp.
  sconjLift7 Hp.
  sconjLift7 Hp.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplL (sconjAssocR (sconjAssocR Hp))); clear h Hp; intros h Hp.
  rewrite hpredTrueEq.
  setoid_rewrite <- (addRemoveEqv Ha0inB).
  apply objHpIntro.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  unfold okObjAt, okFieldAt.
  linLogIntro.
  2: apply Hp.
  left; trivial.
  setoid_rewrite (addRemoveEqv Ha0inB); assumption.
  rewrite hpredTrueEq.
  setoid_rewrite (addRemoveEqv Ha0inB).
  apply Hp.
  assert (Hin' : set_In (rget s a0) (B |_| (G |_| W))).
  elimSetIn.
  assert (Ha0inGW := set_InUnion_notinL_inR _ _ _ Hin' Ha0notinB).
  (* MARK_PHASE /\ atomP (rget s a1) /\ (rget s a0) in G |_| W *)
  sconjLift6 Hp.
  hypReplace Hp (sconjAssocL (fasElimL Ha0inGW Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hpb := Hp).
  hypReplace Hp (contUpdateSt Hp (rget s a1)).
  split.
  split.
  unfold hget.
  apply (contGetEqL Hp).

  clear Hp; rename Hpb into Hp.
  intros v [Hvin [HvnotinF Hev]].
  assert (Hvin' : set_In v (B |_| (G |_| W))).
  apply set_InUnion_notinR_inL with free; trivial.
  assert (set_In v ((B |_| G) |_| W |_| free)).
  setoid_rewrite Hset_eq; assumption.
  elimSetIn.
  destruct (set_union_elim Hvin') as [HvinB | HvinGW]; clear Hvin'.
  sconjLift8 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HvinB Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgetxl := contGetEqL Hp).
  sconjLift2 Hp.
  assert (Hgetxr := contGetEqL Hp).
  sconjLift4 Hp.
  hypReplace Hp (contUpdateSt Hp (rget s a1)).
  sconjLift3 Hp.
  assert (Hgetxl' := contGetEqL Hp).
  sconjLift3 Hp.
  assert (Hgetxr' := contGetEqL Hp).
  unfold uhset; stSimpl.
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.
  sconjLift3 Hp.
  assert (HvinGW' : set_In v (set_remove (rget s a0) (G |_| W))).
  apply set_remove_intro; trivial.
  hypReplace Hp (sconjAssocL (fasElimL HvinGW' Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgetxl := contGetEqL Hp).
  sconjLift2 Hp.
  assert (Hgetxr := contGetEqL Hp).
  sconjLift4 Hp.
  hypReplace Hp (contUpdateSt Hp (rget s a1)).
  sconjLift3 Hp.
  assert (Hgetxl' := contGetEqL Hp).
  sconjLift3 Hp.
  assert (Hgetxr' := contGetEqL Hp).
  unfold uhset; stSimpl.
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.
  clear Hpb.

  sconjLift7 Hp.
  sconjLift7 Hp.
  sconjLift7 Hp.
  sconjLift7 Hp.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplL (sconjAssocR (sconjAssocR Hp))); clear h Hp; intros h Hp.
  rewrite hpredTrueEq.
  setoid_rewrite <- (addRemoveEqv Ha0inGW).
  apply objHpIntro.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  unfold okObjAt, okFieldAt.
  linLogIntro.
  2: apply Hp.
  left; trivial.
  assumption.
  elimtype False; apply H; trivial.
  rewrite Hphase in Hphase'; elim Hphase'; reflexivity.
  rewrite Hphase in Hphase'; elim Hphase'; reflexivity.
  compare r ra; intro.
  subst r; reflexivity.
  stSimpl; reflexivity.
  inversion Bt.
  (* ~ MARK_PHASE *)
  rewrite addiuStep.
  destruct Hp as [sweepAt Hp].
  sconjLift4 Hp.
  hypReplace Hp (sconjAssocL (fasElimL Hin Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  rewrite (swStep Hp); trivial.
  rewrite addiuStep; stSimpl.
  doBtestSimpl.
  
  clear Bt B G W free freePtr ohpSt ohpEnd Ha1 Hset_eq Hin sweepAt x 
    H3 x0 H6 Hp.
  unfold return2Guar, return2Pre.
  split; trivial.
  stSimpl; simpl.
  intros.
  subst s'.
  unfold djikstraBarrierGuar, djikstraBarrierPreP; stSimpl.
  split; intros.
  destruct H as  [[Hphase' _] | [_ [Ha1 [Hset_eq [Hin [sweepAt Hp]]]]]].
  rewrite Hphase' in Hphase; elim Hphase; reflexivity.
  split; intro Hphase'.
  rewrite Hphase' in Hphase; elim Hphase; reflexivity.
  sconjLift4 Hp.
  hypReplace Hp (sconjAssocL (fasElimL Hin Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hpb := Hp).
  hypReplace Hp (contUpdateSt Hp (rget s a1)).
  unfold BrNotMPHp, setHp; stSimpl.
  split.
  split.
  unfold hget.
  apply (contGetEqL Hp).

  clear Hp; rename Hpb into Hp.
  intros v [Hvin [HvnotinF Hev]].
  assert (Hvin' : set_In v (B |_| W)).
  apply set_InUnion_notinR_inL with free; trivial.
  assert (set_In v (B |_| W |_| free)).
  setoid_rewrite Hset_eq; assumption.
  elimSetIn.
  unfold hget, uhset; stSimpl.
  destruct (set_union_elim Hvin') as [HvinB | HvinW]; clear Hvin'.
  sconjLift3 Hp.
  assert (HvinB' : set_In v (set_remove (rget s a0) B)).
  apply set_remove_intro; trivial.
  hypReplace Hp (sconjAssocL (fasElimL HvinB' Hp)).
  linLogElim Hp.
  assert (Hgetxl := contGetEqL Hp).
  sconjLift2 Hp.
  assert (Hgetxr := contGetEqL Hp).
  sconjLift4 Hp.
  hypReplace Hp (contUpdateSt Hp (rget s a1)).
  sconjLift3 Hp.
  assert (Hgetxl' := contGetEqL Hp).
  sconjLift3 Hp.
  assert (Hgetxr' := contGetEqL Hp).
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.
  sconjLift7 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HvinW Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hgetxl := contGetEqL Hp).
  sconjLift2 Hp.
  assert (Hgetxr := contGetEqL Hp).
  sconjLift4 Hp.
  hypReplace Hp (contUpdateSt Hp (rget s a1)).
  sconjLift3 Hp.
  assert (Hgetxl' := contGetEqL Hp).
  sconjLift3 Hp.
  assert (Hgetxr' := contGetEqL Hp).
  rewrite Hgetxl; rewrite Hgetxl'; rewrite Hgetxr; rewrite Hgetxr'.
  split; reflexivity.
  clear Hpb.

  exists sweepAt.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplL (sconjAssocR (sconjAssocR Hp))); clear h Hp; intros h Hp.
  rewrite hpredTrueEq.
  setoid_rewrite <- (addRemoveEqv Hin).
  apply objHpIntro.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  unfold okObjAt, okFieldAt.
  linLogIntro.
  2: apply Hp.
  setoid_rewrite (addRemoveEqv Hin); assumption.
  setoid_rewrite (addRemoveEqv Hin); assumption.
  rewrite hpredTrueEq.
  setoid_rewrite (addRemoveEqv Hin); assumption.
  compare r ra; intro.
  subst r; trivial.
  stSimpl; reflexivity.
  rewrite Bt in Hphase; elim Hphase; reflexivity.
Qed.

Definition Original_Pointer_Saved s s' B G W :=
  forall v, 
    hget s (rget s a0) = Some v /\
    ~ atomP v /\ set_In v W ->
    exists G',
      set_In v G' /\
      s' |||- objHdrs G' MARKED ** setHp G' s' **
	objHp ((B |_| G) |_| W) (v::empty_set) ** top.

Definition No_Pointer_FromBToW' s B W :=
  forall v v1 v2,
    set_In v B /\ hget s v = Some v1 /\
    hget s (4 + v) = Some v2 /\ 
    (~ atomP v1 \/ ~ atomP v2) ->
    ~ (s |||- objHp W (v::empty_set) ** top).

Definition No_Pointer_FromBToW s s' B G W free :=
  No_Pointer_FromBToW' s B W /\
  exists B', exists G', exists W',
    set_eqv ((B |_| G) |_| W) ((B' |_| G') |_| W') /\
    s' |||- objHdrs B' MARKED ** objHdrs G' MARKED ** 
      objHdrs (W' |_| free) UNMARK ** setHp G' s' ** top /\
    No_Pointer_FromBToW' s' B' W'.    

Lemma djikstra_MP_NoBtoW : forall s freePtr B G W free A,
  BrMPHp s freePtr B G W free A ->
  No_Pointer_FromBToW' s B W.
Proof.
  intros.
  intros v v1 v2 [HvinB [Hv [Hv' Hna]]] Hpv.
  unfold objHp, okObjAt, okFieldAt, fas, fas' in Hpv.
  simpl in Hpv.
  unfold hget in * .
  linLogElim Hpv.
  assert (Hv1 := contGetEqL Hpv).
  rewrite Hv in Hv1; inversion Hv1.
  subst x; clear Hv1.
  sconjLift2 Hpv.
  assert (Hv2 := contGetEqL Hpv).
  simpl in Hv'.
  rewrite Hv' in Hv2; inversion Hv2.
  subst x0; clear Hv2 Hpv.
  rename H into Hp.
  sconjLift5 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HvinB Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hv1 := contGetEqL Hp).
  rewrite Hv in Hv1; inversion Hv1.
  subst x; clear Hv Hv1.
  sconjLift2 Hp.
  assert (Hv2 := contGetEqL Hp).
  simpl in Hv2.
  rewrite Hv' in Hv2; inversion Hv2.
  subst x0; clear Hv' Hv2.

  destruct Hna as [Hnav1 | Hnav2].
  destruct H4 as [Ha | Hv1inW].
  apply Hnav1; trivial.
  destruct H3 as [Ha | Hv1inBG].
  apply Hnav1; trivial.
  assert (Hv1hd : Heap.get (heapOf s) (v1 - 4) = Some 1).
  destruct (set_union_elim Hv1inBG) as [Hv1inB | Hv1inG].
  sconjLift4 Hp.
  hypReplace Hp (sconjAssocL (fasElimL Hv1inB Hp)).
  apply (contGetEqL Hp).
  sconjLift5 Hp.
  hypReplace Hp (sconjAssocL (fasElimL Hv1inG Hp)).
  apply (contGetEqL Hp).
  sconjLift6 Hp.
  hypReplace Hp (sconjAssocL (fasElimL (set_union_intro1 _ _ free Hv1inW) Hp)).
  assert (Hv1hd' := contGetEqL Hp).
  rewrite Hv1hd' in Hv1hd; inversion Hv1hd.

  destruct H7 as [Ha | Hv2inW].
  apply Hnav2; trivial.
  destruct H8 as [Ha | Hv2inBG].
  apply Hnav2; trivial.
  assert (Hv2hd : Heap.get (heapOf s) (v2 - 4) = Some 1).
  destruct (set_union_elim Hv2inBG) as [Hv2inB | Hv2inG].
  sconjLift4 Hp.
  hypReplace Hp (sconjAssocL (fasElimL Hv2inB Hp)).
  apply (contGetEqL Hp).
  sconjLift5 Hp.
  hypReplace Hp (sconjAssocL (fasElimL Hv2inG Hp)).
  apply (contGetEqL Hp).
  sconjLift6 Hp.
  hypReplace Hp (sconjAssocL (fasElimL (set_union_intro1 _ _ free Hv2inW) Hp)).
  assert (Hv2hd' := contGetEqL Hp).
  rewrite Hv2hd' in Hv2hd; inversion Hv2hd.
Qed.

Lemma djikstra_NotMP_NoBtoW : forall s ohpSt ohpEnd freePtr sweepAt B W free A,
  BrNotMPHp s ohpSt ohpEnd freePtr sweepAt B W free A ->
  No_Pointer_FromBToW' s B W.
Proof.
  intros.
  intros v v1 v2 [HvinB [Hv [Hv' Hna]]] Hpv.
  unfold objHp, okObjAt, okFieldAt, fas, fas' in Hpv.
  simpl in Hpv.
  unfold hget in * .
  linLogElim Hpv.
  assert (Hv1 := contGetEqL Hpv).
  rewrite Hv in Hv1; inversion Hv1.
  subst x; clear Hv1.
  sconjLift2 Hpv.
  assert (Hv2 := contGetEqL Hpv).
  simpl in Hv'.
  rewrite Hv' in Hv2; inversion Hv2.
  subst x0; clear Hv2 Hpv.
  rename H into Hp.

  sconjLift4 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HvinB Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  assert (Hv1 := contGetEqL Hp).
  rewrite Hv in Hv1; inversion Hv1.
  subst x; clear Hv Hv1.
  sconjLift2 Hp.
  assert (Hv2 := contGetEqL Hp).
  simpl in Hv2.
  rewrite Hv' in Hv2; inversion Hv2.
  subst x0; clear Hv' Hv2.

  destruct Hna as [Hnav1 | Hnav2].
  destruct H4 as [Ha | Hv1inW].
  apply Hnav1; trivial.
  destruct H3 as [Ha | Hv1inB].
  apply Hnav1; trivial.
  sconjLift5 Hp.
  hypReplace Hp (sconjAssocL (fasElimL (set_union_intro1 _ _ free Hv1inW) Hp)).
  sconjLift6 Hp.
  hypReplace Hp (sconjAssocL (fasElimL Hv1inB Hp)).
  sconjLift3 Hp.
  assert (v1 - 4 <> v1 -4).
  eapply sconjNeq.
  linLogIntro.
  apply Hp.
  apply H; reflexivity.

  destruct H7 as [Ha | Hv2inW].
  apply Hnav2; trivial.
  destruct H8 as [Ha | Hv2inB].
  apply Hnav2; trivial.
  sconjLift5 Hp.
  hypReplace Hp (sconjAssocL (fasElimL (set_union_intro1 _ _ free Hv2inW) Hp)).
  sconjLift6 Hp.
  hypReplace Hp (sconjAssocL (fasElimL Hv2inB Hp)).
  sconjLift3 Hp.
  assert (v2 - 4 <> v2 -4).
  eapply sconjNeq.
  linLogIntro.
  apply Hp.
  apply H; reflexivity.
Qed.

Theorem WFBarrier : forall s s' ohpSt ohpEnd freePtr B G W free A,
  BarrierPreP s ohpSt ohpEnd freePtr B G W free A -> 
  yussaBarrierGuar s s' \/ djikstraBarrierGuar s s' -> 
  BarrierHpInv s s' ohpSt ohpEnd free /\
  (rget s rphase = MARK_PHASE ->
    Original_Pointer_Saved s s' B G W \/ No_Pointer_FromBToW s s' B G W free). 
Proof.
  intros.
  destruct H0 as [HG | HG].
  destruct HG as [HGBr _].
  compare (rget s rphase) 0; intro Hphase.
  hypReplace HGBr (HGBr _ _ _ _ _ _ _ _ H).
  destruct HGBr as [v [Hva0 [HFV [HinMP _]]]].
  hypReplace HinMP (HinMP Hphase).
  destruct (even_odd_dec v) as [Hveven | Hvodd].
  assert (Hvnotatom : ~ atomP v).
  unfold atomP.
  intro.
  apply not_even_and_odd with v; trivial.
  clear Hveven.
  destruct HinMP as [_ Hna].
  hypReplace Hna (Hna Hvnotatom).
  destruct (set_In_dec v W) as [HvinW | HvnotinW].
  destruct Hna as [HMv _].
  hypReplace HMv (HMv HvinW).
  destruct HMv as [HpInv Hp].
  split; trivial.
  intro Hphase'.
  left.
  intros v' [Hva0' HvinW'].
  rewrite Hva0' in Hva0; inversion Hva0.
  subst v'; clear Hva0' Hva0 Hphase' HvinW'.
  assert (HvnotinG : ~ set_In v G).
  intro HvinG.
  destruct H as [Hpre | Hpre].
  destruct Hpre as [_ [_ [_ [_ Hp0]]]].
  sconjLift2 Hp0.
  hypReplace Hp0 (sconjAssocL (fasElimL HvinG Hp0)).
  sconjLift4 Hp0.
  hypReplace Hp0 (sconjAssocL (fasElimL (set_union_intro1 _ _ free HvinW) Hp0)).
  sconjLift3 Hp0.
  assert (v - 4 <> v - 4).
  eapply sconjNeq.
  linLogIntro.
  apply Hp0.
  apply H; reflexivity.
  destruct Hpre as [Hphase' _].
  rewrite Hphase in Hphase'; elim Hphase'; reflexivity.
  sconjLift2 Hp.
  assert (HvinvG : set_In v (set_add v G)).
  elimSetIn.
  exists (set_add v G).
  split; trivial.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  sconjLift3 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift4 Hp.
  assert (Hpp : h ||- objHp ((B |_| G) |_| W) (G |_| W) ** top).
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  rewrite hpredTrueEq.
  setoid_replace (G |_| W) with (set_add v G |_| set_remove v W).
  split; intro; elimSetIn.
  setoid_replace ((B |_| G) |_| W) with
    ((B |_| set_add v G) |_| set_remove v W).
  split; intro; elimSetIn.
  apply Hp.
  linLogIntro.
  clear Hp; rename Hpp into Hp.
  hypReplace Hp (sconjAssocL (fasElimL (set_union_intro2 _ G _ HvinW) Hp)).
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  unfold objHp, fas, fas'; simpl.
  apply unitIR; trivial.
  linLogIntro.
  destruct Hna as [_ HMnv].
  hypReplace HMnv (HMnv HvnotinW).
  destruct HMnv as [HpInv _].
  split; trivial.
  intro Hphase'.
  left.
  intros v' [Hva0' [_ HvinW]].
  rewrite Hva0' in Hva0; inversion Hva0.
  subst v'.
  elimtype False; apply HvnotinW; trivial.
  assert (Hatom : atomP v).
  trivial.
  clear Hvodd.
  destruct HinMP as [Ha _].
  hypReplace Ha (Ha Hatom).
  destruct Ha.
  split; trivial.
  intro Hphase'.
  left.
  intros v' [Hva0' [Hvnotatom _]].
  rewrite Hva0' in Hva0; inversion Hva0.
  subst v'.
  elimtype False; apply Hvnotatom; trivial.
  hypReplace HGBr (HGBr _ _ _ _ _ _ _ _ H).
  destruct HGBr as [v [Hva0 [HFV [_ HNMP]]]].
  hypReplace HNMP (HNMP Hphase).
  destruct HNMP.
  split; trivial.
  intro Hphase'.
  rewrite Hphase' in Hphase; elim Hphase; reflexivity.
  (* Djikstra Barrier *)
  assert (Hoks : No_Pointer_FromBToW' s B W).
  destruct H as [Hpre | Hpre].
  destruct Hpre as [_ [_ [_ [_ Hp]]]].
  eapply djikstra_MP_NoBtoW.
  apply Hp.

  destruct Hpre as [_ [_ [_ [_ [sweepAt Hp]]]]].
  eapply djikstra_NotMP_NoBtoW.
  apply Hp.
  
  destruct HG as [HGBr _].
  compare (rget s rphase) 0; intro Hphase.
  hypReplace HGBr (HGBr _ _ _ _ _ _ _ _ H).
  destruct HGBr as [HinMP _].
  hypReplace HinMP (HinMP Hphase).
  destruct (even_odd_dec (rget s a1)) as [Ha1even | Ha1odd].
  assert (Ha1notatom : ~ atomP (rget s a1)).
  unfold atomP.
  intro.
  apply not_even_and_odd with (rget s a1); trivial.
  clear Ha1even.
  destruct HinMP as [_ Hna].
  hypReplace Hna (Hna Ha1notatom).
  destruct (set_In_dec (rget s a1) W) as [Ha1inW | Ha1notinW].
  destruct Hna as [Hp _].
  hypReplace Hp (Hp Ha1inW).
  destruct Hp as [HpInv Hp].
  split; trivial.
  intro Hphase'.
  right.
  split; trivial.
  exists B; exists (set_add (rget s a1) G); 
    exists (set_remove (rget s a1) W).
  split.
  split; intro; elimSetIn.
  split.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  linLogIntro.
  apply djikstra_MP_NoBtoW with freePtr (set_add (rget s a1) G) free A; trivial.

  destruct Hna as [_ Hp].
  hypReplace Hp (Hp Ha1notinW).
  destruct Hp as [HpInv Hp].
  split; trivial.
  intro Hphase'.
  right.
  split; trivial.
  exists B; exists G; exists W.
  split.
  apply set_eqvRefl.
  split.
  unfold BrMPHp in Hp.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  linLogIntro.

  eapply djikstra_MP_NoBtoW.
  apply Hp.
  assert (Ha1atom : atomP (rget s a1)).
  apply Ha1odd.
  clear Ha1odd.
  destruct HinMP as [Ha _].
  hypReplace Ha (Ha Ha1atom).
  destruct Ha as [HpInv Hp].
  split; trivial.
  intro Hphase'.
  right.
  split; trivial.
  exists B; exists G; exists W.
  split.
  apply set_eqvRefl.
  split.
  unfold BrMPHp in Hp.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  linLogIntro.
  eapply djikstra_MP_NoBtoW.
  apply Hp.
  hypReplace HGBr (HGBr _ _ _ _ _ _ _ _ H).
  destruct HGBr as [_ HNMP].
  hypReplace HNMP (HNMP Hphase).
  destruct HNMP as [HpInv [sweepAt Hp]].
  
  split; trivial.
  intro Hphase'.
  rewrite Hphase' in Hphase; elim Hphase; reflexivity.
Qed.

