Require Export MarkSweepLemma.
Require Import Coq.Arith.Div2.
Require Export BuffArrayStack.
Require Export StackSet.

(** ** auxiliary definitions *)

(** Header Size *)
Notation HDSIZE := 4.

(* Mark bit *)
Notation UNMARK := 0.
Notation MARKED := 1.

(** Divide by 12 *)
Fixpoint div3 (x : nat) {struct x} : nat :=
  match x with
    | O => O
    | S O => O
    | S (S O) => O
    | S (S (S x')) => S (div3 x')
  end. 

Definition div12 x := div2 (div2 (div3 x)).

(** Generate a set of MarkSweep Objects *)
Definition rangeMSObjs (regMin regMax: nat) :=
  set_map (fun k => regMin + HDSIZE + 12 * k)
    (setZeroRange (div12 (regMax - regMin))).

(** MarkSweep headers for objects *)
Definition objHdrs objs n := 
  foralls x in objs, (x - HDSIZE) |-> n.

Definition bobjHdrs objs p m n := 
  foralls x in objs, (x - HDSIZE) |-> 
    if blt_nat x p then m else n.

(** Free list of objects *)
Fixpoint freeList (x : nat) (objs : set) {struct objs} : hpred :=
  match objs with
    | nil => !(x = 0)
    | a::objs' => lexists x', !(x = a) ** !(x <> 0) **
	(a |->_ ** (addrOffset + a) |-> x') ** freeList x' objs'
  end.

(** Callee saved regs without ra *)
Definition calleeSaved' :=
  regInList (s0::s1::s2::s3::s4::s5::s6::s7::nil).

Definition calleeSavedOk' s s' :=
  regPres calleeSaved' s s'.


(** ** Labels and register alias *)

(** Various Returns *)
Notation RETURN := 22.
Notation RETURN0 := 39.
Notation RETURN1 := 23.
Notation RETURN2 := 24.
Notation RETURN3 := 25.

(** markField() *)
Notation MARKFIELD  := 26.

(** mark() *)
Notation MARK := 27.
Notation MARKLOOP := 28.
Notation MARKLOOPBODY := 29.
Notation MARKFIRST := 30.
Notation MARKSECOND := 31.

(** sweep() *)
Notation SWEEP := 32.
Notation SWEEPLOOP := 33.
Notation SWEEPNEXT := 34.
Notation SWEEPADD := 35.
Notation SWEEPFIN := 36.

(** markSweepGC() *)
Notation MARKSWEEPGC := 37.
Notation MARKSWEEPSWEEP := 38.

(** Registers *)
Notation rtemp := ast.
Notation rroot := t0.
Notation rgcInfo := s7.
Notation rsweep := t3.
Notation rhpEnd := t4.
Notation rfree := t2.

(** Control stack *)
Notation rcst1 := t5.
Notation rcst2 := t6.
Notation rcst3 := t7.
Notation rcst0 := t8.

(** ** Codes and Specifications *)

(** *** Various returns *)
Definition returnPre (s : state) := True.
Definition returnGuar := fun s s':state => s = s'.

Definition returnImpl :=
  jr ra.

Definition return0Pre (s : state) := True.
Definition return0Guar s s' := rset s ra (rget s rcst0) = s'.

Definition return0Impl :=
  addiu ra rcst0 0;
  jr ra.

Definition return1Pre (s : state) := True.
Definition return1Guar s s' := rset s ra (rget s rcst1) = s'.

Definition return1Impl :=
  addiu ra rcst1 0;
  jr ra.

Definition return2Pre (s : state) := True.
Definition return2Guar s s' := rset s ra (rget s rcst2) = s'.

Definition return2Impl :=
  addiu ra rcst2 0;
  jr ra.

Definition return3Pre (s : state) := True.
Definition return3Guar s s' := rset s ra (rget s rcst3) = s'.

Definition return3Impl :=
  addiu ra rcst3 0;
  jr ra.

(** *** markField() *)
(* markField pre-condition:
  1. arg0 is not an object pointer, or
  2. arg0 is an object pointer and greater than four,
    the heap is composed with three parts
*)
Definition markFieldPre s :=
  atomP (rget s a0) \/ 
  (rget s a0 >= HDSIZE /\ ~ atomP (rget s a0) /\ 
    exists G, 
      (set_In (rget s a0) G /\
	s |||- objHdrs G MARKED **
	  setHp G s ** top) 
      \/
      (~set_In (rget s a0) G /\
	s |||- rget s a0 - HDSIZE |->_ ** objHdrs G MARKED **
	  setHp G s ** top)).
      
(* markField guarantees:
  1. if arg0 is not an object pointer, or points to a marked object,
    the heap is unchanged.
  2. otherwise the markbit is updated, G' = arg0::G, 
    other parts of the heap is unchanged.
  3. all regs except {ast, a0, rtop, rcst3} are unchanged.
*)
Definition markFieldGuar s s' :=
  ((atomP (rget s a0) \/ 
    s |||- rget s a0 - HDSIZE |-> MARKED ** top) -> 
  heapOf s' = heapOf s /\ rget s rtop = rget s' rtop) /\
  (forall A G, ~ atomP (rget s a0) -> 
    s |||- rget s a0 - HDSIZE |-> UNMARK **
      objHdrs G MARKED **
      buffStackSet (rget s rbot) (rget s rtop) (rget s rbuff) G ** A ->
      s' |||- objHdrs (set_add (rget s a0) G) MARKED **
	buffStackSet (rget s' rbot) (rget s' rtop) 
	(rget s' rbuff) (set_add (rget s a0) G) ** A) /\
  (forall r, r <> ast -> r <> a0 -> r <> rtop -> r <> rcst3 -> 
    rget s' r = rget s r).

Definition markFieldImpl :=
  addiu rcst3 ra 0;
  andi1 r31 a0;
  bne r31 zero RETURN3;;
  addiu r31 zero 4;
  subu a0 a0 r31;
  lw ast 0 a0;
  addiu r31 zero 1;
  beq ast r31 RETURN3;; 
  sw r31 0 a0;
  addiu a0 a0 4;
  jal STACKPUSH RETURN3.

(** *** mark() *)

Definition markPreP 
  s objs ohpSt ohpEnd freePtr stBot stBuff A A':=
  ptrP ohpSt /\ ptrP ohpEnd /\ 
  okFieldVal objs (rget s rroot) /\
  set_eqv objs (rangeMSObjs ohpSt ohpEnd) /\  
  s |||- objHdrs objs UNMARK ** 
    buffStackSet stBot stBot stBuff nil **
    rget s rgcInfo |->, (ohpSt::ohpEnd::freePtr::stBot::stBuff::nil) ** 
      A' ** (A && objHp objs objs).

Definition markPost 
  s B W ohpSt ohpEnd freePtr stBot stBuff A A':=
  ptrP ohpSt /\ ptrP ohpEnd /\ 
(*   set_In (rget s rroot) B /\ *)
  okFieldVal B (rget s rroot) /\
  set_eqv (set_union B W) (rangeMSObjs ohpSt ohpEnd) /\ 
  s |||- objHdrs W UNMARK ** objHdrs B MARKED ** 
    buffStackSet stBot stBot stBuff nil ** 
    rget s rgcInfo |->, (ohpSt::ohpEnd::freePtr::stBot::stBuff::nil) ** 
      A' ** (A && (objHp B B ** objHp (rangeMSObjs ohpSt ohpEnd) W)). 

(* well formed markHeap *)
Definition markPre s := 
  exists objs, exists ohpSt, exists ohpEnd, 
    exists freePtr, exists stBot, exists stBuff,
      markPreP s objs ohpSt ohpEnd freePtr stBot stBuff top top.

(* the objHp is untouched by mark, and W is garbage.
  There may be garbage in B *)
Definition markGuar s s' := 
  (forall objs ohpSt ohpEnd freePtr stBot stBuff A A',
    markPreP s objs ohpSt ohpEnd freePtr stBot stBuff A A' ->
    exists B, exists W, 
      markPost s' B W ohpSt ohpEnd freePtr stBot stBuff A A') /\
  calleeSavedOk s s' /\ rget s rcst1 = rget s' rcst1 /\ rget s rcst0 = rget s' rcst0 /\ 
  rget s' rroot = rget s rroot.

Definition markImpl :=
  addiu rcst2 ra 0;
  lw rbot 12 rgcInfo;
  lw rbuff 16 rgcInfo;
  addiu rtop rbot 0; (* stack_init() *)
  addiu a0 rroot 0;
  jal MARKFIELD MARKLOOP.

Definition markLoopInv s B G W ohpSt ohpEnd freePtr A A' :=
  ptrP ohpSt /\ ptrP ohpEnd /\ 
(*   set_In (rget s rroot) (set_union B G) /\ *)
  okFieldVal (set_union B G) (rget s rroot) /\
  (atomP (rget s rroot) -> G = empty_set) /\
  set_eqv (set_union (set_union B G) W) (rangeMSObjs ohpSt ohpEnd) /\ 
  s |||- objHdrs B MARKED ** 
    objHdrs G MARKED ** objHdrs W UNMARK **
    buffStackSet (rget s rbot) (rget s rtop) (rget s rbuff) G **
    rget s rgcInfo |->, 
      (ohpSt::ohpEnd::freePtr::(rget s rbot)::(rget s rbuff)::nil) ** 
      A' ** (A && (objHp (set_union B G) B ** objHp (rangeMSObjs ohpSt ohpEnd) (set_union G W))).

Definition markLoopPre s :=
  exists B, exists G, exists W,  
    exists ohpSt, exists ohpEnd, exists freePtr,
      markLoopInv s B G W ohpSt ohpEnd freePtr top top.

Definition markLoopGuar s s' := 
  (forall B G W ohpSt ohpEnd freePtr A A',
    markLoopInv s B G W ohpSt ohpEnd freePtr A A' ->
    exists B, exists W,  
      markPost s' B W ohpSt ohpEnd freePtr 
      (rget s rbot) (rget s rbuff) A A') /\
  calleeSavedOk' s s' /\ rget s rcst1 = rget s' rcst1 /\ rget s rcst0 = rget s' rcst0 /\ 
  rget s rcst2 = rget s' ra /\ rget s rroot = rget s' rroot.  

Definition markLoopImpl :=
  jal STACKEMPTY MARKLOOPBODY.

Definition markLoopBodyPre s :=
  exists B, exists G, exists W,
    exists ohpSt, exists ohpEnd, exists freePtr,
    sbool (rget s v0) (G = empty_set) /\
    markLoopInv s B G W ohpSt ohpEnd freePtr top top.

Definition markLoopBodyGuar s s' := 
  (forall B G W ohpSt ohpEnd freePtr A A',
    sbool (rget s v0) (G = empty_set) ->
    markLoopInv s B G W ohpSt ohpEnd freePtr A A' ->
    exists B, exists W, 
      markPost s' B W ohpSt ohpEnd freePtr 
      (rget s rbot) (rget s rbuff) A A') /\
  calleeSavedOk' s s' /\ rget s rcst1 = rget s' rcst1 /\ rget s rcst0 = rget s' rcst0 /\ 
  rget s rcst2 = rget s' ra  /\ rget s rroot = rget s' rroot.  

Definition markLoopBodyImpl :=
  bne v0 zero RETURN2;;
  jal STACKPOP MARKFIRST.

Definition markFirstPreP s B G W ohpSt ohpEnd freePtr A A' :=
  ptrP ohpSt /\ ptrP ohpEnd /\ 
  set_In (rget s rroot) (set_add (rget s v0) (set_union B G)) /\ 
  set_In (rget s v0) (rangeMSObjs ohpSt ohpEnd) /\
  set_eqv (set_add (rget s v0) (set_union (set_union B G) W)) (rangeMSObjs ohpSt ohpEnd) /\ 
  s |||- (rget s v0 - addrOffset |-> MARKED) ** 
    objHdrs B MARKED ** objHdrs G MARKED ** objHdrs W UNMARK **
    buffStackSet (rget s rbot) (rget s rtop) (rget s rbuff) G **
    rget s rgcInfo |->, 
      (ohpSt::ohpEnd::freePtr::(rget s rbot)::(rget s rbuff)::nil) ** 
      A' ** (A && (objHp (set_add (rget s v0) (set_union B G)) B ** 
	objHp (rangeMSObjs ohpSt ohpEnd) (set_union G W) ** 
	objHp (rangeMSObjs ohpSt ohpEnd) ((rget s v0)::empty_set))).

Definition markFirstPre s :=
  exists B, exists G, exists W,
    exists ohpSt, exists ohpEnd, exists freePtr, 
      markFirstPreP s B G W ohpSt ohpEnd freePtr top top.

Definition markFirstGuar s s' := 
  (forall B G W ohpSt ohpEnd freePtr A A',
    markFirstPreP s B G W ohpSt ohpEnd freePtr A A' ->
    exists B, exists W, 
      markPost s' B W ohpSt ohpEnd freePtr 
      (rget s rbot) (rget s rbuff) A A') /\
  calleeSavedOk' s s' /\ rget s rcst1 = rget s' rcst1 /\ rget s rcst0 = rget s' rcst0 /\ 
  rget s rcst2 = rget s' ra /\ rget s rroot = rget s' rroot.  

Definition markFirstImpl :=
  lw a0 0 v0;
  jal MARKFIELD MARKSECOND.

Definition markSecondPreP s B G W ohpSt ohpEnd freePtr A A' :=
  exists lchd,
    ptrP ohpSt /\ ptrP ohpEnd /\ 
    set_In (rget s rroot) (set_add (rget s v0) (set_union B G)) /\ 
    set_In (rget s v0) (rangeMSObjs ohpSt ohpEnd) /\
    set_eqv (set_add (rget s v0) (set_union (set_union B G) W)) (rangeMSObjs ohpSt ohpEnd) /\ 
    hget s (rget s v0) = Some lchd /\
    okFieldVal (set_add (rget s v0) (set_union B G)) lchd /\ 
    s |||- (rget s v0 - addrOffset |-> MARKED) ** 
      objHdrs B MARKED ** objHdrs G MARKED ** objHdrs W UNMARK **
      buffStackSet (rget s rbot) (rget s rtop) (rget s rbuff) G **
      rget s rgcInfo |->, 
	(ohpSt::ohpEnd::freePtr::(rget s rbot)::(rget s rbuff)::nil) ** 
	A' ** (A && (objHp (set_add (rget s v0) (set_union B G)) B  ** 
	  objHp (rangeMSObjs ohpSt ohpEnd) (set_union G W) ** objHp (rangeMSObjs ohpSt ohpEnd) ((rget s v0)::empty_set))).

Definition markSecondPre s :=
  exists B, exists G, exists W,
    exists ohpSt, exists ohpEnd, exists freePtr, 
      markSecondPreP s B G W ohpSt ohpEnd freePtr top top.

Definition markSecondGuar s s' := 
  (forall B G W ohpSt ohpEnd freePtr A A',
    markSecondPreP s B G W ohpSt ohpEnd freePtr A A' ->
    exists B, exists W, 
      markPost s' B W ohpSt ohpEnd freePtr 
      (rget s rbot) (rget s rbuff) A A') /\
  calleeSavedOk' s s' /\ rget s rcst1 = rget s' rcst1 /\ rget s rcst0 = rget s' rcst0 /\ 
  rget s rcst2 = rget s' ra  /\ rget s rroot = rget s' rroot.

Definition markSecondImpl :=
  lw a0 4 v0;
  jal MARKFIELD MARKLOOP.


(** *** sweep() *)

Definition sweepPreP 
  s B W  ohpSt ohpEnd freePtr stBot stBuff A A' :=
  ptrP ohpSt /\ ptrP ohpEnd /\ 
(*   set_In (rget s rroot) B /\  *)
  okFieldVal B (rget s rroot) /\
  set_eqv (set_union B W) (rangeMSObjs ohpSt ohpEnd) /\  
  s |||- objHdrs B MARKED ** objHdrs W UNMARK **
    buffStackSet stBot stBot stBuff nil **
    rget s rgcInfo |->, (ohpSt::ohpEnd::freePtr::stBot::stBuff::nil) ** 
      A' ** objHp (rangeMSObjs ohpSt ohpEnd) W ** (A && objHp B B).

Definition sweepPre s :=
  exists B, exists W, exists ohpSt, 
    exists ohpEnd, exists freePtr, exists stBot, exists stBuff,
      sweepPreP s B W ohpSt ohpEnd freePtr stBot stBuff top top.

Definition sweepPost
  s objs free ohpSt ohpEnd freePtr stBot stBuff A A' :=
  ptrP ohpSt /\ ptrP ohpEnd /\ 
  okFieldVal objs (rget s rroot) /\
  set_eqv (set_union objs free) (rangeMSObjs ohpSt ohpEnd) /\
  s |||- objHdrs (rangeMSObjs ohpSt ohpEnd) UNMARK **
    buffStackSet stBot stBot stBuff nil **
    rget s rgcInfo |->, (ohpSt::ohpEnd::freePtr::stBot::stBuff::nil) ** 
      A' ** freeList freePtr free ** (A && objHp objs objs).

Definition sweepGuar s s' :=
  (forall B W ohpSt ohpEnd freePtr stBot stBuff A A',
    sweepPreP s B W ohpSt ohpEnd freePtr stBot stBuff A A' ->
    exists free', exists freePtr',
      sweepPost s' B free' ohpSt ohpEnd freePtr' stBot stBuff A A') /\
  calleeSavedOk s s' /\ rget s rcst1 = rget s' rcst1 /\ rget s rcst0 = rget s' rcst0 /\ 
  rget s' rroot = rget s rroot.

Definition sweepImpl :=
  lw rsweep 0 rgcInfo;
  addiu rsweep rsweep 4;
  lw rhpEnd 4 rgcInfo;
  addiu rfree zero 0;
  j SWEEPLOOP.

Definition sweepLoopInv
  s B W objs free swept ohpSt freePtr stBot stBuff A A' :=
  ptrP ohpSt /\ ptrP (rget s rhpEnd) /\ 
(*   set_In (rget s rroot) B /\ *)
  okFieldVal B (rget s rroot) /\
  set_diff W swept = W /\
  objs = rangeMSObjs ohpSt (rget s rhpEnd) /\
  swept = rangeMSObjs ohpSt (7 + rget s rsweep) /\
  (exists k, rget s rsweep = ohpSt + addrOffset + 12 * k) /\
  set_subset free swept /\
  set_eqv (set_union (set_union B W) free) objs /\  
  s |||- bobjHdrs B (rget s rsweep) UNMARK MARKED ** 
    objHdrs (set_union W free) UNMARK **
    buffStackSet stBot stBot stBuff nil **
    rget s rgcInfo |->, 
      (ohpSt::(rget s rhpEnd)::freePtr::stBot::stBuff::nil) ** 
      A' ** objHp objs W ** freeList (rget s rfree) free ** 
      (A && objHp B B).

Definition sweepLoopPre s :=
  exists B, exists W, exists objs, exists free, exists swept, 
    exists ohpSt, exists freePtr, exists stBot, exists stBuff,
      sweepLoopInv s B W objs free swept 
      ohpSt freePtr stBot stBuff top top.

Definition sweepLoopGuar s s' :=
  (forall B W objs free swept ohpSt freePtr stBot stBuff A A',
    sweepLoopInv s B W objs free swept ohpSt freePtr stBot stBuff A A' ->
    exists free', exists freePtr',
      sweepPost s' B free' ohpSt (rget s rhpEnd) freePtr' stBot stBuff A A') /\
  calleeSavedOk s s' /\ rget s rcst1 = rget s' rcst1 /\ rget s rcst0 = rget s' rcst0 /\ 
  rget s' rroot = rget s rroot.

Definition sweepLoopImpl :=
  addiu ast rsweep 8;
  sltu ast rhpEnd ast;
  bne ast zero SWEEPFIN;;
  addiu ast zero 4;
  subu ast rsweep ast;
  lw v0 0 ast;
  beq v0 zero SWEEPADD;;
  sw zero 0 ast;
  j SWEEPNEXT.

(* move obj(rsweep) from W to free *)
Definition sweepAddPreP 
  s B W objs free swept ohpSt freePtr stBot stBuff A A' :=
  ptrP ohpSt /\ ptrP (rget s rhpEnd) /\ 
(*   set_In (rget s rroot) B /\  *)
  okFieldVal B (rget s rroot) /\
  set_diff W swept = W /\  
  objs = rangeMSObjs ohpSt (rget s rhpEnd) /\
  swept = rangeMSObjs ohpSt (7 + rget s rsweep) /\
  (exists k, rget s rsweep = ohpSt + addrOffset + 12 * k) /\
  set_subset free swept /\
  set_In (rget s rsweep) W /\
  set_eqv (set_union (set_union B W) free) objs /\  
  s |||- bobjHdrs B (rget s rsweep) UNMARK MARKED ** 
    objHdrs (set_union W free) UNMARK **
    buffStackSet stBot stBot stBuff nil **
    rget s rgcInfo |->, (ohpSt::(rget s rhpEnd)::freePtr::stBot::stBuff::nil) ** 
      A' ** objHp objs W ** freeList (rget s rfree) free ** 
      (A && objHp B B).

Definition sweepAddPre s :=
  exists B, exists W, exists objs, exists free, exists swept, 
    exists ohpSt, exists freePtr, exists stBot, exists stBuff,
      sweepAddPreP s B W objs free swept 
      ohpSt freePtr stBot stBuff top top.

Definition sweepAddGuar s s' :=
  (forall B W objs free swept ohpSt freePtr stBot stBuff A A',
    sweepAddPreP s B W objs free swept ohpSt freePtr stBot stBuff A A' ->
    exists free', exists freePtr',
      sweepPost s' B free' ohpSt (rget s rhpEnd) freePtr' stBot stBuff A A') /\
  calleeSavedOk s s' /\ rget s rcst1 = rget s' rcst1 /\ rget s rcst0 = rget s' rcst0 /\ 
  rget s' rroot = rget s rroot.

Definition sweepAddImpl :=
  sw rfree 4 rsweep;
  addiu rfree rsweep 0;
  j SWEEPNEXT.

Definition sweepNextPreP 
  s B W objs free swept ohpSt freePtr stBot stBuff A A' :=
  ptrP ohpSt /\ ptrP (rget s rhpEnd) /\ 
(*   set_In (rget s rroot) B /\  *)
  okFieldVal B (rget s rroot) /\
  set_diff W swept = W /\
  objs = rangeMSObjs ohpSt (rget s rhpEnd) /\
  swept = rangeMSObjs ohpSt (19 + rget s rsweep) /\
  (exists k, rget s rsweep = ohpSt + addrOffset + 12 * k) /\
  set_subset free swept /\
  set_eqv (set_union (set_union B W) free) objs /\  
  s |||- bobjHdrs B (12 + rget s rsweep) UNMARK MARKED ** 
    objHdrs (set_union W free) UNMARK **
    buffStackSet stBot stBot stBuff nil **
    rget s rgcInfo |->, (ohpSt::(rget s rhpEnd)::freePtr::stBot::stBuff::nil) ** 
      A' ** objHp objs W ** freeList (rget s rfree) free ** 
      (A && objHp B B).

Definition sweepNextPre s :=
  exists B, exists W, exists objs, exists free, exists swept, 
    exists ohpSt, exists freePtr, exists stBot, exists stBuff,
	sweepNextPreP s B W objs free swept 
	ohpSt freePtr stBot stBuff top top.

Definition sweepNextGuar s s' :=
  (forall B W objs free swept ohpSt freePtr stBot stBuff A A',
    sweepNextPreP s B W objs free swept ohpSt freePtr stBot stBuff A A' ->
    exists free', exists freePtr',
      sweepPost s' B free' ohpSt (rget s rhpEnd) freePtr' stBot stBuff A A') /\
  calleeSavedOk s s' /\ rget s rcst1 = rget s' rcst1 /\ rget s rcst0 = rget s' rcst0 /\ 
  rget s' rroot = rget s rroot.

Definition sweepNextImpl :=
  addiu rsweep rsweep 12;
  j SWEEPLOOP.

Definition sweepFinPreP 
  s B objs free ohpSt freePtr stBot stBuff A A':=
  ptrP ohpSt /\ ptrP (rget s rhpEnd) /\ 
(*   set_In (rget s rroot) B /\  *)
  okFieldVal B (rget s rroot) /\
  objs = rangeMSObjs ohpSt (rget s rhpEnd) /\
  set_eqv (set_union B free) objs /\  
  s |||- objHdrs objs UNMARK **
    buffStackSet stBot stBot stBuff nil **
    rget s rgcInfo |->, (ohpSt::(rget s rhpEnd)::freePtr::stBot::stBuff::nil) ** 
      A' ** freeList (rget s rfree) free ** (A && objHp B B).

Definition sweepFinPre s :=
  exists B, exists objs, exists free,
    exists ohpSt, exists freePtr, exists stBot, exists stBuff,
	sweepFinPreP s B objs free ohpSt freePtr stBot stBuff top top.

Definition sweepFinGuar s s' :=
  (forall B objs free ohpSt freePtr stBot stBuff A A',
    sweepFinPreP s B objs free ohpSt freePtr stBot stBuff A A' ->
    exists free', exists freePtr',
      sweepPost s' B free' ohpSt (rget s rhpEnd) freePtr' stBot stBuff A A') /\
  calleeSavedOk s s' /\ rget s rcst1 = rget s' rcst1 /\ rget s rcst0 = rget s' rcst0 /\ 
  rget s' rroot = rget s rroot.

Definition sweepFinImpl :=
  sw rfree 8 rgcInfo;
  jr ra.

(** *** markSweepGC() *)

Notation markSweepGCPre := markPre.

Definition markSweepGCGuar s s' :=
  (forall objs ohpSt ohpEnd freePtr stBot stBuff A,
    markPreP s objs ohpSt ohpEnd freePtr stBot stBuff 
    (objHp objs objs) A ->
    exists objs', exists free', exists freePtr', 
      (forall x, set_In x objs' -> hget s' x = hget s x /\
	hget s' (addrOffset + x) = hget s (addrOffset + x)) /\
      sweepPost s' objs' free' ohpSt ohpEnd freePtr' stBot stBuff 
      (objHp objs' objs') A) /\ rget s' rcst0 = rget s rcst0 /\
  calleeSavedOk s s' /\ rget s' rroot = rget s rroot.

Definition markSweepGCImpl :=
  addiu rcst1 ra 0;
  jal MARK MARKSWEEPSWEEP.

Definition markSweepSweepImpl :=
  jal SWEEP RETURN1.

Notation markSweepSweepPre := sweepPre.

Definition markSweepSweepGuar s s' :=
  (forall B W ohpSt ohpEnd freePtr stBot stBuff A,
    sweepPreP s B W ohpSt ohpEnd freePtr stBot stBuff 
    (objHp B B) A ->
    exists free', exists freePtr', 
      (forall x, set_In x B -> hget s' x = hget s x /\
	hget s' (addrOffset + x) = hget s (addrOffset + x)) /\
      sweepPost s' B free' ohpSt ohpEnd freePtr' stBot stBuff 
      (objHp B B) A) /\ rget s' rcst0 = rget s rcst0 /\
  calleeSavedOk' s s' /\ rget s rcst1 = rget s' ra /\
  rget s' rroot = rget s rroot.

(** ** Some tactics *)

Ltac regSolver := 
  intro r;intros;induction r;stSimpl; auto.

Ltac regEqSolver :=
  match goal with
    | [Hr : _ |- rget _ ?R = rget _ ?R ] => 
      rewrite (Hr R); try (discriminate || reflexivity)
  end. 

Ltac regRewrite R :=
  match goal with
    | [Hr : _ |- _ ] => 
      rewrite (Hr R); try discriminate
end.

Ltac regRewriteH R H :=
  match goal with
    | [Hr : _ H : _ |- _ ] => 
      rewrite (Hr R) in H; try discriminate
end.

Ltac calleeSavedOkSolver' :=
  intros r Hx; destruct r; simpl; trivial;
    destruct Hx as [H | [H | [H | [H | [
      H | [H | [H | [H | H]]]]]]]];
    try (discriminate H ||
      elimtype False; apply H). 

Ltac calleeSavedOkSolver :=
  intros r Hx; destruct r; simpl; trivial;
    destruct Hx as [H | [H | [H | [H | [
      H | [H | [H | [H | [H | H]]]]]]]]];
      try (discriminate H || 
	elimtype False; apply H). 

(** ** Well-formedness theorem for return blocks *)

Definition returnP : chspec := fun f => NoneT _.

Ltac PsimplReturn := idtac.

Theorem returnImplOk : 
  iseqOkS returnP returnPre returnGuar returnImpl.
Proof.
  unfold returnImpl.
  iseqWeakSolver PsimplReturn.
  intuition.
Qed.

Theorem return0ImplOk : 
  iseqOkS returnP return0Pre return0Guar return0Impl.
Proof.
  unfold return0Impl.
  iseqWeakSolver PsimplReturn.
  unfold return0Pre; unfold return0Guar.
  intros.
  rewrite addiuStep;stSimpl.
  intuition.
Qed.

Theorem return1ImplOk : 
  iseqOkS returnP return1Pre return1Guar return1Impl.
Proof.
  unfold return1Impl.
  iseqWeakSolver PsimplReturn.
  unfold return1Pre; unfold return1Guar.
  intros.
  rewrite addiuStep;stSimpl.
  intuition.
Qed.

Theorem return2ImplOk : 
  iseqOkS returnP return2Pre return2Guar return2Impl.
Proof.
  unfold return2Impl.
  iseqWeakSolver PsimplReturn.
  unfold return2Pre; unfold return2Guar.
  intros.
  rewrite addiuStep;stSimpl.
  intuition.
Qed.

Theorem return3ImplOk : 
  iseqOkS returnP return3Pre return3Guar return3Impl.
Proof.
  unfold return3Impl.
  iseqWeakSolver PsimplReturn.
  unfold return3Pre; unfold return3Guar.
  intros.
  rewrite addiuStep;stSimpl.
  intuition.
Qed.


(** ** Div12 *)

Lemma div12EqP : forall n, div12(12 + n) = S (div12 n).
Proof.
  compute; auto.
Qed.

Lemma div12EqT : forall n, div12(12 * n) = n.
Proof.
  induction n; auto.
  replace (12 * S n) with (12 + 12 * n).
  rewrite div12EqP; auto.
  omega.
Qed.

Lemma div3IntroM : forall k n, 
  k = n - 3 -> n >= 3 -> S (div3 k) = div3 n.
Proof.
  intros.
  assert (n = 3 + k).
  omega.
  rewrite H1; simpl; omega.
Qed.

Lemma div3IntorE : forall k n,
  3 * k <= n < 3 * S k -> div3 n = k.
Proof.
  induction k; intros.
  assert (n = 0 \/ n = 1 \/ n = 2).
  omega.
  destruct H0 as [H0 | [H0 | H0]];
    subst n; compute; auto.
  assert (3 * k <= n - 3 < 3 * (S k)).
  omega.
  assert (X := IHk _ H0).
  assert (S (div3 (n - 3)) = div3 n).
  apply div3IntroM; auto.
  omega.
  subst k; auto.
Qed.
 
Lemma div12IntroM : forall k n, 
  k = n - 12 -> n >= 12 -> S (div12 k) = div12 n.
Proof.
  intros.
  assert (n = 12 + k).
  omega.
  rewrite H1; rewrite div12EqP; reflexivity.
Qed.

Lemma div12IntroE : forall k n,
  12 * k <= n < 12 * S k -> div12 n = k.
Proof.
  induction k; intros.
  assert (n = 0 \/ n = 1 \/ n = 2 \/ n = 3 \/
    n = 4 \/ n = 5 \/ n = 6 \/ n = 7 \/
    n = 8 \/ n = 9 \/ n = 10 \/ n = 11).
  omega.
  decompose [or] H0; try (subst n; compute; auto).
  assert (12 * k <= n - 12 < 12 * (S k)).
  omega.
  assert (X := IHk _ H0).
  assert (S (div12 (n - 12)) = div12 n).
  apply div12IntroM; auto.
  omega.
  subst k; auto.
Qed.

Lemma exists12 : forall n, exists k,
  12 * k <= n < 12 * S k.
Proof.
  induction n.
  exists 0;omega.
  destruct IHn as [k Hk].
  compare n ((12 * S k) - 1); intro e.
  exists (S k); omega.
  exists k;omega.
Qed.

Lemma div12Le : forall n, 12 * (div12 n) <= n.
Proof.
  intros.
  destruct (exists12 n) as [k1 Hk1]. 
  rewrite (div12IntroE k1 n Hk1).
  omega.
Qed.
    
Lemma div12IntroLe : forall n m,
  n <= m -> div12 n <= div12 m.
Proof.
  intros.
  destruct (exists12 n) as [k1 Hk1]. 
  destruct (exists12 m) as [k2 Hk2]. 
  assert (k1 <= k2).
  omega.
  rewrite (div12IntroE k1 n Hk1).
  rewrite (div12IntroE k2 m Hk2).
  apply H0.
Qed.

Lemma div12IntroLt : forall n m, 
  12 * S n <= m  -> n < div12 m.
Proof.
  intros.
  assert (X := div12IntroLe _ _ H).
  rewrite div12EqT in X.
  omega.
Qed.


Lemma calleeSavedOkTrans' : forall s s' s'', 
  calleeSavedOk' s s' -> calleeSavedOk' s' s'' -> calleeSavedOk' s s''.
Proof.
  intros s s' s'' H1 H2 r.
  generalize (H1 r).
  generalize (H2 r).
  destruct r;simpl;intuition.
Qed.


Lemma raAndOthers : forall s s',
  calleeSavedOk' s s' -> rget s ra = rget s' ra -> calleeSavedOk s s'.
Proof.
  intros s s' H' Hra r Hr.  
  generalize (H' r) Hr.
  unfold calleeSaved'; unfold calleeSaved.
  destruct r;simpl;trivial;intro X;intro Hx;
    destruct Hx as [He | Hx]; try inversion_clear He; try apply (X Hx).
Qed.

Lemma othersNotRa : forall s s',
  calleeSavedOk s s' -> calleeSavedOk' s s'.
Proof.
  intros s s' H' r Hr.  
  generalize (H' r) Hr.
  unfold calleeSaved'; unfold calleeSaved.
  destruct r;simpl;trivial;intro X;intro Hx;
    apply X;right;apply Hx.
Qed.

(** ** rangeMSObjs *)

Lemma MSPtrGtFour : forall p x y objs, set_In p objs -> 
  objs = rangeMSObjs x y -> p >= 4.
Proof.
  intros p x y.
  unfold rangeMSObjs. 
  induction (div12 (y - x)).
  intros objs Hin e;rewrite e in Hin.
  inversion_clear Hin.
  intros objs Hin e; rewrite e in Hin.
  inversion_clear Hin.
  omega.
  eapply IHn;auto.
  apply H.
Qed.

Lemma MSPtrNotAtom : forall p x y objs, set_In p objs -> 
  ptrP x -> objs = rangeMSObjs x y -> ~ atomP p.
Proof.
  intros p x y.
  unfold rangeMSObjs. 
  induction (div12 (y - x)).
  intros objs Hin Hp e;rewrite e in Hin.
  inversion_clear Hin.
  intros objs Hin Hp e; rewrite e in Hin.
  destruct (oddNotEven p) as [[_ X] _];apply X;clear X.
  simpl in Hin.
  destruct Hin.
  rewrite <- H.
  clear IHn e objs y p H.
  induction Hp.
  induction n.
  simpl.
  repeat constructor.
  simpl in IHn.
  inversion_clear IHn.
  inversion_clear H.
  inversion_clear H0.
  inversion_clear H.
  replace  (0 + 4 + (S n + (S n + (S n + (S n + 
    (S n + (S n + (S n + (S n + (S n + (S n + (
      S n + (S n + 0))))))))))))) with (16 + 12 * n).
  simpl.
  repeat constructor.
  apply H0.
  omega.
  simpl.
  repeat constructor.
  apply IHHp.
  destruct (oddNotEven p) as [[X _] _].
  apply X;clear X.
  apply IHn with (objs := set_map 
    (fun k : POnat.POnat.A => x + 4 + 12 * k) 
    (setZeroRange n));auto.
Qed.

Lemma rangeMSObjsIntro : forall n st ed k,
  ed >= n + 8 -> n = st + 4 + 12 * k ->
  set_In n (rangeMSObjs st ed).
Proof.
  intros.
  rewrite H0.
  unfold rangeMSObjs.
  apply set_mapIn with 
    (f := fun k => st + 4 + 12 * k).
  apply setZeroRangeIntro.
  rewrite H0 in H.
  assert (12 * (S k) <= ed - st).
  omega.
  apply div12IntroLt; apply H1.
  split; intros.
  subst z;reflexivity.
  omega.
Qed.


Lemma rangeMSObjsElim : forall n st ed,
  set_In n (rangeMSObjs st ed) ->
  ed >= n + 8 /\ exists k, n = st + 4 + 12 * k.
Proof.
  intros.
  unfold rangeMSObjs in H.
  destruct (set_mapIn' H)
    as [k [Hk1 Hk2]].
  split;intros. 
  subst z;reflexivity.
  omega.
  assert (Hk3 := setZeroRangeElim _ _ Hk2).
  assert (S k <= div12 (ed - st)).
  omega.
  assert (X := div12Le (ed - st)).
  assert (S k <= ed - st).
  omega.
  split.
  omega.
  exists k; apply Hk1.
Qed.


Lemma rangeMSObjsMaxNotIn : forall st ed,
  ~ set_In ed (rangeMSObjs st ed).
Proof.
  intros st ed Hin.
  destruct (rangeMSObjsElim _ _ _ Hin) as [Hed _].
  omega.
Qed.

Lemma rangeMSObjsSubSet : forall st ed ed', 
  ed <= ed' ->
  set_subset (rangeMSObjs st ed) (rangeMSObjs st ed').
Proof.
  intros st ed ed' lt x Hin.
  destruct (rangeMSObjsElim _ _ _ Hin) as [lt' [k Hk]].
  assert (ed' >= x + 8).
  omega.
  apply rangeMSObjsIntro with k; assumption.
Qed.

Lemma rangeMSObjsRemoveMax : forall st ed x k,
  x = st + 4 + 12 * k -> ed < x + 8 ->
  12 + ed >= x + 8 ->
  x::(rangeMSObjs st ed) = rangeMSObjs st (12 + ed).
Proof.
  intros.
  unfold rangeMSObjs.
  replace (12 + ed - st) with (12 + (ed - st)).
  rewrite div12EqP.
  unfold setZeroRange; fold setZeroRange.
  rewrite set_mapCons. 
  assert (div12 (ed - st) = k).
  apply div12IntroE.
  omega.
  rewrite H2; rewrite <- H; reflexivity.
  omega.
Qed.


(** ** objHdrs *)

Add Morphism objHdrs : objHdrsMorph0.
Proof.
  unfold objHdrs.
  intros s s' n h E Hp.
  setoid_rewrite <- E.
  assumption.
Qed.


Definition objHdrsC s n h := h ||- objHdrs s n.

Add Morphism objHdrsC : objHdrsMorph.
Proof.
  unfold objHdrsC.
  unfold hpredTrue.
  intros s s' n h E H.
  setoid_rewrite <- E.
  assumption.
Qed.

Lemma objHdrsCEq : forall s n h,
  (h ||- objHdrs s n) = objHdrsC s n h.
Proof.
  intros; reflexivity.
Qed.

Lemma objHdrsIntro : forall h s x n,
  h ||- x - 4 |-> n ** objHdrs s n ->
    h ||- objHdrs (set_add x s) n.
Proof.
  intros h s x n H.
  unfold objHdrs.
  apply fasIntro.
  assumption.
  intro I.
  hypReplace H (sconjImplR H _ (fun h => fasElim (h := h) I)).
  assert (h ||- ((x - 4|-> n ** x - 4|-> n) ** top)).
  apply (sconjImpl (sconjAssocR H)); trivial.
  intros; linLogIntroOld.
  assert (Hp : h ||- (!False ** top) ** top).
  apply (sconjImplL H0); clear h H H0; intros h H.
  linLogIntroOld; eapply sconjDisj; eauto.
  linLogElimOld Hp; assumption.
Qed.


Lemma objHdrsElim : forall h s x n,
  set_In x s ->
  h ||- objHdrs s n ->
    h ||- x - 4 |-> n ** objHdrs (set_remove x s) n.
Proof.
  unfold objHdrs.
  intros h s x n I Hp.
  hypReplace Hp (fasElim I Hp).
  assumption.
Qed.


Lemma objHdrsDisj : forall h s1 s2 n1 n2 A,
  h ||- objHdrs s1 n1 ** objHdrs s2 n2 ** A ->
    set_disj s1 s2.
Proof.
  unfold objHdrs.
  intros h s1 s2 n1 n2 A Hp.
  eapply (fasFasDisjF (fun x => x - 4)).
  apply Hp.

  simpl; intros; linLogIntroOld; exists n1.
  hypReplace H (unitIR H).
  apply (sconjImplR H); intros; linLogIntroOld.

  simpl; intros; linLogIntroOld; exists n2.
  hypReplace H (unitIR H).
  apply (sconjImplR H); intros; linLogIntroOld.

Qed.


Lemma objHdrsMerge : forall h s1 s2 n,
  h ||- objHdrs s1 n ** objHdrs s2 n ->
  h ||- objHdrs (set_union s1 s2) n.
Proof.
  intros h s1 s2 n H.
  rewrite objHdrsCEq.
  setoid_rewrite (set_canonEqv s1).
  rewrite <- objHdrsCEq.
  assert (C := set_canonCanonP s1).
  
  assert (h ||- objHdrs (set_canon s1) n ** objHdrs s2 n).
  apply (sconjImplL H); clear h H; intros h H.
  rewrite objHdrsCEq.
  setoid_rewrite <- (set_canonEqv s1).
  exact H.

  clear H; rename H0 into H.
  generalize dependent s2.
  generalize dependent h.
  generalize dependent (set_canon s1); clear s1.
  induction 1; simpl; intros h s2 H'.
  apply unitEL; simpl.
  apply (sconjImpl H'); clear h H'; intros h H.
  apply H.
  assumption.
  apply  objHdrsIntro. 
  assert (I : set_In a (set_add a s)).
  left; reflexivity.

  assert (h ||- (a - 4 |-> n ** objHdrs s n) ** objHdrs s2 n).
  apply (sconjImplL H'); clear H'; intros h' H'.
  unfold objHdrs in H'.
  assert (H2 := fasElim I H'); clear H'.
  apply (sconjImplR H2); clear h' H2; intros h' H'.
  simpl in H'.
  destruct (POnat.POnat.eqDec a a).
  assert (~set_In a s).
  intro X.
  assert (Q := H _ X).
  unfold POnat.POnat.LT in Q; omega.
  rewrite <- (set_removeEq H0).
  assumption.
  elimtype False; apply n0; reflexivity.
  clear H' I.
  linLogElimOld H0.

  apply (sconjImplR H0).
  intros.
  apply IHC.
  assumption.
Qed.

Add Morphism bobjHdrs : bodjHdrsMorph.
Proof.
  unfold bobjHdrs.
  intros s s' n1 n2 n3 h E Hp.
  setoid_rewrite <- E.
  assumption.
Qed.


Lemma bobjHdrsElim : forall s h m n p,
  (forall x, set_In x s -> x < p) ->
  h ||- bobjHdrs s p m n ->
    h ||- objHdrs s m.
Proof.
  unfold bobjHdrs; unfold objHdrs; unfold fas; intros.
  assert (forall x, set_In x (set_canon s) -> x < p).
  intros.
  apply H.
  apply (set_canon_elim _ _ H1).
  rename H1 into Hx.
  clear H.
  generalize h m n p H0 Hx.
  clear h m n p H0 Hx.
  induction (set_canon s); intros.
  apply H0.
  assert (IN : set_In a (a::s0)).
  left; reflexivity.
  simpl in H0.
  simpl.
  apply (sconjImpl H0); clear h H0; intros h Hh.
  rewrite blt_true in Hh; auto.
  eapply IHs0.
  apply Hh.
  intros.
  apply Hx.
  right; exact H.
Qed.

Lemma bobjHdrsIntroGe : forall s h m n p,
  (forall x, set_In x s -> x >= p) ->
  h ||- objHdrs s n ->
    h ||- bobjHdrs s p m n.
Proof.
  unfold bobjHdrs; unfold objHdrs; unfold fas; intros.
  assert (forall x, set_In x (set_canon s) -> x >= p).
  intros.
  apply H.
  apply (set_canon_elim _ _ H1).
  rename H1 into Hx.
  clear H.
  generalize h m n p H0 Hx.
  clear h m n p H0 Hx.
  induction (set_canon s); intros.
  apply H0.
  assert (IN : set_In a (a::s0)).
  left; reflexivity.
  simpl in H0.
  simpl.
  apply (sconjImpl H0); clear h H0; intros h Hh.
  rewrite blt_false.
  apply Hh.
  apply Hx; assumption.
  apply IHs0.
  exact Hh.
  intros.
  apply Hx; right; exact H.
Qed.

Lemma bobjHdrsIntro1 : forall h s x p m n,
  x >= p ->
  h ||- x - 4 |-> n ** bobjHdrs s p m n ->
    h ||- bobjHdrs (set_add x s) p m n.
Proof.
  intros h s x p m n X H.
  unfold bobjHdrs.
  apply fasIntro.
  rewrite blt_false; assumption.
  intro I.
  hypReplace H (sconjImplR H _ (fun h => fasElim (h := h) I)).
  assert (h ||- ((x - 4|-> n ** x - 4|-> n) ** top)).
  apply (sconjImpl (sconjAssocR H)); trivial.
  rewrite blt_false; trivial.
  intros; linLogIntroOld.
  assert (h ||- (x - 4 |->_ ** x - 4 |->_) ** top).
  linLogIntroOld.
  exists n.
  apply (sconjImplR (sconjAssocL H0)).
  clear H0; intros.
  linLogIntroOld.
  exists n.
  apply H0.
  apply (sconjNeq (sconjAssocL H1) (refl_equal (x - 4))).
Qed.

Lemma step12Lt : forall x n k k1,
  x = n + 12 * k1 -> x <> (n + 12 * k) ->
  blt_nat x (n + 12 * k) = blt_nat x (12 + n + 12 * k).
Proof.
  intros.
  destruct (POnat.POnat.compareDec x (n + 12 * k)).
  destruct s.
  repeat (rewrite blt_true); trivial.
  omega.
  elimtype False; apply (H0 e).
  repeat (rewrite blt_false); trivial.
  assert (k1 >= S k).
  omega.
  omega.
  omega.
Qed.

Lemma bobjHdrsStep : forall s p m n h k,
  ~ set_In (p + 12 * k) s ->
  (forall x, set_In x s -> exists k', x = p + 12 * k') ->
  h ||- bobjHdrs s (p + 12 * k) m n -> 
    h ||- bobjHdrs s (12 + p + 12 *k) m n.
Proof.
  unfold bobjHdrs; unfold objHdrs; unfold fas; intros.
  assert (forall x, set_In x (set_canon s) -> 
    exists k' : nat, x = p + 12 * k').
  intros.
  apply H0.
  apply (set_canon_elim _ _ H2).
  assert (~ set_In (p + 12 * k) (set_canon s)).
  intro I; apply H.
  apply (set_canon_elim _ _ I).
  generalize h m n p k H1 H2 H3 .
  clear h m n p H0 H1 H2 H3 H k.
  induction (set_canon s);
    intros h m n p k Hs Hx NI.
  apply Hs.

  assert (set_In a (a::s0)).
  left; apply (refl_equal a).
  unfold fas' in *; fold fas' in *; idtac.
  apply (sconjImpl Hs); clear h Hs; intros h Hh.
  destruct (Hx a H) as [k' Hk'].
  rewrite <- (step12Lt a p k k'); trivial.
  intro X.
  rewrite <- X in NI.
  elimtype False; apply (NI H).
  apply IHs0; trivial.
  intros; apply Hx; right; exact H0.
  intro I; apply NI; right; exact I.
Qed.

Lemma rangeMSObjsNoDup : forall x y, 
  set_noDupP (rangeMSObjs x y).
Proof.
  intros x y.
  unfold rangeMSObjs.
  unfold set_noDupP.
  induction (div12 (y - x)).
  simpl;auto.
  unfold setZeroRange;fold setZeroRange.
  rewrite set_mapCons.
  apply set_noDupAdd;auto.
  clear IHn y.
  apply set_mapNotIn with 
    (f := fun k => x + 4 + 12 * k).
  intro X.
  assert (H := setZeroRangeElim _ _ X).
  omega.
  intros.
  split;intro e.
  rewrite e;reflexivity.
  omega.
Qed.
