Require Export WeakSCAP.
Require Export LinLog.
Require Export Machine.
Require Export List.
Require Export SpecUtil.
Require Export GC.
Require Export NatListSet.
Require Export NatUtil.
Require Export EqNat.
Require Export StateTheory.
Require Import Heap.

(** ** Callee saved registers *)

Set Implicit Arguments.
Unset Strict Implicit.

Lemma calleeSavedOkTrans : forall s s' s'', 
  calleeSavedOk s s' -> calleeSavedOk s' s'' -> calleeSavedOk s s''.
Proof.
  intros s s' s'' H1 H2 r.
  generalize (H1 r).
  generalize (H2 r).
  destruct r;simpl;intuition.
Qed.

(** ** StateTheory *)

Lemma heapGetSoundL : forall h1 h2 h x y, 
  Heap.heap_join h1 h2 h -> 
  Heap.get h1 x = Some y ->
  Heap.get h x = Some y.
Proof.
  unfold Heap.get.
  intros h1 h2 h x y.
  case (isPtr x); intros.
  eapply Heap.HeapMap.ABJoin.mapJoinSoundLeft.
  apply H.
  apply H0.
  apply H0.
Qed.

(** ** NatUtils *)

Lemma atomPBool : forall x, 
  (odd x -> isOdd x = true) /\ (even x -> isEven x = true).
Proof.
  induction x.
  split;intro Hx.
  inversion_clear Hx.
  simpl;reflexivity.
  split;intro Hx;inversion_clear Hx;intuition.
Qed.
  
Lemma atomIsOdd : forall x, atomP x -> isOdd x = true.
Proof.
  intro x;destruct (atomPBool x);trivial.
Qed.

Lemma isOddNotEven : forall x, isOdd x = negb (isEven x).
Proof.
  induction x.
  auto.
  simpl.
  rewrite IHx.
  destruct (isEven x);auto.
Qed.

Lemma oddNotEven : forall x, 
  (~ odd x <-> even x) /\ (~ even x <-> odd x).
Proof.
  induction x;split;split;intros;
    try destruct IHx as [[He1 He2] [He3 He4]].
  constructor.
  intros Ho.
  inversion_clear Ho.
  elimtype False;apply H;constructor.
  inversion_clear H.
  constructor.
  intuition.
  intro Hx.
  inversion_clear H.
  inversion_clear Hx.
  apply He4;auto.
  intuition.
  intro Hx. 
  inversion_clear H.
  inversion_clear Hx.
  apply He4;auto.
Qed.
    
Lemma notAtomNotOdd : forall x, 
  ~ atomP x -> isOdd x = false.
Proof.
  intros.
  destruct (atomPBool x).
  rewrite (isOddNotEven x).
  unfold atomP in H.
  assert (even x).
  destruct x;constructor.
  destruct (oddNotEven x) as [_ [Hx _]];apply Hx;clear Hx.
  intro Hx;apply H.
  constructor;apply Hx.
  rewrite (H1 H2);auto.
Qed.

Lemma boolAtomP : forall x, 
  (isOdd x = true -> odd x) /\ (isEven x = true -> even x).
Proof.
  induction x;intuition.
  inversion_clear H.
Qed.

Lemma negb_intro' : forall (A B : Type) (b: bool), 
  (if b then A else B) -> if negb b then B else A.
Proof.
  intros A B b.
  destruct b; simpl; trivial.
Qed.

(** ** CanonListSet *)

Lemma neqNotIn : forall x s, 
  (forall y, set_In y s -> x <> y) -> ~ set_In x s.
Proof.
  intros x s Hy Hx.
  apply (Hy x Hx);reflexivity.
Qed.

Lemma setRangeRemoveMin : forall m n,
  set_remove m (setRange m n) = setRange (S m) n.
Proof.
  unfold setRange; simpl.
  intros; apply removeIsDiffStep.
Qed.


Lemma notEmptyIsCons : forall objs, objs <> empty_set -> 
  exists x, exists objs', objs = x::objs'.
Proof.
  intros objs Hne.
  destruct objs.
  elimtype False;auto.
  exists a;exists objs;tauto.
Qed.

Lemma set_mapCons : forall x s f, 
  set_map f (x::s) = ((f x)::(set_map f s)).
Proof.
  intros.
  simpl;reflexivity.
Qed.

Lemma set_mapNotIn : forall f x s, 
  ~ set_In x s -> 
  (forall y z : nat, y = z <-> f y = f z) -> 
  ~ set_In (f x) (set_map f s).
Proof.
  intros f x s Hin He.
  intros X;apply Hin.
  clear Hin; induction s.
  inversion_clear X.
  rewrite set_mapCons in X.
  destruct X;[left | right].
  destruct (He a x) as [_ X];auto.
  auto.
Qed.


Lemma set_subsetTrans : forall s1 s2 s3,
  set_subset s1 s2 -> set_subset s2 s3 -> set_subset s1 s3.
Proof.
  intros s1 s2 s3 S1 S2 a A1.
  apply (S2 a (S1 a A1)).
Qed.

Lemma eqvUnitIsUnit : forall s, set_eqv nil s -> s = nil.
Proof.
  destruct s; intros.
  reflexivity.
  destruct (H a).
  assert (X := H1 (or_introl _ (refl_equal a))).
  inversion_clear X.
Qed.

Lemma set_unionEmpty : forall s, s = set_union s empty_set.
Proof.
  induction s;auto.
  simpl.
  rewrite <- IHs;auto.
Qed.

Lemma set_mapIn : forall f x s, set_In x s -> 
  (forall y z : nat, y = z <-> f y = f z) -> set_In (f x) (set_map f s).
Proof.
  intros f x s Hin He.
  induction s.
  inversion_clear Hin.
  rewrite set_mapCons.
  destruct Hin;[left | right].
  destruct (He a x) as [_ X];auto.
  auto.
Qed.

Lemma set_mapIn' : forall f x s, 
  set_In x (set_map f s) ->
  (forall y z : nat, y = z <-> f y = f z) -> 
  exists y, x = f y /\ set_In y s.
Proof.
  intros f x s Hin He.
  induction s.
  inversion_clear Hin.
  rewrite set_mapCons in Hin.
  destruct Hin. 
  exists a.
  split; [rewrite H | left]; reflexivity.
  destruct (IHs H) as [y [X1 X2]].
  exists y.
  split.
  rewrite X1; reflexivity.
  right; exact X2.
Qed.

Lemma set_map_remove : forall s x f, 
  (forall y z : nat, y = z <-> f y = f z) ->
  set_remove (f x) (set_map f s) = set_map f (set_remove x s).
Proof.
  induction s; intros.
  reflexivity.
  simpl.
  destruct (POnat.POnat.eqDec x a).
  subst a.
  rewrite <- IHs; trivial.
  destruct (POnat.POnat.eqDec (f x) (f x)).
  clear e.
  reflexivity.
  elimtype False; apply n; reflexivity.
  rewrite set_mapCons.
  destruct (POnat.POnat.eqDec (f x) (f a)).
  elimtype False; apply n.
  destruct (H x a); auto.
  rewrite IHs; trivial.
Qed.


(** ** LinLog *)
Lemma fasIntroL : forall a s p h A,
  h ||- (p a ** fas p s) ** A -> ~set_In a s ->
    h ||- fas p (set_add a s) ** A.
Proof.
  intros.
  eapply sconjImplL.
  apply H.
  intros.
  apply fasIntro;auto.
Qed.

Lemma sconjNeq : forall x y h A, 
  h ||- (x |->_) ** (y |->_) ** A -> x <> y.
Proof.
  intros.
  linLogElim H.
  destruct (sconjAssocR H) as 
    [h' [h'' [_ [Hh _]]]];clear H.
  destruct (sconjElim Hh) as 
    [h1 [h2 [Hjoin [H1 H2]]]];clear h''.
  assert (Heap.get h1 x = Some x0).
  apply contGetEq;apply H1.
  unfold Heap.get in H.
  assert (isPtr x = true).
  destruct (isPtr x);[trivial | discriminate H].
  assert (Heap.HeapMap.ABJoin.inDom (div4 x) h1).
  exists x0.
  rewrite H0 in H;apply H.
  assert (Heap.get h2 y = Some x1).
  apply contGetEq;apply H2.
  unfold Heap.get in H4.
  assert (isPtr y = true).
  destruct (isPtr y);[trivial | discriminate H4].
  assert (Heap.HeapMap.ABJoin.inDom (div4 y) h2).
  exists x1.
  rewrite H5 in H4;apply H4.
  assert (XXX := 
    Heap.HeapMap.ABJoin.joinedMapsDisjointL Hjoin H3 ).
  intro e;subst x;auto.
Qed.

Lemma aconjIntro1 : forall h A, h ||- A -> h ||- A && A.
Proof.
  intros.
  apply aconjIntro; assumption.
Qed.

Lemma aconjERL : forall A B C h, 
  h ||- A ** (B && C) -> h ||- A ** B.
Proof.
  intros.
  destruct (sconjElim H) as [h1 [h2 [J [HA HB]]]].
  apply sconjIntro with h1 h2; trivial. 
  apply (aconjEL HB).
Qed.

Lemma aconjERR : forall A B C h, 
  h ||- A ** (B && C) -> h ||- A ** C.
Proof.
  intros.
  destruct (sconjElim H) as [h1 [h2 [J [HA HB]]]].
  apply sconjIntro with h1 h2; trivial. 
  apply (aconjER HB).
Qed.

Lemma aconjGR : forall A B h, 
  h ||- A ** B -> h ||- A ** (B && B).
Proof.
  intros.
  destruct (sconjElim H) as [h1 [h2 [J [HA HB]]]].
  apply sconjIntro with h1 h2; trivial. 
  apply (aconjIntro HB HB).
Qed.

Ltac sconjLR H :=
  match type of H with
    | _ ||- sconj _ _  => repeat (hypReplace H (sconjAssocR H))
  end.

Ltac sconjRL H :=
  match type of H with
    | _ ||- sconj _ _  => repeat (hypReplace H (sconjAssocL H))
  end.

Ltac aconjG H :=
  match type of H with
    | _ ||- sconj _ _  => sconjLR H; 
      hypReplace H (aconjGR H); sconjRL H
  end.

Ltac aconjE1 H :=
  match type of H with
    | _ ||- sconj _ _  => 
      let Hh := fresh "Hh" in
	(sconjLR H; assert (Hh := aconjERL H); sconjRL Hh)
  end.

Ltac aconjE2 H :=
  match type of H with
    | _ ||- sconj _ _  => 
      let Hh := fresh "Hh" in
	(sconjLR H; assert (Hh := aconjERR H); sconjRL Hh)
  end.

Hint Resolve set_union_intro1 set_union_intro2
  set_add_intro1 set_add_intro2 
  set_remove_intro: setIn.

Ltac elimSetIn' :=
  match goal with
    | [Hs: set_In ?x (set_add ?y ?A) |- set_In ?x _ ] =>
      let e := fresh "e" in 
	let H := fresh "H" in
	  (destruct (set_add_elim Hs) as [e | H]; clear Hs;
	    [(subst x || rewrite e) | idtac]; elimSetIn')
    | [Hs: set_In ?x (?y :: ?A) |- set_In ?x _ ] =>
      let e := fresh "e" in 
	let H := fresh "H" in
	  (destruct (set_add_elim Hs) as [e | H]; clear Hs;
	    [(subst x || rewrite e) | idtac]; elimSetIn')
    | [Hs: set_In ?x (set_union ?A ?B) |- set_In ?x _ ] =>
      destruct (set_union_elim Hs); clear Hs; elimSetIn'
    | [Hs : set_In ?x (set_remove ?y ?A) |- set_In ?x _ ] =>
      destruct (set_remove_elim _ _ _ Hs); clear Hs; elimSetIn'
    | [Hs: set_In ?x empty_set |- _ ] =>
      inversion_clear Hs
    | _ => auto with setIn
  end.

Ltac elimSetIn :=
  match goal with 
    | [|-context [(set_remove ?y ?A)]] => 
      match goal with
	| [|- set_In ?x _ ] => 
	  let e := fresh "e" in 
	    (compare x y; intro e; 
	      [(subst x || rewrite e) |idtac]; elimSetIn')
      end
    | [Heqv : set_eqv (set_union ?A ?B) ?C |- set_In ?x ?C ] => 
      rewrite <- Heqv; elimSetIn'
    | [Heqv : set_eqv ?C (set_union ?A ?B) |- set_In ?x ?C ] => 
      rewrite Heqv; elimSetIn'
    | [Heqv : set_eqv (set_add ?A ?B) ?C |- set_In ?x ?C ] => 
      rewrite <- Heqv; elimSetIn'
    | [Heqv : set_eqv ?C (set_add ?A ?B) |- set_In ?x ?C ] => 
      rewrite Heqv; elimSetIn'
    | _ => elimSetIn'
  end.


(** ** Others *)

Lemma getCanon : forall h x,
  Heap.HeapMap.map_get h x = 
  Heap.HeapMap.map_get (Heap.heap_canon h) x.
Proof.
  intros h x.
  rewrite <- (Heap.HeapMap.canonStep h x 
    (Heap.HeapMap.ASet.set_canon (Heap.HeapMap.map_dom h))).
  reflexivity.
  intros Hx.
  assert (exists y,Heap.HeapMap.map_get h x = Some y).
  destruct ( Heap.HeapMap.map_get h x).
  exists b;auto.
  elimtype False;auto.
  destruct H as [y Hy];clear Hx.
  induction h.
  simpl in Hy;discriminate Hy.
  destruct a.
  simpl in Hy.
  simpl.
  destruct (POnat.POnat.eqDec x a).
  subst a.
  apply set_insert_intro1.
  apply set_insert_intro2.
  apply IHh;auto.
Qed.

Lemma inObjHpGetSome' : forall x s objs objs', 
  heapOf s ||- objHp objs' objs ** top -> set_In x objs -> 
    (exists lchd, exists A, heapOf s ||- (x |-> lchd) ** 
      A /\ okFieldVal objs' lchd) /\ 
    (exists rchd, exists B, heapOf s ||- (4 + x |-> rchd) ** 
      B /\ okFieldVal objs' rchd).
Proof.
  unfold objHp; intros.
  generalize (fasElimL H0 H).
  simpl; unfold okObjAt; unfold okFieldAt.
  intro Hx;linLogElim Hx.
  split.
  exists x0;exists (4 + x |-> x1 **
    (foralls x in set_remove x objs,
      (lexists y, !okFieldVal objs' y ** x |-> y) **
      (lexists y, !okFieldVal objs' y ** 4 + x |-> y)) 
    ** top).
  split;[apply Hx | apply H5].
  assert (Hx' := sconjAssocL (sconjComm Hx)).
  exists x1;exists (((foralls x in set_remove x objs,
    (lexists y, !okFieldVal objs' y ** x |-> y) **
    (lexists y, !okFieldVal objs' y ** 4 + x |-> y)) 
  ** top) ** x |-> x0).
  split;[apply Hx' | apply H8].
Qed.

Lemma inObjHpGetSome : forall x s objs, 
  heapOf s ||- objHp objs objs ** top -> 
    set_In x objs -> 
    (exists lchd, hget s x = Some lchd) /\ 
    (exists rchd, hget s (4 + x) = Some rchd).
Proof.
  unfold objHp; intros.
  generalize (fasElimL H0 H).
  simpl; unfold okObjAt; unfold okFieldAt.
  intro Hx;linLogElim Hx.
  split.
  exists x0;unfold hget;apply (contGetEqL Hx).
  exists x1;unfold hget;apply (contGetEqL  
    (sconjAssocL (sconjComm Hx))).
Qed.


(* for VGC *)

Ltac sconjLift n :=
  match eval compute in n with
    | O => smartAssocR
    | _ => smartAssocR; apply sconjComm; sconjLift (pred n)
  end.

Ltac sconjLiftH n H :=
  match eval compute in n with
    | O => smartAssocL H
    | _ => smartAssocL H; hypReplace H (sconjComm H); sconjLiftH (pred n) H
  end.

Lemma heap_canonEqv' : forall h,
  heap_eqv h (heap_canon h).
Proof.
  intros h k.
  assert (E := heap_canonEqv h (4 * k)).
  unfold get in E.
  rewrite (isPtrEq (ptrP_intro _ (refl_equal (4 * k)))) in E.
  rewrite div44eq in E.
  apply E.
Qed.

Lemma inMapExists : forall (A B:Set) (s: list A) (x:B) (f:A->B),
  In x (List.map f s) -> exists y, x = f y /\ In y s.
Proof.
  induction s; simpl; intros.
  tauto.
  destruct H.
  exists a; intuition.
  destruct (IHs _ _ H) as [y [H1 H2]].
  exists y; intuition.
Qed.

Lemma set_mapCanonP : forall s f,
  set_canonP s ->
  (forall a b, a < b -> f a < f b) ->
  set_canonP (set_map f s).
Proof.
  intros.
  induction H.
  simpl; constructor.
  simpl.
  constructor; intuition.
  destruct (set_map_elim H2) as [x [Hx Hx']].
  hypReplace H1 (H1 _ Hx').
  subst a'.
  apply (H0 _ _ H1).
Qed.

Lemma set_canonPEqvEq : forall s s',
  set_canonP s ->
  set_canonP s' ->
  set_eqv s s' ->
  s = s'.
Proof.
  intros.
  rewrite <- (set_canonCanonPEq H).
  rewrite <- (set_canonCanonPEq H0).
  apply (set_canonEqvEq H1).
Qed.

Lemma set_canonMap : forall s f,
  (forall a b, a < b -> f a < f b) ->
  set_canon (set_map f s) = set_map f (set_canon s).
Proof.
  intros.
  eapply set_canonPEqvEq.
  apply set_canonCanonP.
  apply set_mapCanonP.
  apply set_canonCanonP.
  apply H.
  split; intros.
  destruct (set_map_elim (set_canon_elim _ _ H0)) as [x [H1 H2]].
  eapply set_map_intro.
  apply H1.
  apply (set_canon_intro _ _ H2).
  destruct (set_map_elim H0) as [x [H1 H2]].
  apply set_canon_intro.
  eapply set_map_intro.
  apply H1.
  apply (set_canon_elim _ _ H2).
Qed.

Lemma heap_eqvElim : forall h h' l,
  heap_eqv h h' -> get h l = get h' l.
Proof.
  intros.
  unfold get.
  destruct (isPtr l); auto.
Qed.

Lemma objHpNotInDom : forall objs h x,
  ~ set_In x objs -> 
  ~ set_In x (List.map (fun x=> 4 + x) objs) ->
  h ||- fas (fun x => x |->__) objs ->
    get h x = None.
Proof.
  unfold fas; intros.
  assert (exists s, s = set_canon objs).
  exists (set_canon objs); reflexivity.
  destruct H2 as [s E].
  assert (~ set_In x (set_canon objs)).
  intro F; apply H; apply set_canon_elim; exact F.
  clear H.
  assert (~ set_In x (List.map (fun x: nat => 4 + x) (set_canon objs))).
  intro F; apply H0. 
  rewrite <- set_canonMap in F.
  apply set_canon_elim; exact F.
  intros; omega.
  clear H0.
  generalize h x H H2 H1 .
  rewrite <- E.
  clear h x H1 objs E H2 H.
  induction s; intros.
  simpl in H1.
  unfold hpredTrue in H1; unfold unit in H1.
  rewrite (heap_eqvElim x (heap_canonEqv' h)).
  rewrite <- (heap_eqvElim x H1).
  unfold get; destruct (isPtr x); simpl; trivial.
  assert (x <> a /\ ~ set_In x s).
  split; intro F; apply H2; elimSetIn.
  clear H2; destruct H0.
  assert (x <> 4 + a /\ ~ set_In x (List.map (fun x=>4+x) s)).
  split; intro F; apply H.
  eapply set_map_intro.
  apply F.
  elimSetIn.
  destruct (set_map_elim F) as [b [XX YY]].
  apply set_map_intro with (f := fun x => 4 + x) (b := b); elimSetIn.
  clear H; destruct H3.
  unfold fas' in H1; fold fas' in H1.
  destruct (sconjElim H1) as [ha [hb [J [Hpa Hpb]]]]; clear H1.
  unfold get.
  destruct (ptrPDec x).
  rewrite (isPtrEq H1).
  eapply (HeapMap.ABJoin.mapJoinSoundNone J).
  destruct (sconjElim Hpa) as [h1 [h2 [J1 [Hp1 Hp2]]]]; clear Hpa.
  clear IHs Hpb.
  linLogElim Hp1.
  linLogElim Hp2.
  unfold cont in * ; unfold hpredTrue in * .
  destruct Hp1; destruct Hp2.
  destruct (ptrP_elim H1).
  destruct (ptrP_elim H4).
  destruct (ptrP_elim H6).
  eapply (HeapMap.ABJoin.mapJoinSoundNone J1);
    rewrite heap_canonEqv'.
  rewrite H5.
  simpl.
  subst x a; repeat (rewrite div44eq).
  destruct POnat.POnat.eqDec.
  subst x2.
  elimtype False; apply H0; reflexivity.
  reflexivity.
  rewrite H7.
  subst x; rewrite H10.
  rewrite div44eq.
  simpl.
  replace (x4 + (x4 + (x4 + (x4 + 0)))) with (4 * x4);
    [rewrite div44eq | omega].
  destruct POnat.POnat.eqDec.
  subst x2.
  elimtype False; apply H; auto.
  reflexivity.
  hypReplace IHs (IHs _ _ H3 H2 Hpb).
  unfold get in IHs.
  rewrite (isPtrEq H1) in IHs.
  apply IHs.
  rewrite (isPtrEq' H1); reflexivity.
Qed.

Lemma existsEqvHp : forall s s' vh objs,
  s' |||- heap_eqv vh ** top ->
    vh ||- objHp objs objs ->
      (forall x, set_In x objs ->
	hget s' x = hget s x /\ hget s' (4 + x) = hget s (4 + x)) ->
      s |||- heap_eqv vh ** top.
Proof.
  intros s s' vh objs Hp' V E.
  assert (subHeap vh (heapOf s)).
  red; intros.
  destruct (sconjElim Hp') as [ha [hb [J [Hpa Hpb]]]].
  unfold hpredTrue in Hpa.
  destruct (set_In_dec l objs).
  hypReplace E (E _ s0).
  generalize (heap_joinSub J l).
  rewrite heap_canonEqv.
  rewrite <- (heap_eqvElim l Hpa).
  intros.
  hypReplace H0 (H0 v H).
  destruct E.
  unfold hget in * .
  rewrite <- H1; exact H0.
  unfold objHp in V.
  destruct (set_In_dec l (List.map (fun x => 4 + x) objs)).
  unfold set_In in s0.
  destruct (inMapExists s0) as [y [XX YY]].
  hypReplace E (E _ YY).
  generalize (heap_joinSub J (4 + y)).
  rewrite heap_canonEqv.
  rewrite <- (heap_eqvElim (4 + y) Hpa).
  subst l.
  intros.
  hypReplace H0 (H0 v H).
  destruct E.
  unfold hget in * .
  rewrite <- H2; exact H0.
  assert (vh ||- fas (fun x => x |->__) objs).
  eapply fasImpl.
  apply V.
  intros.
  unfold okObjAt in H0; unfold okFieldAt in H0; linLogElim H0.
  linLogIntro.
  instantiate (1 := x1).
  instantiate (1 := x0).
  apply H0.
  hypReplace H0 (objHpNotInDom n n0 H0).
  rewrite H in H0; discriminate.
  apply subHeapImpl with vh; auto.
  unfold hpredTrue.
  apply heap_canonEqv'.
Qed.


Lemma reachableNotAtom : forall x y h,
  x <> y -> reachable h x y -> ptrP x.
Proof.
  intros.
  induction H0.
  elimtype False; apply H; trivial.
  unfold get in H0.
  destruct (ptrPDec x); trivial.
  rewrite (isPtrEq' H2) in H0; discriminate.
  unfold get in H0.
  destruct (ptrPDec x); trivial.
  assert (~ ptrP (4 + x)).
  intro F; apply H2.
  inversion_clear F; trivial.
  rewrite (isPtrEq' H3) in H0; discriminate.
Qed.

Lemma ptrNotAtom : forall n,
  ptrP n -> ~ atomP n.
Proof.
  red; intros.
  induction H. 
  inversion H0.
  apply IHptrP.
  inversion_clear H0.
  inversion_clear H1.
  inversion_clear H0.
  inversion_clear H1.
  apply H0.
Qed.

Lemma reachableInObjHp : forall objs h root,
  h ||- objHp objs objs ** top ->
    set_In root objs -> 
    forall x, ptrP x -> reachable h root x -> set_In x objs.
Proof.
  intros.
  induction H2; trivial; unfold objHp in H.
  apply IHreachable; auto.
  hypReplace H (fasElimL H0 H).
  unfold okObjAt in H; unfold okFieldAt in H.
  linLogElim H.
  assert (get h x = Some x0).
  apply (contGetEqL H).
  rewrite H2 in H4; injection H4; intro; subst.
  unfold okFieldVal in H8.
  destruct H8; trivial.
  compare x0 y; intro R; subst.
  hypReplace H1 (ptrNotAtom H1).
  tauto.
  hypReplace H1 (reachableNotAtom R H3).
  hypReplace H1 (ptrNotAtom H1).
  tauto.
  apply IHreachable.
  hypReplace H (fasElimL H0 H).
  unfold okObjAt in H; unfold okFieldAt in H.
  linLogElim H.
  assert (get h (4 + x) = Some x1).
  sconjLiftH 1 H.
  apply (contGetEqL H).
  rewrite H2 in H4; injection H4; intro; subst.
  destruct H11; trivial.
  compare x1 y; intro R; subst.
  hypReplace H1 (ptrNotAtom H1).
  tauto.
  hypReplace H1 (reachableNotAtom R H3).
  hypReplace H1 (ptrNotAtom H1).
  tauto.
  auto.
Qed.

Lemma inObjHpIsPtr : forall x h objs,
  h ||- objHp objs objs ** top -> set_In x objs -> ptrP x.
Proof.
  intros.
  hypReplace H (fasElimL H0 H).
  unfold okObjAt in H; unfold okFieldAt in H.
  linLogElim H.
  destruct (sconjElim H) as [ha [hb [J [Hpa _]]]].
  unfold cont in Hpa.
  unfold hpredTrue in Hpa.
  tauto.
Qed.

Lemma reachableEqv : forall h h' x y, 
  reachable h' x y -> 
  heap_eqv h h' ->
  reachable h x y.
Proof.
  intros.
  induction H.
  constructor.
  rewrite <- (heap_eqvElim x H0) in H.
  apply (reachStep1 _ H IHreachable).
  rewrite <- (heap_eqvElim (4 + x) H0) in H.
  apply (reachStep2 _ H IHreachable).
Qed.

Lemma subHeapEqv : forall h1 h2 h,
  subHeap h1 h -> heap_eqv h1 h2 -> subHeap h2 h.
Proof.
  red; intros.
  hypReplace H (H l v).
  rewrite (heap_eqvElim l H0) in H; auto.
Qed.
  
Lemma contContFalse : forall h a x y A,
  h ||-  a |-> x ** a |-> y ** A -> False.
Proof.
  intros.
  hypReplace H (sconjAssocR H).
  destruct H as [h0 [h1 [_ [Hp _ ]]]].
  clear h h1 A.
  destruct (sconjElim Hp) as [ha [hb [J [[A Hpa] [B Hpb]]]]].
  inversion_clear J.
  hypReplace H (H (div4 a)).
  replace (HeapMap.MapDef.get ha (div4 a)) with (Some x) in H.
  replace (HeapMap.MapDef.get hb (div4 a)) with (Some y) in H.
  inversion H.
  rewrite (heap_canonEqv' hb).
  rewrite Hpb.
  simpl; destruct POnat.POnat.eqDec; auto; tauto.
  rewrite (heap_canonEqv' ha).
  rewrite Hpa.
  simpl; destruct POnat.POnat.eqDec; auto; tauto.
Qed.

