Require Import WeakSCAP.
Require Import LinLog.
Require Import Machine.
Require Import List.
Require Import SpecUtil.
Require Import GC.
Require Import NatListSet.
Require Import NatUtil.
Require Import EqNat.
Require Import StateTheory.
Require Import BuffArrayStack.


(* set implemented on top of a stack *)

Definition buffStackSet x y z E :=
  lexists E', !(set_eqv E E' /\ set_noDupP E') **
    buffArrayStack x y z E'.

Definition setHp objs s :=
  buffStackSet (rget s rbot) (rget s rtop) (rget s rbuff) objs.

Notation setPre :=
  (fun s => exists objs, s |||- setHp objs s ** top).


(** *** Set specifications *)

(** isEmpty() *)
Definition stackIsEmptyPre s :=
  exists objs, s |||- setHp objs s ** top.

Definition stackIsEmptyGuar s s' :=
  (forall objs A, s |||- setHp objs s ** A ->
    sbool (rget s' v0) (objs = empty_set)) /\ 
  heapOf s = heapOf s' /\
  (forall r, r <> v0 -> rget s r = rget s' r).


(** push(), no duplication is allowed *)
Definition stackPushPre s :=
  exists objs, ~ set_In (rget s a0) objs /\
    s |||- setHp objs s ** top.

Definition stackPushGuar s s' :=
  (forall objs A, s |||- setHp objs s ** A -> 
    ~ set_In (rget s a0) objs ->
    s' |||- setHp (set_add (rget s a0) objs) s' ** A) /\
  (forall r, r <> rtop -> r <> rtemp -> rget s r = rget s' r).


(** pop() *)
Definition stackPopPre s :=
  exists objs, (objs <> empty_set) /\
    s |||- setHp objs s ** top.

Definition stackPopGuar s s' :=
  (forall objs A, s |||- setHp objs s ** A -> 
    objs <> empty_set ->
    set_In (rget s' v0) objs /\
    s' |||- setHp (set_remove (rget s' v0) objs) s' ** A) /\
  (forall r, r <> v0 -> r <> rtop -> rget s r = rget s' r).


(* fairly weak, but all we need for this *)
Notation specWeaker :=
  (fun (p:spred) (g:guar) (p':spred) (g':guar) => 
    (forall s, p s -> p' s) /\ 
    (forall s s', g' s s' -> g s s')).


Definition BSSC a b c (d:set) h := 
  h ||- buffStackSet a b c d.

Definition BSSCeq : forall a b c d h,
  (h ||- buffStackSet a b c d) = BSSC a b c d h.
Proof.
  reflexivity.
Qed.


Add Morphism BSSC : BSSCmorph.
  unfold BSSC; unfold buffStackSet.
  intros a b c X X' h Xeq Hp.
  linLogElimOld Hp.
  rename x into X2; destruct H1 as [X2eq X2nd].
  existsl X2; linLogIntroOld.
  constructor.
  setoid_rewrite <- Xeq; assumption.
  assumption.
  assumption.
Qed.


Lemma set_eqvEmptyEmpty : forall x,
  set_eqv x empty_set ->
  x = nil.
Proof.
  destruct x.
  auto.
  intros E; elimtype False.
  destruct (E a) as [X _].
  apply X; apply set_add_intro1; reflexivity.
Qed.


Theorem emptyStackIsEmptySet : 
  specWeaker stackIsEmptyPre stackIsEmptyGuar isEmptyPre isEmptyGuar.
Proof.
  simpl.
  unfold stackIsEmptyPre; unfold isEmptyPre.
  unfold stackIsEmptyGuar; unfold isEmptyGuar.
  unfold setHp; unfold buffStackSet.
  unfold stackHp.
  
  constructor.
  intros s [objs Hp].
  linLogElimOld Hp.
  rename x into E; destruct H3 as [Seq Snd].
  exists E; assumption.
  
  intros s s' [F [Heq Req]].
  constructor.
  intros objs A Q.
  linLogElimOld Q.
  rename x into E; destruct H3 as [Seq Snd].
  assert (sbool (rget s' v0) (E = nil)).
  apply (F E A); clear F.
  assumption.
  clear F Q.
  destruct (rget s' v0); simpl; simpl in H.
  intro.
  subst.
  apply H.
  apply set_eqvEmptyEmpty.
  apply set_eqvSymm; assumption.
  subst.
  apply set_eqvEmptyEmpty.
  assumption.

  intuition.

Qed.


Theorem pushStackIsPushSet : 
  specWeaker stackPushPre stackPushGuar pushPre pushGuar.
Proof.
  simpl.
  unfold stackPushPre; unfold pushPre.
  unfold stackPushGuar; unfold pushGuar.
  unfold setHp; unfold buffStackSet; unfold stackHp.

  constructor.
  intros s [objs [Ni Hp]].
  linLogElimOld Hp.
  rename x into E; destruct H3 as [Seq Snd].
  exists E; assumption.
  
  intros s s' [F Req].
  constructor.
  2:assumption.
  intros objs A Q Ni.
  linLogElimOld Q.
  rename x into E; destruct H3 as [Seq Snd].
  assert (Q' := F _ _ Q); clear F Q.
  linLogIntroOld.
  existsl (set_add (rget s a0) E); linLogIntroOld.
  constructor.
  setoid_rewrite Seq; apply set_eqvRefl.
  apply set_noDupAdd.
  assumption.
  intro X; apply Ni; setoid_rewrite Seq; assumption.
  assumption.
Qed.


Theorem popStackIsPopSet : 
  specWeaker stackPopPre stackPopGuar popPre popGuar.
Proof.
  simpl.
  unfold stackPopPre; unfold popPre.
  unfold stackPopGuar; unfold popGuar.
  unfold setHp; unfold buffStackSet; unfold stackHp.

  constructor.
  intros s [objs [Ni Hp]].
  linLogElimOld Hp.
  rename x into E; destruct H3 as [Seq Snd].
  exists E.
  constructor.
  intro R; apply Ni.
  subst; apply set_eqvEmptyEmpty; assumption.
  assumption.

  intros s s' [F Req].
  constructor.
  2:assumption.
  intros objs A Hp Ne.
  linLogElimOld Hp.
  rename x into E; destruct H3 as [Seq Snd].
  destruct E.
  elimtype False; apply Ne.
  apply set_eqvEmptyEmpty; auto.
  destruct (F _ _ _ Hp) as [Hp' Aeq]; clear F Hp.
  subst a.
  constructor.
  setoid_rewrite Seq.
  apply set_add_intro1; reflexivity.

  linLogIntroOld.
  existsl E; linLogIntroOld.
  constructor.
  setoid_rewrite Seq.
  simpl.
  destruct (POnat.POnat.eqDec (rget s' v0) (rget s' v0)).
  clear e.
  rewrite set_removeEq.
  apply set_eqvRefl.
  apply set_noDupElim.
  assumption.
  elimtype False; auto.
  eapply set_noDupRemove; eauto.
  assumption.
Qed.


